/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef CFSRW_H
#define CFSRW_H

#include "nfs3_prot.h" // nfsstat3
#include "ranges.h"
#include "qhash.h"
#include "view.h"
#include "vervec.h"
#include "cfsrw_prot.h"

template<>
struct hashfn<bigint> {
  hash_t operator() (const bigint &id) const { return id.getui (); }
};

class cfsrw_states {
public:
  friend class cfsrw;

  struct attr {
    struct cfsrw_sattr sattr;
    cfsrw_ftype type;
    uint32 ctime;
    bool mver_set;
    vervec mver;
    int32 nlink;
    bool truncated;
    attr();
    void getfa (cfsrw_fattr &) const;
    void update (cfsrw_sattr &);
  };

  struct lookup {
    bigint dir;
    str name;
    ptr<vec<bigint> > ci;
    lookup (bigint d, str n, ptr<vec<bigint> > v)
      : dir(d), name(n), ci(v) {}
    ~lookup () {}
  };

  struct read {
    bigint object;
    uint64 offset;
    uint32 count;
    uint64 size;
    bool valid_size;
    ranges *rng;
    char *buf;
    uint32 rcount;
    bool eof;
    read (bigint i, uint64 o, uint32 c)
      : object(i), offset(o), count(c), size(0), valid_size(false),
        rcount(0), eof(false)
    {
      if (count > 8192) count = 8192;
      rng = New ranges;
      buf = New char[count];
    }
    ~read () { if (buf) delete buf; if (rng) delete rng; }
  };

  struct readdir {
    unsigned int count;
    uint64 cookie;
    unsigned int n;
    cfsrw_readdir_res *res;
    rpc_ptr<cfs_rpc_dirent> *direntp;
    vec<recordp> changed;

    readdir (uint64 c, unsigned cnt, cfsrw_readdir_res *r)
    {
      res = r;
      res->reply.eof = false;
      direntp = &(res->reply.entries);
      cookie = c;
      count = cnt;
      n = 0;
    }
    ~readdir() {}
  };

  typedef callback<void, ptr<attr>, cfsrw_status>::ref cb_attr;
  typedef callback<void, bigint, cfsrw_status>::ref cb_fh;
  typedef callback<void, recordp, cfsrw_status>::ref cb_lookup;
  typedef callback<void, ptr<read>, cfsrw_status>::ptr cb_read;
  typedef callback<void, ptr<readdir>, cfsrw_status>::ref cb_readdir;
};
 
class write_log_record;
class iterinfo;

class cfsrw
{
private:
  ref<view> _view;
  bigint _root;
  vervec _version;
  bigint _verf;
  qhash<bigint, bool> _locks;

  void find_inode (vervec start, vervec stop,
                   bigint object, ptr<cfsrw_states::attr>,
		   cfsrw_states::cb_attr cb);
  void find_inode_log_iterator (bigint, cfsrw_states::cb_attr,
                                ptr<cfsrw_states::attr>, iterinfo*);
  void found_inode (cfscallp c, bool s, ptr<cfsrw_states::attr> ca,
                    cfsrw_status status);

  void find_parent (vervec start, vervec stop, bigint object,
                    cfsrw_states::cb_fh cb);
  void find_parent_log_iterator (bigint, cfsrw_states::cb_fh, iterinfo*);
  void found_parent (cfscallp c, bool s, bigint parent, cfsrw_status status);
  
  void lookup (vervec start, vervec stop, bigint dir, str name, 
               cfsrw_states::cb_lookup cb, bool check_intend = false);
  void lookup_log_iterator (cfsrw_states::cb_lookup cb,
                            ref<cfsrw_states::lookup> lu, iterinfo*);
  void lookup_found (cfscallp c, bool s, recordp e, cfsrw_status status);

  void read (vervec start, vervec stop,
	     ptr<cfsrw_states::read>, cfsrw_states::cb_read cb);
  void do_read_got_chunk (ptr<cfsrw_states::read>,
                          cfsrw_states::cb_read, iterinfo*);
  void sfrxx (ptr<cfsrw_states::read>, write_log_record *,
              uint64, uint64);
  void read_done (cfscallp c, bool s,
                  ptr<cfsrw_states::read>, cfsrw_status status);
  void readlink_done (cfscallp c, bool s,
                      ptr<cfsrw_states::read> rs, cfsrw_status status);
  
  void scan_dir_log_iterator (bigint dir, ptr<cfsrw_states::readdir>,
                              cfsrw_states::cb_readdir cb, iterinfo*);
  bool readdir_check_dir_entry (ptr<cfsrw_states::readdir>,
                                bigint dir, cfsrw_states::cb_readdir cb,
				str name, bigint object, bool finished);
  void readdir (vervec start, vervec stop,
	        bigint dir, ptr<cfsrw_states::readdir>,
		cfsrw_states::cb_readdir);
  void readdir_done (cfscallp c, bool s, ptr<cfsrw_states::readdir>,
                     cfsrw_status);

  void nores_reply  (cfscallp c, log_status);
  void create_reply (cfscallp c, ptr<unsigned> status,
                     ptr<int> n, log_status);
  void symlink_reply (cfscallp c, ptr<unsigned> status,
                      ptr<int> n, log_status);

  bool test_and_set_excl (bigint object);
  void clear_excl (bigint object);
  void cancel_zombie_intend (cfscallp c, bool again, log_status);
  void excl_cancel (cfscallp c, cfsrw_status s, log_status);
  void excl_intend (cfscallp c, log_status);
  void excl_fetch (cfscallp c, vervec start, log_status s);
  void excl_found (cfscallp c, bool checked_snapshot, bool again,
                   recordp e, cfsrw_status status);

public:
  cfsrw (ref<view> v);
  bigint verf () const { return _verf; }

  void dispatch (cfscallp c);
};

#endif

