/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <dhash/dhash.h>
#include "definc.h"
#include "lrucache.h"
#include "ns_cfsrw.h"
#include "cfs.h"

//
// caching policy
//
// three separate LRU caches. one for log records. one for cfs blocks.
// and one for all the other type of dhash blocks.
//
// we cache both immutable log records and public key log head
// records. we only cache public key log head records that we insert.
// this means we will NOT cache any log head records that we did not
// insert. furthermore, cached log head records expire after 60
// seconds. cached immutable log records are returned to the caller
// when retrieved. copies of cached log head records are returned to
// the caller when retrieved.
//
// we provide two methods for retrieving cfs blocks, one that makes a
// copy of the block before returning (retrieve_cfs), and one that
// does not (retrieve_cfs_readonly).
//
// the has_cfs (bigint id) method returns true if the cfs block with
// the given id exists in the cache (i.e.  it has been recently
// inserted or retrieved). calling this method also moves the cfs
// block to the tail of the LRU list (i.e. it is touched). the
// has_block (bigint id) and has_record (bigint id) methods have
// similar semantics.
//


#define PK_CACHETIME 60  // cache pk blocks for 60 seconds

static int precsz = ENV("CFSRW_PRINT_RECORD_SIZE", 0);
static int pcsize = ENV("CFSRW_PCACHE", 32);
static int rcsize = ENV("CFSRW_RCACHE", 512);
static int scsize = ENV("CFSRW_SCACHE", 256);
static int bcsize = ENV("CFSRW_BCACHE", 16);

static int pk_use_backup   = ENV("CFSRW_PK_USE_BACKUP", 1);
static int pk_skip_lookup  = ENV("CFSRW_PK_SKIPLOOKUP", 1);
static int rec_skip_lookup = ENV("CFSRW_REC_SKIPLOOKUP", 0);


struct cached_block {
  enum cached_type {
    DHASH = 0,
    LOG,
    PKEY,
    CFS
  };
  unsigned short type;
  unsigned short touches;
  long version;
  time_t written;
  ptr<dhash_block> block;
  ptr<log_record> record;
  ptr<cfs_data> cfs;
  
  cached_block () { touches = 0; written = 0; version = 0; }
};

struct pk_insert_state {
  bigint hash;
  sfs_pubkey2 key;
  sfs_sig2 sig;
  long version;
  str data;
  recordp record;
  nscb_insert_t cb;
  bool use_cached_successor;
    
  pk_insert_state (bigint h, sfs_pubkey2 k, sfs_sig2 s, long v,
                   str d, recordp r, nscb_insert_t c, bool ucs)
    : hash(h), key(k), sig(s), version(v), data(d), record(r),
      cb(c), use_cached_successor(ucs) { }
};

static lrucache<bigint, cached_block, hashID> backup_pk_cache (pcsize);
static lrucache<bigint, cached_block, hashID> record_cache (rcsize);
static lrucache<bigint, cached_block, hashID> cfs_cache (scsize);
static lrucache<bigint, cached_block, hashID> block_cache (bcsize);
static dhashclient *dhc = 0;

static int record_insert = 0;
static int record_fetch = 0;
static int cached_record_fetch = 0;
static int cfs_insert = 0;
static int cfs_fetch = 0;
static int cached_cfs_fetch = 0;
static int ch_insert = 0;
static int pk_insert = 0;
static int pk_fetch = 0;

static ptr<cfs_data>
cfs_decode (const char *d, size_t len)
{
  ptr<cfs_data> data = New refcounted<cfs_data>;
  xdrmem x (d, len, XDR_DECODE);
  if (xdr_cfs_data (x.xdrp (), data))
    return data;
  else
    return NULL;
}

static ptr<cfs_data>
cfs_decode (ptr<dhash_block> blk)
{
  return cfs_decode (blk->data, blk->len);
}

void
ns::init()
{
  char *sn = getenv ("LSD_SOCKET");
  if (sn == 0) {
    warnx << "LSD_SOCKET is not set\n";
    exit(1);
  }
  dhc = New dhashclient (getenv ("LSD_SOCKET"));
}

static void
retrieve_block_cb (nscb_retrieve_t cb, bigint id,
                   dhash_stat status, ptr<dhash_block> b, route rt)
{
  if (!status && b && bcsize && !(block_cache[id])) {
    cached_block blk;
    blk.type = cached_block::DHASH;
    blk.block = b;
    blk.record = 0;
    blk.cfs = 0;
    block_cache.insert (id, blk);
  }
  if (status || !b) {
    ptr<ns_block> nsb = 0;
    cb (nsb);
  }
  else {
    ptr<ns_block> nsb = New refcounted<ns_block> (b->data, b->len);
    cb (nsb);
  }
}

static void
return_block (nscb_retrieve_t cb, ptr<dhash_block> b)
{
  assert (b);
  ptr<ns_block> nsb = New refcounted<ns_block> (b->data, b->len);
  cb (nsb);
}

void
ns::retrieve_block (bigint id, nscb_retrieve_t cb)
{
  cached_block *b = block_cache[id];
  if (b && bcsize) {
    if (b->type == cached_block::DHASH) {
      route rt;
      timecb (tsnow, wrap (return_block, cb, b->block));
      return;
    }
    else
      block_cache.remove (id);
  }
  dhc->retrieve (id, wrap (retrieve_block_cb, cb, id));
}

static void
retrieve_record_cb (int tries, cb_recordp cb, bigint id,
                    dhash_stat status, ptr<dhash_block> b, route rt)
{
  if (!status && b) {
    ptr<log_record> e = log_record::parse (str (b->data, b->len));
    if (e && rcsize && !(record_cache[id])) {
      cached_block blk;
      blk.type = cached_block::LOG;
      blk.block = 0;
      blk.record = e;
      blk.cfs = 0;
      record_cache.insert (id, blk);
      cb (e);
    }
    else
      cb (e);
  }
  else
    cb ((ptr<log_record>)0);
}

static void
return_record (cb_recordp cb, ptr<log_record> e)
{
  cb (e);
}

void
ns::retrieve_record (bigint id, cb_recordp cb)
{
  record_fetch++;
  cached_block *b = record_cache[id];
  if (b && rcsize) {
    if (b->type == cached_block::LOG) {
      cached_record_fetch++;
      timecb (tsnow, wrap (return_record, cb, b->record));
      return;
    }
    else
      record_cache.remove (id);
  }
  dhc->retrieve (id, wrap (retrieve_record_cb, 0, cb, id));
}

static void
copy_cfs (nscb_cfs_t cb, ptr<cfs_data> s)
{
  ptr<cfs_data> n = New refcounted<cfs_data> (*s);
  cb (n);
}

static void
retrieve_cfs_cb (nscb_cfs_t cb, bigint id,
                 dhash_stat status, ptr<dhash_block> b, route rt)
{
  if (!status && b) {
    ptr<cfs_data> s = cfs_decode (b);
    if (s && scsize && !(cfs_cache[id])) {
      cached_block blk;
      blk.type = cached_block::CFS;
      blk.block = 0;
      blk.record = 0;
      blk.cfs = s;
      cfs_cache.insert (id, blk);
      copy_cfs (cb, s);
    }
    else
      cb (s);
  }
  else
    cb ((ptr<cfs_data>)0);
}

bool
ns_cfsrw::has_cfs (bigint n)
{
  cached_block *b = cfs_cache[n];
  if (b && b->type == cached_block::CFS) {
    b->touches++;
    if (b->touches < 1024)
      return true;
    else
      cfs_cache.remove (n);
  }
  return false;
}

static void
retrieve_cfs_readonly_cb (nscb_cfs_t cb, bigint id,
                          dhash_stat status, ptr<dhash_block> b, route rt)
{
  if (!status && b) {
    ptr<cfs_data> s = cfs_decode (b);
    if (s && scsize && !(cfs_cache[id])) {
      cached_block blk;
      blk.type = cached_block::CFS;
      blk.block = 0;
      blk.record = 0;
      blk.cfs = s;
      cfs_cache.insert (id, blk);
      cb (s);
    }
    else
      cb (s);
  }
  else
    cb ((ptr<cfs_data>)0);
}

static void
return_cfs (nscb_cfs_t cb, ptr<cfs_data> s)
{
  cb (s);
}

void
ns_cfsrw::retrieve_cfs (bigint id, nscb_cfs_t cb)
{
  cfs_fetch++;
  cached_block *b = cfs_cache[id];
  if (b && scsize) {
    if (b->type == cached_block::CFS) {
      cached_cfs_fetch++;
      timecb (tsnow, wrap (copy_cfs, cb, b->cfs));
      return;
    }
    else
      cfs_cache.remove (id);
  }
  dhc->retrieve (id, wrap (retrieve_cfs_cb, cb, id));
}

void
ns_cfsrw::retrieve_cfs_readonly (bigint id, nscb_cfs_t cb)
{
  cfs_fetch++;
  cached_block *b = cfs_cache[id];
  if (b && scsize) {
    if (b->type == cached_block::CFS) {
      cached_cfs_fetch++;
      timecb (tsnow, wrap (return_cfs, cb, b->cfs));
      return;
    }
    else
      cfs_cache.remove (id);
  }
  dhc->retrieve (id, wrap (retrieve_cfs_readonly_cb, cb, id));
}

static void
insert_ok_cb (nscb_insert_t cb, bigint n)
{
  cb (NS_OK, n);
}

bool
ns::has_block (bigint n)
{
  cached_block *b = block_cache[n];
  if (b && b->type == cached_block::CFS) {
    b->touches++;
    if (b->touches < 1024)
      return true;
    else
      block_cache.remove (n);
  }
  return false;
}

static void
insert_block_cb (ptr<dhash_block> b, nscb_insert_t cb,
                 dhash_stat status, ptr<insert_info> i)
{
  assert (i);
  block_cache.remove (i->key);
  if (status == DHASH_OK && bcsize) {
    cached_block blk;
    blk.type = cached_block::DHASH;
    blk.block = b;
    blk.record = 0;
    blk.cfs = 0;
    block_cache.insert (i->key, blk);
  }
  if (status == DHASH_OK)
    cb (NS_OK, i->key);
  else if (status == DHASH_NOENT)
    cb (NS_NOENT, i->key);
  else
    cb (NS_ERR, i->key);
}

void
ns::insert_block (bigint hash, const char *data, size_t len,
                  nscb_insert_t cb)
{
  if (!has_block (hash)) {
    ch_insert++;
    ptr<dhash_block> b = New refcounted<dhash_block> (data, len);
    dhc->insert (hash, data, len, wrap (insert_block_cb, b, cb));
  }
  else
    timecb (tsnow, wrap (insert_ok_cb, cb, hash));
}

void
ns::insert_block (const char *data, size_t len, nscb_insert_t cb)
{
  ch_insert++;
  ptr<dhash_block> b = New refcounted<dhash_block> (data, len);
  dhc->insert (data, len, wrap (insert_block_cb, b, cb));
}

static void
insert_cfs_cb (ptr<cfs_data> b, nscb_insert_t cb,
               dhash_stat status, ptr<insert_info> i)
{
  assert (i);
  cfs_cache.remove (i->key);
  if (status == DHASH_OK && scsize) {
    cached_block blk;
    blk.type = cached_block::CFS;
    blk.block = 0;
    blk.record = 0;
    blk.cfs = b;
    cfs_cache.insert (i->key, blk);
  }
  if (status == DHASH_OK)
    cb (NS_OK, i->key);
  else if (status == DHASH_NOENT)
    cb (NS_NOENT, i->key);
  else
    cb (NS_ERR, i->key);
}

void
ns_cfsrw::insert_cfs (const char *data, size_t len,
                      nscb_insert_t cb, ptr<cfs_data> b)
{
  if (b == 0)
    b = cfs_decode (data, len);
  ch_insert++;
  dhc->insert (data, len, wrap (insert_cfs_cb, b, cb));
}

void
ns_cfsrw::insert_cfs (bigint hash, const char *data, size_t len,
                      nscb_insert_t cb, ptr<cfs_data> b)
{
  if (!has_cfs (hash)) {
    if (b == 0)
      b = cfs_decode (data, len);
    cfs_insert++;
    dhc->insert (hash, data, len, wrap (insert_cfs_cb, b, cb));
  }
  else
    timecb (tsnow, wrap (insert_ok_cb, cb, hash));
}

static void
insert_record_cb (int tries, recordp e, nscb_insert_t cb,
                  dhash_stat status, ptr<insert_info> i)
{
  assert (i);
  record_cache.remove (i->key);
  if (status == DHASH_OK && rcsize) {
    cached_block blk;
    blk.type = cached_block::LOG;
    blk.block = 0;
    blk.record = e;
    blk.cfs = 0;
    record_cache.insert (i->key, blk);
  }
  if (status == DHASH_OK)
    cb (NS_OK, i->key);
  else if (status == DHASH_NOENT)
    cb (NS_NOENT, i->key);
  else
    cb (NS_ERR, i->key);
}

void
ns::insert_record (const char *data, size_t len,
                   nscb_insert_t cb, recordp e)
{
  if (e == 0) 
    e = log_record::parse (str(data, len));
  assert(!e->template record<head_log_record>());
  record_insert++;
 
  if (precsz)
    warnx << "insert record " << len << " bytes\n";
  bigint id = compute_sha1_hash (data, len);
  int options = 0;
  if (rec_skip_lookup > 0)
    options |= DHASHCLIENT_USE_CACHED_SUCCESSOR;
  dhc->insert (id, data, len, wrap (insert_record_cb, 0, e, cb), options);
}

void
ns::insert_record (bigint id, const char *data, size_t len,
                   nscb_insert_t cb, recordp e)
{
  if (e == 0)
    e = log_record::parse (str(data, len));
  assert(!e->template record<head_log_record>());
  record_insert++;

  if (precsz)
    warnx << "insert record " << len << " bytes\n";
  int options = 0;
  if (rec_skip_lookup > 0)
    options |= DHASHCLIENT_USE_CACHED_SUCCESSOR;
  dhc->insert (id, data, len, wrap (insert_record_cb, 0, e, cb), options);
}

static void
insert_pk_cb (recordp e, long version, nscb_insert_t cb,
              dhash_stat status, ptr<insert_info> i)
{
  assert (i);
  record_cache.remove (i->key);
  if (status == DHASH_OK) {
    if (e) {
      backup_pk_cache.remove (i->key);
      cached_block blk;
      blk.type = cached_block::PKEY;
      blk.block = 0;
      blk.record = e;
      blk.cfs = 0;
      blk.version = version;
      backup_pk_cache.insert (i->key, blk);
    }
    if (rcsize) {
      cached_block blk;
      blk.type = cached_block::PKEY;
      blk.block = 0;
      blk.record = e;
      blk.cfs = 0;
      blk.written = timenow;
      record_cache.insert (i->key, blk);
    }
  }
  if (status == DHASH_OK)
    cb (NS_OK, i->key);
  else if (status == DHASH_NOENT)
    cb (NS_NOENT, i->key);
  else
    cb (NS_ERR, i->key);
}

static void insert_pk (ptr<pk_insert_state> st);

static void
insert_pk_check_cb (ptr<pk_insert_state> st, dhash_stat status,
                    ptr<insert_info> i)
{
  assert (i);
  if (status != DHASH_OK && st->use_cached_successor) {
    warnx << "fast insert of public key failed, try regular insert\n";
    st->use_cached_successor = false;
    insert_pk (st);
    return;
  }
  insert_pk_cb (st->record, st->version, st->cb, status, i);
}

static void
insert_pk (ptr<pk_insert_state> st)
{
  int options = 0;
  if (pk_skip_lookup > 0)
    options |= DHASHCLIENT_USE_CACHED_SUCCESSOR;
  dhc->insert (st->hash, st->key, st->sig,
               st->data.cstr (), st->data.len (), st->version,
               wrap (insert_pk_check_cb, st), options);
}

static void
update_backup_pk_cache (bigint id, ptr<log_record> e, long version)
{
  assert (e);

  backup_pk_cache.remove (id);
  cached_block blk;
  blk.type = cached_block::PKEY;
  blk.block = 0;
  blk.record = e;
  blk.cfs = 0;
  blk.version = version;
  backup_pk_cache.insert (id, blk);
}

static void
fetch_loghead_for_insert (bigint hash, ptr<pk_insert_state> st,
                          dhash_stat status, ptr<dhash_block> b, route rt)
{
  if (!status && b) {
    ptr<log_record> r = log_record::parse (str (b->data, b->len));
    if (r)
      update_backup_pk_cache (hash, r, b->version);
  }

  long v = 0;
  cached_block *blk = backup_pk_cache [hash];
  if (blk)
    v = blk->version+1;
  else
    warn << "cannot fetch " << hash << ", set version to 0\n";
  st->version = v;
  insert_pk (st);
}

void
ns::insert_loghead (bigint hash, sfs_pubkey2 k, sfs_sig2 s,
                    const char *data, size_t len,
		    nscb_insert_t cb, recordp e)
{
  str d (data, len);
  if (e == 0)
    e = log_record::parse (d);
  assert(e->template record<head_log_record>());
  pk_insert++;

  ptr<pk_insert_state> st = New refcounted<pk_insert_state>
    (hash, k, s, 0, d, e, cb, pk_skip_lookup > 0);
  cached_block *blk = backup_pk_cache [hash];
  if (blk) {
    st->version = blk->version+1;
    insert_pk (st);
  }
  else
    dhc->retrieve 
      (hash, wrap (fetch_loghead_for_insert, hash, st),
       DHASHCLIENT_NO_RETRY_ON_LOOKUP);
}

void
ns::insert_loghead (sfs_pubkey2 k, sfs_sig2 s,
                    const char *data, size_t len,
		    nscb_insert_t cb, recordp e)
{
  strbuf b;
  ptr<sfspub> pk = sfscrypt.alloc (k);
  pk->export_pubkey (b, false);
  str pk_raw = b;
  bigint id = compute_sha1_hash (pk_raw.cstr (), pk_raw.len ());
  insert_loghead (id, k, s, data, len, cb, e);
}

static void
copy_record (cb_recordp cb, ptr<log_record> e)
{
  ptr<log_record> n = log_record::dup (e);
  if (!n)
    warnx << "dup failed in cache\n";
  cb (n);
}

typedef callback<void, recordp>::ptr cbptr_recordp;

static void 
retrieve_pk_cb (bigint id, cbptr_recordp cb,
                dhash_stat status, ptr<dhash_block> b, route rt)
{
  if (!status && b) {
    ptr<log_record> e = log_record::parse (str (b->data, b->len));
    if (!e)
      warnx << "parse failed in cache, not returning a log record\n";
    else {
      bool update = true;
      cached_block *blk = backup_pk_cache[id];
      if (blk && blk->type == cached_block::PKEY) {
	recordp r = blk->record;
	if (blk->version > b->version) {
          warnx << "stale log record v" << b->version
	        << ", using cached version v" << blk->version << "\n";
	  e = r;
	  update = false;
	}
      }
      if (update)
	update_backup_pk_cache (id, e, b->version);
    }
    if (cb)
      cb (e);
  }
  else if (cb) {
    recordp r = 0;
    if (pk_use_backup) {
      cached_block *blk = backup_pk_cache[id];
      if (blk && blk->type == cached_block::PKEY) {
	warnx << "use cached pk block " << id << "\n";
	r = blk->record;
      }
    }
    cb (r);
  }
}

static void 
retrieve_pk_fast_cb (bigint id, cb_recordp cb,
                     dhash_stat status, ptr<dhash_block> b, route rt)
{
  if ((!status && b) || status == DHASH_RETRY)
    retrieve_pk_cb (id, cb, status, b, rt);
  else {
    warnx << "fast retrieve on keyhash failed, try regular retrieve\n";
    dhc->retrieve (id, wrap (retrieve_pk_cb, id, cb));
  }
}

void
ns::retrieve_loghead (bigint id, cb_recordp cb)
{
  cached_block *b = record_cache[id];
  if (rcsize && b && b->type == cached_block::PKEY) {
    // if we recently wrote the block, return the cached version
    if (timenow - b->written < PK_CACHETIME) {
      // warnx << "pk block " << id << " recently written\n";
      timecb (tsnow, wrap (copy_record, cb, b->record));
      return;
    }
    record_cache.remove (id);
  }
  pk_fetch++;

  int options = 0;
  if (pk_skip_lookup > 0)
    options |= DHASHCLIENT_USE_CACHED_SUCCESSOR;
  dhc->retrieve (id, wrap (retrieve_pk_fast_cb, id, cb), options);
}

bool
ns::has_record (bigint n)
{
  cached_block *b = record_cache[n];
  if (b && b->type == cached_block::LOG) {
    b->touches++;
    if (b->touches < 1024)
      return true;
    else
      record_cache.remove (n);
  }
  return false;
}

void
ns::stat()
{
  warnx << "record_insert " << record_insert << "\n";
  warnx << "record_fetch " << record_fetch << "\n";
  warnx << "cached_record_fetch " << cached_record_fetch << "\n";
  warnx << "cfs_insert " << cfs_insert << "\n";
  warnx << "cfs_fetch " << cfs_fetch << "\n";
  warnx << "cached_cfs_fetch " << cached_cfs_fetch << "\n";
  warnx << "pk_fetch " << pk_fetch << "\n";
  warnx << "pk_insert " << pk_insert << "\n";
  warnx << "ch_insert " << ch_insert << "\n";
}

