/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <time.h>
#include "definc.h"
#include "pack.h"
#include "cfsrw_rec.h"

ptr<log_record>
log_record::parse (str ins)
{
  ptr<log_record> e = 0;
  xin *in = New xin(ins);
  
  bigint prev;
  bigint head;
  int seqnum;
  int type;
  int sec;
  int vsize;
  vervec version;
 
  if (!in->rbigint (prev))
    goto bad;
  if (!in->rbigint (head))
    goto bad;
  if (!in->rint (seqnum))
    goto bad;
  if (!in->rint (type))
    goto bad;
  if (!in->rint (sec))
    goto bad;
  if (!in->rint (vsize))
    goto bad;
  for (int i=0; i<vsize; i++) {
    bigint h;
    bigint l;
    int s;
    if (!in->rbigint (h))
      goto bad;
    if (!in->rbigint (l))
      goto bad;
    if (!in->rint (s))
      goto bad;
    version.add (h, l, s);
  }

  if (type == log_record::T_END)
    e = New refcounted<end_log_record> (in);
  else if (type == log_record::T_HEAD)
    e = New refcounted<head_log_record> (in);
  else if (type == cfsrw_log_record::T_INODE)
    e = New refcounted<inode_log_record> (in);
  else if (type == cfsrw_log_record::T_LINK)
    e = New refcounted<link_log_record> (in);
  else if (type == cfsrw_log_record::T_UNLINK)
    e = New refcounted<unlink_log_record> (in);
  else if (type == cfsrw_log_record::T_RENAME)
    e = New refcounted<rename_log_record> (in);
  else if (type == cfsrw_log_record::T_WRITE)
    e = New refcounted<write_log_record> (in);
  else if (type == cfsrw_log_record::T_INTEND)
    e = New refcounted<intend_log_record> (in);
  else if (type == cfsrw_log_record::T_ATTR)
    e = New refcounted<attr_log_record> (in);
  else
    warn << "log_record::parse: unknown record type " << type << "\n";
  delete in;

  if (e) {
    e->_prev = prev;
    e->_head = head;
    e->_record_type = type;
    e->_seqnum = (uint)seqnum;
    e->_ts = sec;
    e->_version = version;
    return e;
  }
  else
    return 0;
    
bad:  
  delete in;
  return 0;
}
 
ptr<log_record>
log_record::dup(ptr<log_record> oldr)
{
  ptr<log_record> r = 0;
  if (oldr->record_type () == log_record::T_END)
    r = New refcounted<end_log_record>
      (*oldr->template record<end_log_record> ());
  else if (oldr->record_type () == log_record::T_HEAD)
    r = New refcounted<head_log_record>
      (*oldr->template record<head_log_record> ());
  else if (oldr->record_type () == cfsrw_log_record::T_INODE)
    r = New refcounted<inode_log_record>
      (*oldr->template record<inode_log_record> ());
  else if (oldr->record_type () == cfsrw_log_record::T_LINK)
    r = New refcounted<link_log_record>
      (*oldr->template record<link_log_record> ());
  else if (oldr->record_type () == cfsrw_log_record::T_UNLINK)
    r = New refcounted<unlink_log_record>
      (*oldr->template record<unlink_log_record> ());
  else if (oldr->record_type () == cfsrw_log_record::T_RENAME)
    r = New refcounted<rename_log_record>
      (*oldr->template record<rename_log_record> ());
  else if (oldr->record_type () == cfsrw_log_record::T_WRITE)
    r = New refcounted<write_log_record>
      (*oldr->template record<write_log_record> ());
  else if (oldr->record_type () == cfsrw_log_record::T_INTEND)
    r = New refcounted<intend_log_record>
      (*oldr->template record<intend_log_record> ());
  else if (oldr->record_type () == cfsrw_log_record::T_ATTR)
    r = New refcounted<attr_log_record>
      (*oldr->template record<attr_log_record> ());
  else
    assert (0);
  return r;
}


attr_log_record::attr_log_record (bigint object, cfsrw_sattr attr)
  : log_record(cfsrw_log_record::T_ATTR), _attr(attr)
{
  _object = object;
}

attr_log_record::attr_log_record(xin *in)
{
  in->rbigint (_object);
  _attr.xinxin(in);
}

str
attr_log_record::serialize() const
{
  strbuf sb;

  serialize_meta(sb);
  sbigint (sb, _object);
  _attr.serialize(sb);
  return sb;
}


inode_log_record::inode_log_record (cfsrw_ftype object_type, bigint object)
  : log_record (cfsrw_log_record::T_INODE), _attr ()
{
  _object_type = object_type;
  _object = object;
}

inode_log_record::inode_log_record (cfsrw_ftype object_type, bigint object,
                                    cfsrw_sattr sa)
  : log_record (cfsrw_log_record::T_INODE), _attr (sa)
{
  _object_type = object_type;
  _object = object;
}

inode_log_record::inode_log_record (cfsrw_ftype object_type)
  : log_record (cfsrw_log_record::T_INODE), _attr ()
{
  _object_type = object_type;
  // random new i-number.
  char xxb[20];
  rnd.getbytes(xxb, 20);
  mpz_set_rawmag_be(&_object, xxb, 20);
}

inode_log_record::inode_log_record (cfsrw_ftype object_type, cfsrw_sattr sa)
  : log_record (cfsrw_log_record::T_INODE), _attr (sa)
{
  _object_type = object_type;
  // random new i-number.
  char xxb[20];
  rnd.getbytes(xxb, 20);
  mpz_set_rawmag_be(&_object, xxb, 20);
}

inode_log_record::inode_log_record (xin *in)
{
  int i;
  in->rint(i);
  _object_type = (cfsrw_ftype) i;
  in->rbigint (_object);
  _attr.xinxin(in);
}

str
inode_log_record::serialize () const
{
  strbuf sb;
  serialize_meta(sb);
  sint(sb, _object_type);
  sbigint (sb, _object);
  _attr.serialize(sb);
  return sb;
}


dirop_log_record::dirop_log_record
  (int record_type, bigint object, cfsrw_ftype object_type,
   bigint dir, str name)
  : log_record (record_type),
    _object (object), _object_type (object_type), _dir(dir), _name(name)
{
}

link_log_record::link_log_record
  (bigint object, cfsrw_ftype object_type, bigint dir, str name)
  : dirop_log_record (cfsrw_log_record::T_LINK, object,
                      object_type, dir, name)
{
}

link_log_record::link_log_record(xin *in)
{
  in->rbigint (_object);
  int i;
  in->rint(i);
  _object_type = (cfsrw_ftype) i;
  in->rbigint (_dir);
  in->rstr(_name);
}

str
link_log_record::serialize() const
{
  strbuf sb;
  serialize_meta(sb);
  sbigint (sb, _object);
  sint(sb, (int) _object_type);
  sbigint (sb, _dir);
  sstr(sb, _name);
  return sb;
}

unlink_log_record::unlink_log_record
  (bigint object, cfsrw_ftype object_type, bigint dir, str name)
  : dirop_log_record (cfsrw_log_record::T_UNLINK, object,
                      object_type, dir, name)
{
}

unlink_log_record::unlink_log_record(xin *in)
{
  in->rbigint (_object);
  int i;
  in->rint(i);
  _object_type = (cfsrw_ftype) i;
  in->rbigint (_dir);
  in->rstr(_name);
}

str
unlink_log_record::serialize() const
{
  strbuf sb;
  serialize_meta(sb);
  sbigint (sb, _object);
  sint(sb, (int) _object_type);
  sbigint (sb, _dir);
  sstr(sb, _name);
  return sb;
}

rename_log_record::rename_log_record
  (bigint object, cfsrw_ftype object_type,
   bigint from_dir, str from_name,
   bigint dir, str name, bigint target)

  : dirop_log_record (cfsrw_log_record::T_RENAME, object,
                      object_type, dir, name),
    _from_dir (from_dir), _from_name (from_name),
    _target_exists (true), _target (target)
{
}

rename_log_record::rename_log_record
  (bigint object, cfsrw_ftype object_type,
   bigint from_dir, str from_name,
   bigint dir, str name)
  : dirop_log_record (cfsrw_log_record::T_RENAME, object,
                      object_type, dir, name),
    _from_dir (from_dir), _from_name (from_name), _target_exists (false)
{
}

rename_log_record::rename_log_record(xin *in)
{
  in->rbigint (_object);
  int i;
  in->rint(i);
  _object_type = (cfsrw_ftype) i;
  in->rbigint (_from_dir);
  in->rstr(_from_name);
  in->rbigint (_dir);
  in->rstr(_name);
  int n;
  in->rint(n);
  if (n)
    _target_exists = true;
  else
    _target_exists = false;
  in->rbigint (_target);
}

str
rename_log_record::serialize() const
{
  strbuf sb;
  serialize_meta(sb);
  sbigint (sb, _object);
  sint(sb, (int) _object_type);
  sbigint (sb, _from_dir);
  sstr(sb, _from_name);
  sbigint (sb, _dir);
  sstr(sb, _name);
  if (_target_exists)
    sint(sb, 1);
  else
    sint(sb, 0);
  sbigint (sb, _target);
  return sb;
}


intend_log_record::intend_log_record(bigint v, bigint d, str name, bool c)
  : log_record(cfsrw_log_record::T_INTEND), _verf(v), _dir(d), _name(name)
{
  if (c)
    _canceled = 1;
  else
    _canceled = 0;
}

intend_log_record::intend_log_record(xin *in)
{
  in->rbigint (_verf);
  in->rbigint (_dir);
  in->rstr(_name);
  in->rint(_canceled);
}

str
intend_log_record::serialize() const
{
  strbuf sb;
  serialize_meta(sb);
  sbigint (sb, _verf);
  sbigint (sb, _dir);
  sstr(sb, _name);
  sint(sb, _canceled);
  return sb;
}


write_log_record::write_log_record(bigint file, uint64 offset, str data)
  : log_record(cfsrw_log_record::T_WRITE), _data(data)
{
  _file = file;
  _offset = offset;
}

write_log_record::write_log_record(xin *in)
{
  in->rbigint (_file);
  in->rquad(_offset);
  in->rstr(_data);
}

str
write_log_record::serialize() const
{
  strbuf sb;
  serialize_meta(sb);
  sbigint (sb, _file);
  squad(sb, _offset);
  sstr(sb, _data);
  return sb;
}

