/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef CFSRW_RECORD_H
#define CFSRW_RECORD_H

#include "definc.h"
#include "vervec.h"
#include "pack.h"
#include "record.h"
#include "cfsrw_prot.h"  // cfsrw_sattr

class attr_log_record;
class write_log_record;
class inode_log_record;
class intend_log_record;
class dirop_log_record;
class link_log_record;
class unlink_log_record;
class rename_log_record;

class cfsrw_log_record {
public:
  static const int T_ATTR   = 10;
  static const int T_WRITE  = 11;
  static const int T_INODE  = 12;
  static const int T_INTEND = 13;
  static const int T_LINK   = 14;
  static const int T_UNLINK = 15;
  static const int T_RENAME = 16;
};

class attr_log_record : public log_record
{
private:
  bigint _object;
  cfsrw_sattr _attr;

public:
  attr_log_record (const attr_log_record& r) { *this = r; }
  attr_log_record (bigint, cfsrw_sattr);
  attr_log_record (xin *);

  bigint object () const { return _object; }
  cfsrw_sattr* attr () { return &_attr; }
  str serialize () const;
};

class inode_log_record : public log_record
{
private:
  cfsrw_ftype _object_type;
  bigint _object;
  cfsrw_sattr _attr;

public:
  inode_log_record (const inode_log_record& r) { *this = r; }
  inode_log_record (cfsrw_ftype);
  inode_log_record (cfsrw_ftype, cfsrw_sattr);
  inode_log_record (cfsrw_ftype, bigint);
  inode_log_record (cfsrw_ftype, bigint, cfsrw_sattr);
  inode_log_record (xin *);

  cfsrw_ftype object_type () const { return _object_type; }
  bigint object () const { return _object; }
  cfsrw_sattr* attr () { return &_attr; }
  str serialize () const;
};

class intend_log_record : public log_record
{
private:
  bigint _verf;
  bigint _dir;
  str _name;
  int _canceled;

public:
  intend_log_record (const intend_log_record& r) { *this = r; }
  intend_log_record (bigint, bigint, str, bool);
  intend_log_record (xin *);

  str serialize () const;
  str name () const { return _name; }
  bigint verf () const { return _verf; }
  bigint dir () const { return _dir; }
  bool canceled () const { return (_canceled == 1); }
};

class write_log_record : public log_record
{
private:
  bigint _file;
  uint64 _offset;
  str _data;

public:
  write_log_record (const write_log_record& r) { *this = r; }
  write_log_record (bigint, uint64, str);
  write_log_record (xin *);

  bigint file () const { return _file; }
  uint64 offset () const { return _offset; }
  uint32 count () const { return _data.len(); }
  str data () const { return _data; }
  str serialize () const;
};

class dirop_log_record : public log_record
{
protected:
  bigint _object;
  cfsrw_ftype _object_type;
  bigint _dir;
  str _name;
  
  dirop_log_record (int record_type, bigint object, cfsrw_ftype object_type, 
                    bigint dir, str name);
  dirop_log_record () {}  // used by copy constructors of subclasses

public:
  virtual str serialize () const = 0;
  
  bigint dir () const { return _dir; }
  bigint object () const { return _object; }
  str name () const { return _name; }
  cfsrw_ftype object_type () const { return _object_type; }
};

class link_log_record : public dirop_log_record
{
public:
  link_log_record (const link_log_record &r) { *this = r; }
  link_log_record (bigint object, cfsrw_ftype object_type,
                   bigint dir, str name);
  link_log_record (xin *);
  str serialize () const;
};

class unlink_log_record : public dirop_log_record
{
public:
  unlink_log_record (const unlink_log_record &r) { *this = r; }
  unlink_log_record (bigint object, cfsrw_ftype object_type,
                     bigint dir, str name);
  unlink_log_record (xin *);
  str serialize () const;
};

class rename_log_record : public dirop_log_record
{
private:
  bigint _from_dir;
  str _from_name;
  bool _target_exists;
  bigint _target;

public:
  rename_log_record (const rename_log_record &r) { *this = r; }
  rename_log_record (bigint object, cfsrw_ftype object_type,
                     bigint from_dir, str from_name,
                     bigint dir, str name, bigint target);
  rename_log_record (bigint object, cfsrw_ftype object_type,
                     bigint from_dir, str from_name,
                     bigint dir, str name);
  rename_log_record (xin *);
  str serialize () const;
  
  bigint from_dir ()  const { return _from_dir; }
  str from_name ()     const { return _from_name; }
  bool target_exists ()    const { return _target_exists; }
  bigint target () const { return _target; }
};

inline dirop_log_record *
dirop_record (ptr<log_record> r)
{
  if (r->record_type() == cfsrw_log_record::T_LINK)
    return r->template record<link_log_record> ();
  else if (r->record_type() == cfsrw_log_record::T_UNLINK)
    return r->template record<unlink_log_record> ();
  else if (r->record_type() == cfsrw_log_record::T_RENAME)
    return r->template record<rename_log_record> ();
  return 0;
}

#endif

