/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <sys/time.h>
#include "definc.h"
#include "parseopt.h"
#include "ns.h"
#include "log.h"
#include "cfsrw_rec.h"
#include "cfsrw_prot.h"

void
get_head_record (bigint id, callback<void, ptr<log_record> >::ref cb)
{
  ns::retrieve_loghead (id, cb);
}

void
get_record (bigint id, callback<void, ptr<log_record> >::ref cb)
{
  ns::retrieve_record (id, cb);
}

bigint
make_block_id (const void *data, int size) 
{
  long *rd = (long *)data;
  char id[sha1::hashsize];
  sha1_hash (id, rd, size);
  
  bigint ID;
  mpz_set_rawmag_be (&ID, id, sizeof (id));
  return ID;
}

void
store_block_cb (callback<void, bigint, int>::ref cb,
                ns_stat status, bigint key)
{
  if (status != NS_OK) {
    warn << "store_block_cb failed\n";
    exit(0);
  }
  cb (key, 1);
}

void
main_cb (bool ishead, int *n, bigint id, int ok)
{
  if (ishead) {
    char raw[20];
    mpz_get_rawmag_be (raw, sizeof (raw), &id);
    str name = armor64A (raw, sizeof (raw));
    strbuf out;
    out << name << "\n";
    out.tosuio ()->output (1);
  }

  *n -= 1;
  if (*n == 0)
    exit (0);
}

bigint
store1 (const str &data, callback<void, bigint, int>::ref cb)
{
  bigint id = make_block_id(data.cstr(), data.len());
  ns::insert_block (data.cstr(), data.len(), wrap(store_block_cb, cb));
  return id;
}

void
store_head (const str &data, callback<void, bigint, int>::ref cb,
            ptr <sfspriv> sk)
{
  assert (sk);
  sfs_sig2 s;
  sk->sign (&s, data);
  sfs_pubkey2 k;
  sk->export_pubkey (&k);
  ns::insert_loghead
    (k, s, data.cstr (), data.len (), wrap (store_block_cb, cb));
}

void
make_new (ptr <sfspriv> sk)
{
  static int nb = 0;
  log_record *e;
  bigint lid;
  uint seq = 0;
 
  strbuf b;
  sk->export_pubkey (b, false);
  str pk_raw = b;
  bigint head = compute_sha1_hash (pk_raw.cstr (), pk_raw.len ());

  nb++;
  vervec v;
  e = New end_log_record ();
  e->set_version (v);
  e->set_head (head);
  e->set_seqnum (seq);
  lid = store1 (e->serialize (), wrap (main_cb, false, &nb));
  seq++;

  struct timeval t;
  gettimeofday (&t, 0);

  cfsrw_sattr attr;
  attr.valid_mode = 1;
  attr.mode = 0755;
  attr.valid_size = 1;
  attr.size = 0;
  attr.valid_mtime = 1;
  attr.mtime = t.tv_sec;
  attr.valid_uid = attr.valid_gid = 1;
  attr.uid = attr.gid = 0;

  nb++;
  e = New inode_log_record (CFSRW_DIR, attr);
  v.clear ();
  v.add (head,lid,seq-1);
  e->set_version (v);
  e->set_prev (lid);
  e->set_head (head);
  e->set_seqnum (seq);
  lid = store1 (e->serialize (), wrap (main_cb, false, &nb));
  bigint root = ((inode_log_record*)e)->object ();
  seq++;

  nb++;
  e = New head_log_record (root);
  v.clear ();
  v.add (head,lid,seq-1);
  e->set_version (v);
  e->set_prev (lid);
  e->set_head (head);
  e->set_seqnum (seq);
  store_head (e->serialize (), wrap (main_cb, true, &nb), sk);

  // warn << "root " << root << "\n";
}

void
new_attach_cb_1 (bigint old, ptr<sfspriv> sk, ptr<log_record> old_e)
{
  strbuf b;
  sk->export_pubkey (b, false);
  str pk_raw = b;
  bigint head = compute_sha1_hash (pk_raw.cstr (), pk_raw.len ());

  if (!old_e)
    return;
  
  static int nb = 0;
  log_record *e;
  bigint lid;
  
  head_log_record *re = old_e->template record<head_log_record> ();
  bigint root = re->root ();

  nb++;
  e = New end_log_record ();
  vervec v;
  e->set_version (v);
  e->set_head (head);
  e->set_seqnum (0);
  lid = store1 (e->serialize (), wrap (main_cb, false, &nb));
  
  nb++;
  e = New head_log_record (root);
  v.clear ();
  v.add (head,lid,0);
  e->set_version (v);
  e->set_prev (lid);
  e->set_head (head);
  e->set_seqnum (1);
  store_head (e->serialize (), wrap (main_cb, true, &nb), sk);
  
  // warn << "root " << root << "\n";
}

void
new_attach (bigint old, ptr <sfspriv> sk)
{
  get_head_record (old, wrap (new_attach_cb_1, old, sk));
}

static void
usage ()
{
  warnx << "usage: " << progname << " [-s <keyfile>] [id]\n";
  warnx << "  " << progname
        << " creates a new head of log. The head of log block is\n"
        << "  stored under the public key specified by the MKLOG_KEYFILE\n"
        << "  environment variable or the -s argument. If a chord ID is\n"
        << "  provided, it is interpreted as the hash of another user's\n"
        << "  public key, and the new log will be part of the same fs.\n"
        << "  Otherwise, a new fs is created\n";
  exit (1);
}

int
main (int argc, char **argv)
{
  char *sk_file = getenv ("MKLOG_KEYFILE");

  setprogname (argv[0]);

  int ch;
  while ((ch = getopt (argc, argv, "s:")) != -1)
    switch (ch) {
    case 's':
      sk_file = optarg;
      break;
    case '?':
    default:
      usage ();
    }
  argc -= optind;
  argv += optind;

  if ( (argc > 2) || !sk_file )
    usage ();

  ptr <sfspriv> sk;
  str key = file2wstr (sk_file);
  if (!key) {
    warn << sk_file << ": " << strerror (errno) << "\n";
    fatal ("errors!\n");
  }
  else if (!(sk = sfscrypt.alloc_priv (key, SFS_SIGN))) {
    warn << "could not decode " << sk_file << "\n";
    warn << key << "\n";
    fatal ("errors!\n");
  }

  ns::init ();
  random_init ();

  if (argc == 1) {
    str dearm = dearmor64A (argv[0]);
    bigint id;
    mpz_set_rawmag_be (&id, dearm.cstr (), dearm.len ());
    new_attach (id, sk);
  }
  else
    make_new (sk);
  
  amain();
}

