/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <sys/time.h>
#include "definc.h"
#include "parseopt.h"
#include "rxx.h"

#include "ns.h"
#include "vervec.h"
#include "logiter.h"
#include "cfsrw_rec.h"
#include "cfsrw_prot.h"
#include "view.h"

ptr<view> _view;

static vec<bigint> _heads;
static vec<bigint> _latest;
static ptr<logiter> _iter;
static bigint _root;
static bool verbose = false;
static int _work = 0;

void
store_cb (bool print, ns_stat status, bigint id)
{
  _work--;
  if (status != NS_OK)
    warn << "failed to store view\n";
  else if (print) {
    char raw[20];
    mpz_get_rawmag_be (raw, sizeof (raw), &id);
    str xx = armor64A (raw, sizeof (raw));
    strbuf out;
    out << xx << "\n";
    out.tosuio ()->output (1);
  }
  if (_work == 0)
    exit (0);
}

void
print_vervec (unsigned n, vervec &v, strbuf &out)
{
  for (unsigned i=0; i<v.size (); i++) {
    for (unsigned j=0; j<n; j++) out << " ";
    out << " verh " << v.head (i)   << "\n";
    for (unsigned j=0; j<n; j++) out << " ";
    out << " verl " << v.latest (i) << "\n";
    for (unsigned j=0; j<n; j++) out << " ";
    out << " vers " << v.seqnum (i) << "\n";
  }
}

void
dump_cb (iterinfo *i)
{
  if (!i->finished) {
    recordp e = i->record;
    assert (e);
    bigint hid = e->head ();
    bigint lid = e->hash ();

    strbuf out;
    char ts[32];
    unsigned t = e->timestamp ();
    ctime_r (&((time_t)t), ts);
    ts[strlen (ts)-1] = '\0';

    out << ts << "   seq " << e->seqnum () << "\n                        "
              << "   log " << e->hash ()   << "\n                        "
	      << "  head " << e->head ()   << "\n";
    vervec v = e->version ();
    print_vervec (25, v, out);
    for (unsigned j=0; j<25; j++) out << " ";

    switch (e->record_type ()) {
      case cfsrw_log_record::T_INODE:
	{
          ptr<inode_log_record> l =
	    *reinterpret_cast<ptr<inode_log_record>*> (&e);
	  switch (l->object_type ()) {
	    case CFSRW_REG:
	      out << "  reg ";
	      break;
	    case CFSRW_DIR:
	      out << "  dir ";
	      break;
	    case CFSRW_LNK:
	      out << "  lnk ";
	      break;
	    default:
	      break;
	  }
	  out << l->object () << "\n";
	}
	break;
      case cfsrw_log_record::T_INTEND:
	{
	  out << "itend ";
          intend_log_record *ie = e->template record<intend_log_record> ();
	  out << ie->name ()  << "\n                            in "
	      << ie->dir ()   << "\n                          verf "
	      << ie->verf ()  << "\n                        cancel "
	      << ie->canceled () << "\n";
	}
	break;
      case cfsrw_log_record::T_RENAME:
	{
	  out << "rname ";
          ptr<rename_log_record> l =
	    *reinterpret_cast<ptr<rename_log_record>*> (&e);
	  out << l->object () << "\n                            in ";
	  out << l->dir ()    << "\n                            as "
	      << l->name ()   << "\n";
	  out << "                           was " << l->from_name () << "\n";
	}
	break;
      case cfsrw_log_record::T_LINK:
	{
	  out << " link ";
          ptr<link_log_record> l =
	    *reinterpret_cast<ptr<link_log_record>*> (&e);
	  out << l->object () << "\n                            in ";
	  out << l->dir ()    << "\n                            as "
	      << l->name ()   << "\n";
	}
	break;
      case cfsrw_log_record::T_UNLINK:
	{
	  out << "ulink ";
          ptr<unlink_log_record> l =
	    *reinterpret_cast<ptr<unlink_log_record>*> (&e);
	  out << l->object () << "\n                            in ";
	  out << l->dir ()    << "\n                            as "
	      << l->name ()   << "\n";
	}
	break;
      case cfsrw_log_record::T_WRITE:
	{
          out << "write ";
          ptr<write_log_record> l =
	    *reinterpret_cast<ptr<write_log_record>*> (&e);
	  out << l->file   () << "\n                            at "
	      << l->offset () << ", " << l->count () << " bytes\n";
	}
	break;
      case cfsrw_log_record::T_ATTR:
	{
	  out << " attr ";
          ptr<attr_log_record> l =
	    *reinterpret_cast<ptr<attr_log_record>*> (&e);
	  out << l->object () << "\n                          size ";
	  if (l->attr ()->valid_size)
	    out << l->attr ()->size << " bytes\n";
	  else
	    out << " not set\n";
	}
	break;
      case log_record::T_END:
      case log_record::T_HEAD:
      default:
	break;
    }
    if (e->record_type () != log_record::T_END &&
	e->record_type () != log_record::T_HEAD)
      out.tosuio ()->output (1);
  }
  else if (i->finished) {
    _work--;
    if (_work == 0)
      exit (0);
  }
}

void
new_view (bigint key, recordp l)
{
  if (l && l->record_type () == log_record::T_HEAD) {
    ref<view_constructor> v = New refcounted<view_constructor> ();
    head_log_record *h = l->template record<head_log_record>();
    v->root (h->root());
    v->add_log (key, key);
    _work++;
    v->store (wrap(store_cb, true));
  }
  else {
    exit(1);
  }
}

void
modify_view (bigint key, bigint latest, log_status status)
{
  if (status) {
    warn << "cannot load view\n";
    exit (1);
  }
  ref<view_constructor> vc = New refcounted<view_constructor> (*_view);
  vc->add_log (key,latest);
  _work++;
  vc->store (wrap (store_cb, true));
}

void
got_view (bigint vid, bool printkey, log_status status)
{
  if (status) {
    warn << "cannot load view\n";
    exit (1);
  }

  strbuf out;
  
  if (!printkey)
    out << "root inum: " << _view->root () << "\n";

  for (unsigned i=0; i<_view->nlogs (); i++) {
    if (!printkey) {
      out << "     head: " << _view->head (i) << "\n";
      out << "   latest: " << _view->latest (i) << "\n";
    }
    else {
      bigint n = _view->head (i);
      char raw[20];
      mpz_get_rawmag_be (raw, sizeof (raw), &n);
      str s = armor64A (raw, sizeof (raw));
      out << s << "\n";
      break;
    }
  }
  out.tosuio ()->output (1);

  if (verbose) {
    _work++;
    New refcounted<logiter> (wrap (&dump_cb), vid);
  }
  else
    exit (0);
}

void
usage ()
{
  warnx << "usage: \n";
  warnx << "  " << progname << " [-v] viewid\n"
        << "    prints out information on the given view. if -v is specified,\n"
        << "    iterates and prints all the log entries as well\n";
  warnx << "  " << progname << " -c key\n"
        << "    creates a new view with the log stored at the given key\n";
  warnx << "  " << progname << " -a key[,latest] viewid\n"
        << "    adds the log stored at the given key to an existing view\n";
  warnx << "  " << progname << " -p viewid\n"
        << "    prints out a public key hash in the view\n";
  warnx << "viewid can be the hash of a view's chord ID, or the mount point\n";
  exit (1);
}

int
main (int argc, char **argv)
{
  bool printkey = false;
  char *newkey = 0;
  char *addkey = 0;
  setprogname (argv[0]);
  int opt = 0;

  int ch;
  while ((ch = getopt (argc, argv, "a:c:vp")) != -1)
    switch (ch) {
    case 'p':
      opt++;
      printkey = true;
      break;
    case 'c':
      opt++;
      newkey = optarg;
      break;
    case 'a':
      opt++;
      addkey = optarg;
      break;
    case 'v':
      verbose = true;
      break;
    case '?':
    default:
      usage ();
    }
  argc -= optind;
  argv += optind;

  if ((newkey && argc > 0) || (!newkey && argc != 1) || opt > 1)
    usage();

  ns::init();

  bigint vid;
  if (argc == 1) {
    char sym[1024];
    str dearm;
    rxx path_regexp ("chord:([0-9a-zA-Z+-]*)[/]*$");
    int r = readlink(argv[0], sym, 1024);
    if (r > 0) sym[r] = '\0';
    if (path_regexp.search (argv[0]) || (r > 0 && path_regexp.search(sym)))
      dearm = dearmor64A(path_regexp[1]);
    else
      dearm = dearmor64A(argv[0]);
    if (dearm) 
      mpz_set_rawmag_be(&vid, dearm.cstr(), dearm.len());
    else {
      warnx << "viewctl: key hash invalid\n";
      exit(1);
    }
  }

  if (newkey) {
    str dearm = dearmor64A(newkey);
    if (dearm) {
      bigint key;
      mpz_set_rawmag_be(&key, dearm.cstr(), dearm.len());
      ns::retrieve_loghead (key, wrap (new_view, key));
    }
    else {
      warnx << "viewctl: key hash invalid\n";
      exit(1);
    }
  }
  else if (addkey) {
    bigint key;
    bigint latest;
    str keystr;
    str lateststr;

    rxx wb_regexp ("([0-9a-zA-Z+-]*),([0-9a-zA-Z+-]*)$");
    if (wb_regexp.search (addkey)) {
      keystr = wb_regexp[1];
      lateststr = wb_regexp[2];
    }
    else {
      keystr = addkey;
      lateststr = addkey;
    }
    str dearm = dearmor64A(keystr);
    if (dearm) {
      mpz_set_rawmag_be(&key, dearm.cstr(), dearm.len());
      dearm = dearmor64A(lateststr);
      mpz_set_rawmag_be(&latest, dearm.cstr(), dearm.len());
      _view = New refcounted<view> (wrap(&modify_view, key, latest), vid);
    }
    else {
      warnx << "viewctl: key hash invalid\n";
      exit(1);
    }
  }
  else
    _view = New refcounted<view> (wrap(&got_view, vid, printkey), vid);
  amain();
}

