/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

/*
 *  Copyright (C) 1998  David Mazieres (dm@uun.org)
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 *  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *  DEALINGS IN THE SOFTWARE.
 */

#include "nfstrans.h"
#include "arpc.h"
#include "qhash.h"
#include "itree.h"
#include "list.h"
#include "attrcache.h"

enum { max_attr_dat = 1024 };

inline bool
operator== (const nfs_fh3 &a, const nfs_fh3 &b)
{
  return a.data.size () == b.data.size ()
    && !memcmp (a.data.base (), b.data.base (), b.data.size ());
}

inline bool
operator!= (const nfs_fh3 &a, const nfs_fh3 &b)
{ 
  return !(a == b);
}

static tailq<cfsrw_attr_cache::attr_dat,
             &cfsrw_attr_cache::attr_dat::lrulink> lrulist;
static u_int num_attr_dat;

cfsrw_attr_cache::attr_dat::attr_dat (cfsrw_attr_cache *c, const nfs_fh3 &f,
                                      const fattr3 *a)
  : cache (c), fh (f)
{
  attr = *a;
  lrulist.insert_tail (this);
  cache->attrs.insert (this);
  num_attr_dat++;
  while (num_attr_dat > implicit_cast<u_int> (max_attr_dat))
    delete lrulist.first;
}

cfsrw_attr_cache::attr_dat::~attr_dat ()
{
  lrulist.remove (this);
  cache->attrs.remove (this);
  num_attr_dat--;
}

void
cfsrw_attr_cache::attr_dat::touch ()
{
  lrulist.remove (this);
  lrulist.insert_tail (this);
}

void
cfsrw_attr_cache::attr_dat::set (const fattr3 *a, const wcc_attr *w)
{
  if (a->mode != attr.mode || a->uid != attr.uid || a->gid != attr.gid)
    access.clear ();

#if 0
  /* Maybe need something like this for non-unix?  We would need to
   * know also if the operation was a SETATTR. */
  else if (a->ctime != attr.ctime && (!w || w->ctime != attr.ctime))
    access.clear ();
#endif

  attr = *a;
}

void
cfsrw_attr_cache::flush_access (const nfs_fh3 &fh, sfs_aid aid)
{
  if (attr_dat *ad = attrs[fh])
    ad->access.remove (aid);
}

void
cfsrw_attr_cache::attr_enter (const nfs_fh3 &fh, const fattr3 *a,
                              const wcc_attr *w)
{
  attr_dat *ad = attrs[fh];
  if (!a) {
    if (ad)
      delete ad;
  }
  else {
    if (!ad)
      vNew attr_dat (this, fh, a);
    else {
      ad->set (a, w);
      ad->touch ();
    }
  }
}

const fattr3 *
cfsrw_attr_cache::attr_lookup (const vervec &v, const nfs_fh3 &fh)
{
  if (version.equals (v)) {
    attr_dat *ad = attrs[fh];
    if (ad) {
      ad->touch ();
      return &ad->attr;
    }
  }
  return NULL;
}

const fattr3 *
cfsrw_attr_cache::last_known_attr_lookup (const nfs_fh3 &fh)
{
  attr_dat *ad = attrs[fh];
  if (ad) {
    ad->touch ();
    return &ad->attr;
  }
  return NULL;
}

void
cfsrw_attr_cache::access_enter (const nfs_fh3 &fh, sfs_aid aid,
                                u_int32_t mask, u_int32_t perm)
{
  if (attr_dat *ad = attrs[fh]) {
    ad->touch ();
    access_dat *ac = ad->access[aid];
    if (!ac)
      ad->access.insert (aid, access_dat (mask, perm));
    else {
      ac->mask |= mask;
      ac->perm = (ac->perm & ~perm) | (mask & perm);
    }
  }
}

int32_t
cfsrw_attr_cache::access_lookup (const vervec& v, const nfs_fh3 &fh,
                                 sfs_aid aid, u_int32_t mask)
{
  if (version.equals (v)) {
    if (attr_dat *ad = attrs[fh]) {
      if (access_dat *ac = ad->access[aid]) {
	if ((mask & ac->mask) == mask) {
	  ad->touch ();
	  return ac->perm & mask;
	}
      }
    }
  }
  return -1;
}

