/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "nfstrans.h"
#include "nfs3_nonnul.h"
#include "cfsrwcd.h"

unsigned cfsrwcd::blocksize = 4096;

#define mkcfscall \
  cc->l = c->l; \
  cc->s = c->s; \
  cc->start = c->start; \
  cc->stop = c->stop; \
  _cfsrw.dispatch (cc);

void
cfsrwcd::set_cfsrw_sattr (cfsrw_sattr &ca, sattr3 &sa)
{
  if (sa.mode.set) {
    ca.valid_mode = 1;
    ca.mode = *sa.mode.val;
  }
  if (sa.uid.set) {
    ca.valid_uid = 1;
    ca.uid = *sa.uid.val;
  }
  if (sa.gid.set) {
    ca.valid_gid = 1;
    ca.gid = *sa.gid.val;
  }
  if (sa.size.set) {
    ca.valid_size = 1;
    ca.size = *sa.size.val;
  }
#if 0
  if (sa.mtime.set) {
    ca.valid_mtime = 1;
    ca.mtime = (sa.mtime).seconds;
  }
#endif
}

uint32
cfsrwcd::access_check (cfsrw_ftype t, uint32 access_req)
{
  uint32 r = 0;
  switch (t) {
  case CFSRW_DIR:
    r = ACCESS3_READ | ACCESS3_LOOKUP | ACCESS3_MODIFY | 
        ACCESS3_EXTEND | ACCESS3_DELETE;
    break;
  case CFSRW_LNK:
    r = ACCESS3_READ;
    break;
  case CFSRW_REG:
    r = ACCESS3_READ | ACCESS3_MODIFY | ACCESS3_EXTEND | ACCESS3_EXECUTE;
    break;
  }
  return (access_req & r);
}


///////////////
//  GETATTR  //
///////////////

void
cfsrwcd::getattr_reply (nfscall *sbp, cfscallp c, cfsrw_status status)
{
  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }
  bigint fh =  nfs2cfsfh (sbp->template getarg<nfs_fh3> ());
  cfsrw_getattr_res *res = static_cast<cfsrw_getattr_res *> (c->res ());
  ptr<fattr3> fa = get_fattr3 (fh, &res->attributes);

  getattr3res nfsres (NFS3_OK);
  (*nfsres.attributes) = *fa;
  reply (sbp, &nfsres);
}


///////////////
//  ACCESS   //
///////////////

void
cfsrwcd::access_reply (nfscall *sbp, cfscallp c, cfsrw_status status)
{
  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }
  access3args *a = sbp->template getarg<access3args> ();
  bigint fh =  nfs2cfsfh (&a->object);
  cfsrw_getattr_res *res = static_cast<cfsrw_getattr_res *> (c->res ());
  ptr<fattr3> fa = get_fattr3 (fh, &res->attributes);

  access3res nfsres (NFS3_OK);
  nfsres.resok->access = access_check (res->attributes.type, a->access);
  nfsres.resok->obj_attributes.set_present(true);
  *(nfsres.resok->obj_attributes.attributes) = *fa;
  reply (sbp, &nfsres);
}


///////////////
//  LOOKUP   //
///////////////

void
cfsrwcd::lookup_dotdot_reply (nfscall *sbp, cfscallp c, cfsrw_status status)
{
  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }
  
  cfsrw_getparent_res *res = static_cast<cfsrw_getparent_res *> (c->res ());
  lookup3res nfsres(NFS3_OK);
  nfs_fh3 fh;
  cfsfh2nfs (res->dir, &fh);
  nfsres.resok->object = fh;
  nfsres.resok->obj_attributes.set_present(false);
  reply (sbp, &nfsres);
}

void
cfsrwcd::lookup_reply (nfscall *sbp, cfscallp c, cfsrw_status status)
{
  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }
  
  cfsrw_lookup_res *res = static_cast<cfsrw_lookup_res *> (c->res ());
  lookup3res nfsres(NFS3_OK);
  nfs_fh3 fh;
  cfsfh2nfs (res->object, &fh);
  nfsres.resok->object = fh;
  nfsres.resok->obj_attributes.set_present(false);
  reply (sbp, &nfsres);
}


///////////////
//  WRITE    //
///////////////

void
cfsrwcd::write_reply (nfscall *sbp, cfscallp c, cfsrw_status status)
{
  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }
  
  write3args *a = sbp->template getarg<write3args> ();
  write3res nfsres(NFS3_OK);
  nfsres.resok->count = a->count;
  nfsres.resok->committed = FILE_SYNC;
  writeverf3 verf3;
  bigint verf = _cfsrw.verf ();
  mpz_get_rawmag_be (verf3.base(), NFS3_WRITEVERFSIZE, &verf);
  nfsres.resok->verf = verf3;
  // because of things like brief periods of inconsistency, we cannot
  // return wcc reliably.
  reply (sbp, &nfsres);
}

void
cfsrwcd::write_found_file (nfscall *sbp, cfscallp c, cfsrw_status status)
{
  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }
  
  write3args *a = sbp->template getarg<write3args> ();
  bigint file = nfs2cfsfh (&(a->file));
    
  cfscallp cc = New refcounted<cfscall>
    (CFSRWPROC_WRITE, wrap (this, &cfsrwcd::write_reply, sbp));
  cfsrw_write_arg *arg = static_cast<cfsrw_write_arg *> (cc->arg ());
  arg->file = file;
  arg->offset = a->offset;
  arg->count = a->count;
  arg->data = a->data;
  mkcfscall;
}


///////////////
//  READ     //
///////////////

void
cfsrwcd::read_reply (nfscall *sbp, cfscallp getattr_c,
                     cfscallp c, cfsrw_status status)
{
  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }
  
  read3args *a = sbp->template getarg<read3args> ();
  cfsrw_read_res *read_res = static_cast<cfsrw_read_res *> (c->res ());
  cfsrw_getattr_res *getattr_res =
    static_cast<cfsrw_getattr_res *> (getattr_c->res ());

  read3res nfsres(NFS3_OK);
  nfsres.resok->count = read_res->count;
  nfsres.resok->eof = read_res->eof;
  nfsres.resok->data = read_res->data;

  bigint fh = nfs2cfsfh (&a->file);
  ptr<fattr3> fa = get_fattr3 (fh, &getattr_res->attributes);
  nfsres.resok->file_attributes.set_present (true);
  *(nfsres.resok->file_attributes.attributes) = *fa;
  
  reply (sbp, &nfsres);
}

void
cfsrwcd::read_found_file (nfscall *sbp, cfscallp c, cfsrw_status status)
{
  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }
      
  read3args *a = sbp->template getarg<read3args> ();
  bigint file = nfs2cfsfh (&a->file);
  uint32 count = a->count;
  if (count > 8192)
    count = 8192;
  
  cfscallp cc = New refcounted<cfscall>
    (CFSRWPROC_READ, wrap (this, &cfsrwcd::read_reply, sbp, c));
  cfsrw_read_arg *arg = static_cast<cfsrw_read_arg *> (cc->arg ());
  arg->file = file;
  arg->offset = a->offset;
  arg->count = a->count;
  mkcfscall;
}


///////////////
//  SETATTR  //
///////////////

void
cfsrwcd::setattr_reply (nfscall *sbp, cfscallp getattr_c,
                        cfscallp c, cfsrw_status status)
{
  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }
  
  setattr3args *a = sbp->template getarg<setattr3args> ();
  wccstat3 nfsres (NFS3_OK);

  bigint fh = nfs2cfsfh (&a->object);
  cfsrw_getattr_res *getattr_res =
    static_cast<cfsrw_getattr_res *> (getattr_c->res ());
  ptr<fattr3> fa = get_fattr3 (fh, &getattr_res->attributes);

  // this is not correct: we can't return wcc information reliably
  // because of things like brief periods of inconsistency. however,
  // it seems that nfs client does not work correctly when setattr
  // does not return wcc information.
  nfsres.wcc->before.set_present (true);
  nfsres.wcc->before.attributes->size = fa->size;
  nfsres.wcc->before.attributes->mtime = fa->mtime;
  nfsres.wcc->before.attributes->ctime = fa->ctime;

  apply_sattr (*fa, a->new_attributes);
  nfsres.wcc->after.set_present (true);
  *(nfsres.wcc->after.attributes) = *fa;

  reply (sbp, &nfsres);
}

void
cfsrwcd::setattr_found_obj (nfscall *sbp, cfscallp c, cfsrw_status status)
{
  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }
  
  setattr3args *a = sbp->template getarg<setattr3args> ();
  bigint obj = nfs2cfsfh (&a->object);
    
  cfscallp cc = New refcounted<cfscall>
    (CFSRWPROC_SETATTR, wrap (this, &cfsrwcd::setattr_reply, sbp, c));
  cfsrw_setattr_arg *arg = static_cast<cfsrw_setattr_arg *> (cc->arg ());
  arg->object = obj;
  set_cfsrw_sattr (arg->new_attributes, a->new_attributes);
  mkcfscall;
}


////////////////
//  READLINK  //
////////////////

void
cfsrwcd::readlink_reply (nfscall *sbp, cfscallp getattr_c,
                         cfscallp c, cfsrw_status status)
{
  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }

  cfsrw_readlink_res *res = static_cast<cfsrw_readlink_res *> (c->res ());
  cfsrw_getattr_res *getattr_res =
    static_cast<cfsrw_getattr_res *> (getattr_c->res ());

  readlink3res nfsres(NFS3_OK);
  nfsres.resok->data = res->path;

  bigint fh = nfs2cfsfh (sbp->template getarg<nfs_fh3> ());
  ptr<fattr3> fa = get_fattr3 (fh, &getattr_res->attributes);
  nfsres.resok->symlink_attributes.set_present (true);
  *(nfsres.resok->symlink_attributes.attributes) = *fa;
  
  reply (sbp, &nfsres);
}

void
cfsrwcd::readlink_found_file (nfscall *sbp, cfscallp c, cfsrw_status status)
{
  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }

  cfsrw_getattr_res *res = static_cast<cfsrw_getattr_res *> (c->res ());
  if (res->attributes.type != CFSRW_LNK) {
    error (sbp, NFS3ERR_INVAL);
    return;
  }

  cfscallp cc = New refcounted<cfscall>
    (CFSRWPROC_READLINK, wrap (this, &cfsrwcd::readlink_reply, sbp, c));
  cfsrw_readlink_arg *arg = static_cast<cfsrw_readlink_arg *> (cc->arg ());
  arg->file = nfs2cfsfh (sbp->template getarg<nfs_fh3> ());
  mkcfscall;
}


/////////////////
//  READDIR    //
/////////////////

void
cfsrwcd::readdir_reply (nfscall *sbp, cfscallp getattr_c,
                        cfscallp c, cfsrw_status status)
{
  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }
 
  readdir3args *a = sbp->template getarg<readdir3args> ();
  cfsrw_getattr_res *getattr_res =
    static_cast<cfsrw_getattr_res *> (getattr_c->res ());
  cfsrw_readdir_res *res = static_cast<cfsrw_readdir_res *> (c->res ());
 
  readdir3res nfsres (NFS3_OK);
  
  bigint fh = nfs2cfsfh (&a->dir);
  ptr<fattr3> fa = get_fattr3 (fh, &getattr_res->attributes);
  nfsres.resok->dir_attributes.set_present (true);
  *(nfsres.resok->dir_attributes.attributes) = *fa;

  // nfsres.resok->cookieverf = res->cookieverf;
  
  nfsres.resok->reply.eof = res->reply.eof;
  rpc_ptr <cfs_rpc_dirent> de = res->reply.entries;
  rpc_ptr <entry3> *direntp = &nfsres.resok->reply.entries;
  while (de) {
    (*direntp).alloc ();
    (*direntp)->name = de->name;
    (*direntp)->fileid = cfsfh2fid (de->fh);
    (*direntp)->cookie = de->cookie;
    direntp = &(*direntp)->nextentry;
    de = de->nextentry;
  }
  reply (sbp, &nfsres);
}

void
cfsrwcd::readdir_found_dir (nfscall *sbp, cfscallp c, cfsrw_status status)
{
  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }

  readdir3args *a = sbp->template getarg<readdir3args> ();
  cfscallp cc = New refcounted<cfscall>
    (CFSRWPROC_READDIR, wrap (this, &cfsrwcd::readdir_reply, sbp, c));
  cfsrw_readdir_arg *arg = static_cast<cfsrw_readdir_arg *> (cc->arg ());
  arg->dir = nfs2cfsfh (&a->dir);
  arg->cookie = a->cookie;
  // arg->cookieverf = cookieverf;
  arg->count = a->count;
  mkcfscall;
}


////////////////////
//  REMOVE/RMDIR  //
////////////////////

void
cfsrwcd::remove_reply (nfscall *sbp, cfscallp c, cfsrw_status status)
{
  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }

  cfsrw_unlink_arg *arg = static_cast<cfsrw_unlink_arg *> (c->arg ());
  wccstat3 nfsres(NFS3_OK);
  nfs_fh3 fh;
  cfsfh2nfs (arg->object, &fh);
  reply (sbp, &nfsres, fh);
}

void
cfsrwcd::rmdir_readdir (nfscall *sbp, bigint object,
                        cfscallp c, cfsrw_status status)
{
  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }

  cfsrw_readdir_res *res = static_cast<cfsrw_readdir_res *> (c->res ());
  if (res->entries > 0) {
    error (sbp, NFS3ERR_NOTEMPTY);
    return;
  }

  diropargs3 *a = sbp->template getarg<diropargs3> ();
  cfscallp cc = New refcounted<cfscall>
    (CFSRWPROC_UNLINK, wrap (this, &cfsrwcd::remove_reply, sbp));
  cfsrw_unlink_arg *arg = static_cast<cfsrw_unlink_arg *> (cc->arg ());

  arg->object = object;
  arg->type = CFSRW_DIR;
  arg->dir = nfs2cfsfh (&a->dir);
  arg->name = a->name;
  mkcfscall;
}

void
cfsrwcd::remove_found_attr (nfscall *sbp, bigint object,
                            cfscallp c, cfsrw_status status)
{
  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }

  cfsrw_getattr_res *res = static_cast<cfsrw_getattr_res *> (c->res ());
  
  if (sbp->proc () == NFSPROC3_RMDIR) {
    if (res->attributes.type != CFSRW_DIR) {
      error (sbp, NFS3ERR_NOTDIR);
      return;
    }

    cfscallp cc = New refcounted<cfscall>
      (CFSRWPROC_READDIR, wrap (this, &cfsrwcd::rmdir_readdir, sbp, object));
    cfsrw_readdir_arg *arg = static_cast<cfsrw_readdir_arg *> (cc->arg ());
    arg->dir = object;
    arg->cookie = 0;
    // arg->cookieverf = cookieverf;
    arg->count = 2048;
    mkcfscall;
  }

  else if (sbp->proc () == NFSPROC3_REMOVE) {
    if (res->attributes.type == CFSRW_DIR) {
      error (sbp, NFS3ERR_ISDIR);
      return;
    }

    diropargs3 *a = sbp->template getarg<diropargs3> ();
    cfscallp cc = New refcounted<cfscall>
      (CFSRWPROC_UNLINK, wrap (this, &cfsrwcd::remove_reply, sbp));
    cfsrw_unlink_arg *arg = static_cast<cfsrw_unlink_arg *> (cc->arg ());

    arg->object = object;
    arg->type = res->attributes.type;
    arg->dir = nfs2cfsfh (&a->dir);
    arg->name = a->name;
    mkcfscall;
  }
}

void
cfsrwcd::remove_found_file (nfscall *sbp, cfscallp c, cfsrw_status status)
{
  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }
  
  cfsrw_lookup_res *res = static_cast<cfsrw_lookup_res *> (c->res ());
  cfscallp cc = New refcounted<cfscall>
    (CFSRWPROC_GETATTR, wrap (this, &cfsrwcd::remove_found_attr,
			      sbp, res->object));
  cfsrw_getattr_arg *arg = static_cast<cfsrw_getattr_arg *>(cc->arg ());
  arg->object = res->object;
  mkcfscall;
}


///////////////
//  RENAME   //
///////////////

void
cfsrwcd::rename_reply (nfscall *sbp, cfscallp c, cfsrw_status status)
{
  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }
  
  rename3res nfsres (NFS3_OK);
  cfsrw_rename_arg *rename_arg = static_cast<cfsrw_rename_arg *> (c->arg ());
  if (rename_arg->target_exists) {
    nfs_fh3 fh;
    cfsfh2nfs(rename_arg->target, &fh);
    reply (sbp, &nfsres, fh);
  }
  else
    reply (sbp, &nfsres);
}

void
cfsrwcd::rename_readdir (nfscall *sbp, cfscallp rename_c,
                         cfscallp c, cfsrw_status status)
{
  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }

  cfsrw_readdir_res *res = static_cast<cfsrw_readdir_res *> (c->res ());
  if (res->entries > 0) {
    error (sbp, NFS3ERR_NOTEMPTY);
    return;
  }

  cfscallp cc = rename_c;
  mkcfscall;
}

void
cfsrwcd::rename_to_attr (nfscall *sbp, cfscallp rename_c,
                         cfscallp c, cfsrw_status status)
{
  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }

  cfsrw_getattr_res *res = static_cast<cfsrw_getattr_res *> (c->res ());
  cfsrw_rename_arg *rename_arg =
    static_cast<cfsrw_rename_arg *> (rename_c->arg ());
  // target and source type not compatible
  if (res->attributes.type != rename_arg->type) {
    error (sbp, NFS3ERR_INVAL);
    return;
  }

  if (res->attributes.type == CFSRW_DIR) {
    // must check if dir is empty or not
    cfscallp cc = New refcounted<cfscall>
      (CFSRWPROC_READDIR, wrap (this, &cfsrwcd::rename_readdir,
				sbp, rename_c));
    cfsrw_readdir_arg *arg = static_cast<cfsrw_readdir_arg *> (cc->arg ());
    arg->dir = rename_arg->target;
    arg->cookie = 0;
    // arg->cookieverf = cookieverf;
    arg->count = 2048;
    mkcfscall;
  }
  else {
    cfscallp cc = rename_c;
    mkcfscall;
  }
}

void
cfsrwcd::rename_to (nfscall *sbp, bigint from, cfsrw_ftype from_type,
                    cfscallp c, cfsrw_status status)
{
  if (status && status != CFSRWERR_NOENT) {
    error (sbp, get_nfsstat3 (status));
    return;
  }

  rename3args *a = sbp->template getarg<rename3args> ();
  cfscallp rename_c = New refcounted<cfscall>
    (CFSRWPROC_RENAME, wrap (this, &cfsrwcd::rename_reply, sbp));
  cfsrw_rename_arg *rename_arg =
    static_cast<cfsrw_rename_arg *> (rename_c->arg ());
  rename_arg->object = from;
  rename_arg->type = from_type;
  rename_arg->from_dir = nfs2cfsfh (&a->from.dir);
  rename_arg->from_name = a->from.name;
  rename_arg->to_dir = nfs2cfsfh (&a->to.dir);
  rename_arg->to_name = a->to.name;

  if (status == CFSRWERR_NOENT) {
    rename_arg->target_exists = false;
    cfscallp cc = rename_c;
    mkcfscall;
  }

  else {
    rename_arg->target_exists = true;
    
    cfsrw_lookup_res *res = static_cast<cfsrw_lookup_res *> (c->res ());
    rename_arg->target = res->object;
  
    cfscallp cc = New refcounted<cfscall>
      (CFSRWPROC_GETATTR, wrap (this, &cfsrwcd::rename_to_attr,
				sbp, rename_c));
    cfsrw_getattr_arg *arg = static_cast<cfsrw_getattr_arg *> (cc->arg ());
    arg->object = res->object;
    mkcfscall;
  }
}

void
cfsrwcd::rename_from_attr (nfscall *sbp, bigint from,
                           cfscallp c, cfsrw_status status)
{
  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }
 
  cfsrw_getattr_res *res = static_cast<cfsrw_getattr_res *> (c->res ());
      
  rename3args *a = sbp->template getarg<rename3args> ();
  cfscallp cc = New refcounted<cfscall>
    (CFSRWPROC_LOOKUP, wrap (this, &cfsrwcd::rename_to, sbp,
			     from, res->attributes.type));
  cfsrw_lookup_arg *arg = static_cast<cfsrw_lookup_arg *> (cc->arg ());
  arg->dir = nfs2cfsfh (&a->to.dir);
  arg->name = a->to.name;
  mkcfscall;
}

void
cfsrwcd::rename_from (nfscall *sbp, cfscallp c, cfsrw_status status)
{
  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }
  
  cfsrw_lookup_res *res = static_cast<cfsrw_lookup_res *> (c->res ());
  cfscallp cc = New refcounted<cfscall>
    (CFSRWPROC_GETATTR, wrap (this, &cfsrwcd::rename_from_attr,
			      sbp, res->object));
  cfsrw_getattr_arg *arg = static_cast<cfsrw_getattr_arg *> (cc->arg ());
  arg->object = res->object;
  mkcfscall;
}


//////////////////////////////////
//  CREATE/MKDIR/SYMLINK/LINK   //
//////////////////////////////////

static inline void
getdirop (nfscall *sbp, bigint &dir, str &name)
{
  if (sbp->proc () == NFSPROC3_CREATE) {
    dir = nfs2cfsfh (&sbp->template getarg<create3args> ()->where.dir);
    name = sbp->template getarg<create3args> ()->where.name;
  }
  else if (sbp->proc () == NFSPROC3_LINK) {
    dir = nfs2cfsfh (&sbp->template getarg<link3args> ()->link.dir);
    name = sbp->template getarg<link3args> ()->link.name;
  }
  else if (sbp->proc () == NFSPROC3_MKDIR) {
    dir = nfs2cfsfh (&sbp->template getarg<mkdir3args> ()->where.dir);
    name = sbp->template getarg<mkdir3args> ()->where.name;
  }
  else {
    dir = nfs2cfsfh (&sbp->template getarg<symlink3args> ()->where.dir);
    name = sbp->template getarg<symlink3args> ()->where.name;
  }
}

void
cfsrwcd::create (nfscall *sbp, cfscallp c)
  // check if directory exists
{
  bigint dir;
  str name;
  getdirop (sbp, dir, name);

  cfscallp cc = New refcounted<cfscall>
    (CFSRWPROC_GETATTR, wrap (this, &cfsrwcd::create_found_dir, sbp));
  cfsrw_getattr_arg *arg = static_cast<cfsrw_getattr_arg *> (cc->arg ());
  arg->object = dir;
  mkcfscall;
}

void
cfsrwcd::create_reply (nfscall *sbp, cfscallp c, cfsrw_status status)
  // after actual dirop is done, cancel exclusive lock
{
  if (status) {
    excl_cancel (sbp, 0, get_nfsstat3 (status), c);
    return;
  }

  switch (sbp->proc ()) {
    case NFSPROC3_CREATE:
    case NFSPROC3_MKDIR:
      {
        cfsrw_create_res *res = static_cast<cfsrw_create_res *> (c->res ());
        excl_cancel (sbp, res->object, NFS3_OK, c);
      }
      break;
    case NFSPROC3_SYMLINK:
      {
        cfsrw_symlink_res *res = static_cast<cfsrw_symlink_res *> (c->res ());
        excl_cancel (sbp, res->object, NFS3_OK, c);
      }
      break;
    case NFSPROC3_LINK:
      excl_cancel (sbp, 0, NFS3_OK, c);
      break;
    default:
      assert(0);
  }
}

void
cfsrwcd::create_found_dir (nfscall *sbp, cfscallp c, cfsrw_status status)
  // found the directory, do actual dirop (except for LINK, for which
  // we must also check to ensure that the object is not a DIR)
{
  if (status) {
    excl_cancel (sbp, 0, NFS3ERR_NOENT, c);
    return;
  }

  bigint dir;
  str name;
  getdirop (sbp, dir, name);

  switch (sbp->proc ()) {
    case NFSPROC3_CREATE:
      {
        create3args *a = sbp->template getarg<create3args> ();
        cfscallp cc = New refcounted<cfscall>
          (CFSRWPROC_CREATE, wrap (this, &cfsrwcd::create_reply, sbp));
        cfsrw_create_arg *arg = static_cast<cfsrw_create_arg *> (cc->arg ());
        arg->dir = dir;
	arg->name = name;
        set_cfsrw_sattr (arg->attributes, *a->how.obj_attributes);
        mkcfscall;
      }
      break;

    case NFSPROC3_MKDIR:
      {
        mkdir3args *a = sbp->template getarg<mkdir3args> ();
        cfscallp cc = New refcounted<cfscall>
          (CFSRWPROC_MKDIR, wrap (this, &cfsrwcd::create_reply, sbp));
        cfsrw_create_arg *arg = static_cast<cfsrw_create_arg *> (cc->arg ());
        arg->dir = dir;
	arg->name = name;
        set_cfsrw_sattr (arg->attributes, a->attributes);
        mkcfscall;
      }
      break;

    case NFSPROC3_SYMLINK:
      {
        symlink3args *a = sbp->template getarg<symlink3args> ();
        cfscallp cc = New refcounted<cfscall>
          (CFSRWPROC_SYMLINK, wrap (this, &cfsrwcd::create_reply, sbp));
        cfsrw_symlink_arg *arg = static_cast<cfsrw_symlink_arg *> (cc->arg ());
        arg->dir = dir;
	arg->name = name;
	arg->linkto = a->symlink.symlink_data;
        set_cfsrw_sattr (arg->attributes, a->symlink.symlink_attributes);
        mkcfscall;
      }
      break;

    case NFSPROC3_LINK:
      {
        link3args *a = sbp->template getarg<link3args> ();
        cfscallp cc = New refcounted<cfscall>
	  (CFSRWPROC_GETATTR, wrap (this, &cfsrwcd::link_found_obj, sbp));
        cfsrw_getattr_arg *arg = static_cast<cfsrw_getattr_arg *>(cc->arg ());
        arg->object = nfs2cfsfh (&a->file);
        mkcfscall;
      }
      break;

    default:
      assert (0);
  }
}

void
cfsrwcd::link_found_obj (nfscall *sbp, cfscallp c, cfsrw_status status)
  // found the object given by the LINK rpc, do the link
{
  if (status) {
    excl_cancel (sbp, 0, get_nfsstat3 (status), c);
    return;
  }

  cfsrw_getattr_res *res = static_cast<cfsrw_getattr_res *> (c->res ());
  if (res->attributes.type == CFSRW_DIR) {
    excl_cancel (sbp, 0, NFS3ERR_ISDIR, c);
    return;
  }

  link3args *a = sbp->template getarg<link3args> ();
  cfscallp cc = New refcounted<cfscall>
    (CFSRWPROC_LINK, wrap (this, &cfsrwcd::create_reply, sbp));
  cfsrw_link_arg *arg = static_cast<cfsrw_link_arg *> (cc->arg ());
  arg->object = nfs2cfsfh (&a->file);
  arg->dir = nfs2cfsfh (&a->link.dir);
  arg->name = a->link.name;
  mkcfscall;
}

void
cfsrwcd::create_unchecked_reply (nfscall *sbp, bigint object,
                                 cfscallp c, cfsrw_status s)
  // unchecked create: after truncate
{
  if (s) {
    error (sbp, get_nfsstat3 (s));
    return;
  }

  diropres3 res (NFS3_OK);
  res.resok->obj.set_present(true);
  nfs_fh3 fh;
  cfsfh2nfs (object, &fh);
  *res.resok->obj.handle = fh;
  reply (sbp, &res);
}

void
cfsrwcd::create_unchecked_lookup (nfscall *sbp, cfscallp c, cfsrw_status s)
  // unchecked create: after lookup, truncate file
{
  if (s) {
    error (sbp, get_nfsstat3 (s));
    return;
  }

  cfsrw_lookup_res *res = static_cast<cfsrw_lookup_res *> (c->res ());
  cfscallp cc = New refcounted<cfscall>
    (CFSRWPROC_SETATTR, wrap (this, &cfsrwcd::create_unchecked_reply,
			      sbp, res->object));
  cfsrw_setattr_arg *arg = static_cast<cfsrw_setattr_arg *> (cc->arg ());
  arg->object = res->object;
  arg->new_attributes.valid_size = 1;
  arg->new_attributes.size = 0;
  mkcfscall;
}

void
cfsrwcd::excl_reply (nfscall *sbp, cfscallp c, cfsrw_status status)
  // after excl attempt: if failed because file exists, and the RPC is
  // a unchecked create, truncate the file. otherwise, if failed,
  // return NFS3ERR_EXISTS (XXX should perhaps back-off and try again
  // if failed because of contention, instead of file exists). if excl
  // succeeded, go on.
{
  if (status == CFSRWERR_EXISTS && sbp->proc () == NFSPROC3_CREATE && 
      sbp->template getarg<create3args> ()->how.mode == UNCHECKED) {

    // do lookup to get the file handle, then truncate the file
    create3args *a = sbp->template getarg<create3args> ();
    cfscallp cc = New refcounted<cfscall>
      (CFSRWPROC_LOOKUP, wrap (this, &cfsrwcd::create_unchecked_lookup, sbp));
    cfsrw_lookup_arg *arg = static_cast<cfsrw_lookup_arg *> (cc->arg ());
    arg->dir = nfs2cfsfh(&a->where.dir);
    arg->name = a->where.name;
    mkcfscall;
    return;
  }

  if (status) {
    error (sbp, get_nfsstat3 (status));
    return;
  }
  create (sbp, c);
}

void
cfsrwcd::excl_cancel_reply (nfscall *sbp, bigint object, nfsstat3 r,
                                 cfscallp c, cfsrw_status status)
  // exclusive lock been canceled
{
  if (!r) {
    switch (sbp->proc ()) {
      case NFSPROC3_CREATE:
      case NFSPROC3_MKDIR:
      case NFSPROC3_SYMLINK:
	{
	  diropres3 res (NFS3_OK);
          res.resok->obj.set_present(true);
	  nfs_fh3 fh;
	  cfsfh2nfs (object, &fh);
          *res.resok->obj.handle = fh;
          reply (sbp, &res);
	}
	break;
      case NFSPROC3_LINK:
	{
	  link3res res (NFS3_OK);
          reply (sbp, &res);
	}
	break;
      default:
	assert (0);
	break;
    }
  }
  else
    error (sbp, r);
}

void
cfsrwcd::excl_cancel (nfscall *sbp, bigint fh, nfsstat3 r, cfscallp c)
  // cancel exclusive lock on dirop
{
  bigint dir;
  str name;
  getdirop (sbp, dir, name);

  cfscallp cc = New refcounted<cfscall>
    (CFSRWPROC_EXCL,
     wrap (this, &cfsrwcd::excl_cancel_reply, sbp, fh, r));
  cfsrw_excl_arg *arg = static_cast<cfsrw_excl_arg *> (cc->arg ());
  arg->dir = dir;
  arg->name = name;
  arg->cancel = true;
  mkcfscall;
}


/////////////////
//  dispatch   //
/////////////////

void
cfsrwcd::dispatch_cfsrw (nfscall *sbp, cfscallp c)
{
  int proc = sbp->proc ();

  switch(proc) {
  case NFSPROC3_GETATTR:
    {
      cfscallp cc = New refcounted<cfscall>
	(CFSRWPROC_GETATTR, wrap (this, &cfsrwcd::getattr_reply, sbp));
      cfsrw_getattr_arg *arg = static_cast<cfsrw_getattr_arg *>(cc->arg ());
      arg->object = nfs2cfsfh (sbp->template getarg<nfs_fh3> ());
      mkcfscall;
    }
    break;

  case NFSPROC3_ACCESS:
    {
      cfscallp cc = New refcounted<cfscall>
	(CFSRWPROC_GETATTR, wrap (this, &cfsrwcd::access_reply, sbp));
      cfsrw_getattr_arg *arg = static_cast<cfsrw_getattr_arg *>(cc->arg ());
      access3args *a = sbp->template getarg<access3args> ();
      arg->object = nfs2cfsfh (&a->object);
      mkcfscall;
    }
    break;

  case NFSPROC3_LOOKUP:
    {
      diropargs3 *dirop = sbp->template getarg<diropargs3> ();
      bigint dirfh = nfs2cfsfh (&(dirop->dir));

      if (dirop->name == ".") {
        lookup3res nfsres(NFS3_OK);
        nfsres.resok->object = dirop->dir;
	reply (sbp, &nfsres);
      }

      else if (dirop->name == "..") {
        cfscallp cc = New refcounted<cfscall>
	  (CFSRWPROC_GETPARENT,
	   wrap (this, &cfsrwcd::lookup_dotdot_reply, sbp));
        cfsrw_getparent_arg *arg =
	  static_cast<cfsrw_getparent_arg *> (cc->arg ());
        arg->dir = dirfh;
        mkcfscall;
      }

      else {
        cfscallp cc = New refcounted<cfscall>
	  (CFSRWPROC_LOOKUP, wrap (this, &cfsrwcd::lookup_reply, sbp));
        cfsrw_lookup_arg *arg = static_cast<cfsrw_lookup_arg *> (cc->arg ());
        arg->dir = dirfh;
	arg->name = dirop->name;
        mkcfscall;
      }
    }
    break;
  
  case NFSPROC3_WRITE:
    {
      write3args *a = sbp->template getarg<write3args> ();
      bigint file = nfs2cfsfh (&a->file);
      
      cfscallp cc = New refcounted<cfscall>
	(CFSRWPROC_GETATTR, wrap (this, &cfsrwcd::write_found_file, sbp));
      cfsrw_getattr_arg *arg = static_cast<cfsrw_getattr_arg *> (cc->arg ());
      arg->object = file;
      mkcfscall;
    }
    break;

  case NFSPROC3_SETATTR:
    {
      setattr3args *a = sbp->template getarg<setattr3args> ();
      bigint object = nfs2cfsfh (&a->object);
      
      cfscallp cc = New refcounted<cfscall>
	(CFSRWPROC_GETATTR, wrap (this, &cfsrwcd::setattr_found_obj, sbp));
      cfsrw_getattr_arg *arg = static_cast<cfsrw_getattr_arg *> (cc->arg ());
      arg->object = object;
      mkcfscall;
    }
    break;

  case NFSPROC3_READ:
    {
      read3args *a = sbp->template getarg<read3args> ();
      bigint file = nfs2cfsfh (&a->file);

      cfscallp cc = New refcounted<cfscall>
	(CFSRWPROC_GETATTR, wrap (this, &cfsrwcd::read_found_file, sbp));
      cfsrw_getattr_arg *arg = static_cast<cfsrw_getattr_arg *> (cc->arg ());
      arg->object = file;
      mkcfscall;
    }
    break;

  case NFSPROC3_READLINK:
    {
      cfscallp cc = New refcounted<cfscall>
	(CFSRWPROC_GETATTR, wrap (this, &cfsrwcd::readlink_found_file, sbp));
      cfsrw_getattr_arg *arg = static_cast<cfsrw_getattr_arg *> (cc->arg ());
      arg->object = nfs2cfsfh (sbp->template getarg<nfs_fh3> ());
      mkcfscall;
    }
    break;

  case NFSPROC3_REMOVE:
  case NFSPROC3_RMDIR:
    {
      diropargs3 *dirop = sbp->template getarg<diropargs3> ();
      bigint dirfh = nfs2cfsfh (&dirop->dir);

      cfscallp cc = New refcounted<cfscall>
	(CFSRWPROC_LOOKUP, wrap (this, &cfsrwcd::remove_found_file, sbp));
      cfsrw_lookup_arg *arg = static_cast<cfsrw_lookup_arg *> (cc->arg ());
      arg->dir = dirfh;
      arg->name = dirop->name;
      mkcfscall;
    }
    break;

  case NFSPROC3_READDIR:
    {
      readdir3args *a = sbp->template getarg<readdir3args> ();
      cfscallp cc = New refcounted<cfscall>
	(CFSRWPROC_GETATTR, wrap (this, &cfsrwcd::readdir_found_dir, sbp));
      cfsrw_getattr_arg *arg = static_cast<cfsrw_getattr_arg *> (cc->arg ());
      arg->object = nfs2cfsfh (&a->dir);
      mkcfscall;
    }
    break;

  case NFSPROC3_RENAME:
    {
      rename3args *a = sbp->template getarg<rename3args> ();
      cfscallp cc = New refcounted<cfscall>
	(CFSRWPROC_LOOKUP, wrap (this, &cfsrwcd::rename_from, sbp));
      cfsrw_lookup_arg *arg = static_cast<cfsrw_lookup_arg *> (cc->arg ());
      arg->dir = nfs2cfsfh (&a->from.dir);
      arg->name = a->from.name;
      mkcfscall;
    }
    break;

  case NFSPROC3_CREATE:
    if (sbp->template getarg<create3args> ()->how.mode == EXCLUSIVE) {
      error (sbp, NFS3ERR_NOTSUPP);
      return;
    }

    // falls thru
  case NFSPROC3_LINK:
  case NFSPROC3_MKDIR:
  case NFSPROC3_SYMLINK:
    {
      bigint dir;
      str name;
      getdirop (sbp, dir, name);

      cfscallp cc = New refcounted<cfscall>
        (CFSRWPROC_EXCL, wrap (this, &cfsrwcd::excl_reply, sbp));
      cfsrw_excl_arg *arg = static_cast<cfsrw_excl_arg *> (cc->arg ());
      arg->dir = dir;
      arg->name = name;
      arg->cancel = false;
      mkcfscall;
    }
    break;

  default:
    sbp->error (NFS3ERR_NOTSUPP);
    return;
  }

  build_snapshot_with_v (c->l, c->start, LOG_OK);
}

void
cfsrwcd::error (nfscall *sbp, nfsstat3 err)
{
  void *resp;

#define mkerr(proc, arg, res)                   \
case proc:                                      \
  {                                             \
    resp = New res (err);                       \
    break;                                      \
  }
  
  switch (sbp->proc ()) {
    NFS_PROGRAM_3_APPLY_NONULL (mkerr);
  default:
    panic ("nfscall::error: invalid proc %d\n", sbp->proc ());
  }

#undef mkerr

  nfs_fh3 fh;
  reply (sbp, resp, fh);
}

#undef mkcfscall


