/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "sfscd_prot.h"
#include "nfs3close_prot.h"
#include "nfs3_nonnul.h"
#include "rxx.h"
#include "nfstrans.h"

#include "definc.h"
#include "ns.h"
#include "log.h"
#include "snapshot_spec.h"

#include "nfsxfattr.h"
#include "fcache.h"
#include "cfsrwcd.h"
#include "cfsa_prot.h"


int cfsrw_verbose = ENV("CFSRW_VERBOSE",0);
int print_stat    = ENV("CFSRW_STAT", 0);
int use_fcache    = ENV("CFSRW_FCACHE", 1);
int use_acache    = ENV("CFSRW_ACACHE", 1);

unsigned ss_thresh = ENV("SNAPSHOT_THRESH",20);
unsigned ss_interval = ENV("SNAPSHOT_INTERVAL",60);

int ncacheflush = 0;
int nsyncwrite = 0;
  
qhash<sfs_aid, ref<aclnt> > cfsrwcd::cahash;

void
cfsrwcd::dispatch (ref<nfsserv> ns, nfscall *sbp)
{
  dispatch_nfs (sbp);
}

void
cfsrwcd::dispatch_nfs (nfscall *sbp)
{
  int proc = sbp->proc();

  if (readonly (sbp)) {
    sbp->error (NFS3ERR_ROFS);
    return;
  }

  if (proc == NFSPROC3_NULL) {
    sbp->reply (NULL);
    return;
  }

  else if (proc == NFSPROC3_FSSTAT) {
    fsstat3res res (NFS3_OK);
    res.resok->obj_attributes.set_present(false);
    res.resok->tbytes = INT64 (0x7fffffffffffffff);
    res.resok->fbytes = INT64 (0x7fffffffffffffff);
    res.resok->abytes = INT64 (0x7fffffffffffffff);
    res.resok->tfiles = INT64 (0x7fffffffffffffff);
    res.resok->ffiles = INT64 (0x7fffffffffffffff);
    res.resok->afiles = INT64 (0x7fffffffffffffff);
    res.resok->invarsec = 0;
    sbp->reply (&res);
    return;
  }

  else if (proc == NFSPROC3_FSINFO) {
    fsinfo3res res (NFS3_OK);
    res.resok->rtmax = blocksize;
    res.resok->rtpref = blocksize;
    res.resok->rtmult = 512;
    res.resok->wtmax = blocksize;
    res.resok->wtpref = blocksize;
    res.resok->wtmult = 4096;
    res.resok->dtpref = blocksize;
    res.resok->maxfilesize = INT64 (0x7fffffffffffffff);
    res.resok->time_delta.seconds = 0;
    res.resok->time_delta.nseconds = 1;
    res.resok->properties = (FSF3_LINK | FSF3_SYMLINK | FSF3_HOMOGENEOUS);
    sbp->reply (&res);
    return;
  }

  if (proc == NFSPROC3_ACCESS && print_stat) {
    access3args *a = sbp->template getarg<access3args> ();
    bigint dir = nfs2cfsfh (&a->object);
    if (dir == _view->root()) {
      ns::stat();
      warn << nsyncwrite << " synchronous writes\n";
      warn << ncacheflush << " cache flushes\n";
    }
  }

  if (use_fcache && use_acache) {
    if (fcache_dispatch (sbp))
      return;
  }

  else if (sbp->proc () == cl_NFSPROC3_CLOSE || 
           sbp->proc () == NFSPROC3_COMMIT) {
    commit3res res (NFS3_OK);
    res.resok->verf = _verf3;
    sbp->reply (&res);
    return;
  }

  _view->version (wrap(this, &cfsrwcd::dispatch_view, sbp));
}

static inline bool
ctime_recent (timespec ctime)
{
  time_t t = tsnow.tv_sec - ctime.tv_sec;
  if (t > CACHE_RECENT)
    return false;
  return true;
}

void
cfsrwcd::dispatch_view (nfscall *sbp, vervec start, log_status status)
{
  if (status) {
    sbp->error (get_nfsstat3 (status));
    return; 
  }
  
  // check if RPC can be answered using the attribute cache

  if (use_acache) {
    // we update the attribute cache after every modification to the
    // file system. if we are the only writer to the file system, then
    // the attribute cache can be used.
    //
    // flush attr cache if
    //  - current fs version does not equal to attr cache version
    //    AND
    //    - current user cannot write to file system OR
    //    - two versions have different sequence numbers other than the
    //      current user's sequence number OR
    //    - attr cache is more than CACHE_RECENT seconds old
    bigint id;
    bool rw = uid (sbp, id);
    if (!_ac.current_version ().equals (start) &&
        (!rw ||
         !_ac.current_version ().equals_except (start, id) ||
         !ctime_recent (_ac.cache_time ())))
      _ac.flush_attr ();
  
    // update the attr cache version to the current version
    _ac.current_version (start);
  
    switch (sbp->proc()) {
    case NFSPROC3_GETATTR:
      {
        nfs_fh3 fh = *sbp->template getarg<nfs_fh3> ();
        const fattr3 *f = _ac.attr_lookup (start, fh);
        if (f) {
	  WARN(1) << "cached getattr: " << fh << "\n";
	  getattr3res res (NFS3_OK);
	  *res.attributes = *reinterpret_cast<const fattr3 *> (f);
	  sbp->reply (&res);
	  return;
        }
        WARN(1) << "sending getattr: " << fh << "\n";
      }
      break;
    case NFSPROC3_ACCESS:
      {
        access3args *a = sbp->template getarg<access3args> ();
        int32_t perm = _ac.access_lookup
	  (start, a->object, sbp->getaid (), a->access);
        if (perm > 0) {
          fattr3 fa = *reinterpret_cast<const fattr3 *>
	    (_ac.attr_lookup (start, a->object));
	  
	  WARN(1) << "cached access: " << a->object << "\n";
          if (fa.type == NF3REG) {
            file_cache *e = file_cache_lookup(a->object);
            if (!e) {
              file_cache_insert (a->object);
              e = file_cache_lookup(a->object);
              assert(e);
              e->open();
            }
            else if (e->is_idle())
              e->open();
	  }
  
          access3res res(NFS3_OK);
          res.resok->obj_attributes.set_present (true);
          *res.resok->obj_attributes.attributes = fa;
          res.resok->access = perm;
          sbp->reply (&res);
          return;
        }
        WARN(1) << "sending access: " << a->object << "\n";
      }
      break;
    default:
      break;
    }
  }

  // just a place holder for snapshot object, starting vector, etc.
  cfscallp c = get_cfscall (sbp, start);
  if (sbp->proc () == NFSPROC3_WRITE || c->s ||
      load_snapshot_and_continue (sbp, c))
    dispatch_cfsrw (sbp, c);
}

cfscallp
cfsrwcd::get_cfscall (nfscall *sbp, vervec start)
{
  cfscallp c = New refcounted<cfscall> ();
  c->l = mylog (sbp);
  c->start = start;
  c->s = 0;
  if (c->l) {
    ptr<snaprec>* ss;
    if ((ss = _snapshots [c->l->head()]) && (*ss)->set) {
      c->s = (*ss)->image;
      c->stop = (*ss)->version;
    }
  }
  return c;
}

// extract attributes from replies from cfsrw and put them into the
// attribute cache

void
cfsrwcd::reply (nfscall *sbp, void *res)
{
  nfs_fh3 fh;
  reply (sbp, res, fh);
}

void
cfsrwcd::reply (nfscall *sbp, void *res, nfs_fh3 a)
{
  if (use_acache) {
    if (use_fcache)
      fcache_check_reply (sbp, res);

    // flush objects this RPC modified
    if (!nfs_constop (sbp->proc ()))
      flushfattr (sbp->proc (), sbp->getvoidarg (), res);

    // these two RPCs may change the attribute of objects whose file
    // handles are not included in either the arg or the res. the file
    // handle is passed back to us from cfsrw.
    if (sbp->proc () == NFSPROC3_REMOVE ||
	sbp->proc () == NFSPROC3_RENAME) {
      WARN(1) << "removing attr for " << a << "\n";
      _ac.attr_enter(a, 0L, 0L);
    }
    
    // enter new attributes into attribute cache; if this rpc modified
    // the file system, flush objects it changed
    getfattr (sbp->proc (), sbp->getaid (), sbp->getvoidarg (), res);
  }

  sbp->reply (res);
}

void
cfsrwcd::flushfattr (unsigned int proc, void *arg, void *res)
{
  xfhvec xv;
  nfs3_getxfh (&xv, proc, arg, res);
  for (xfh *x = xv.base (); x < xv.lim (); x++) {
    WARN(1) << "removing attr for " << *x->fh << "\n";
    _ac.attr_enter(*x->fh, 0L, 0L);
  }
}

void
cfsrwcd::getfattr (unsigned int proc, sfs_aid aid, void *arg, void *res)
{
  xfattrvec xv;
  nfs3_getxfattr (&xv, proc, arg, res);
  for (xfattr *x = xv.base (); x < xv.lim (); x++) {
    const fattr3 *fa = reinterpret_cast<const fattr3 *> (x->fattr);
    if (fa) {
      WARN(1) << "adding attr for " << *x->fh << "\n";
      _ac.attr_enter (*x->fh, fa, x->wattr);
    }
    if (proc == NFSPROC3_ACCESS) {
      access3res *ares = static_cast<access3res *> (res);
      access3args *a = static_cast<access3args *>(arg);
      if (ares->status)
        _ac.flush_access (a->object, aid);
      else
        _ac.access_enter (a->object, aid, a->access, ares->resok->access);
    }
  }
}

bool
cfsrwcd::uid (nfscall *sbp, bigint &u)
{
  ptr<log> l = mylog (sbp);
  if (l) {
    u = l->head ();
    return true;
  }
  return false;
}

bool
cfsrwcd::readonly (nfscall *sbp)
{
  ptr<log> l = mylog (sbp);
  int proc = sbp->proc();
  if (!l && 
      (proc == NFSPROC3_CREATE || proc == NFSPROC3_WRITE ||
       proc == NFSPROC3_REMOVE || proc == NFSPROC3_SETATTR ||
       proc == NFSPROC3_SYMLINK || proc == NFSPROC3_MKDIR ||
       proc == NFSPROC3_RMDIR || proc == NFSPROC3_RENAME ||
       proc == NFSPROC3_LINK || proc == NFSPROC3_MKNOD ||
       proc == NFSPROC3_COMMIT))
    return true;
  return false;
}

ptr<log>
cfsrwcd::mylog (nfscall *sbp)
  // contacts cfsagent and returns a log object that cfsagent can sign
{
  sfs_aid aid = sbp->getaid();
  ptr<log> *p = _logs[aid];
  if (p && *p != 0) {
    ptr<log> l = *p;
    if (!l->readonly())
      return l;
    else
      return 0;
  }

  if (!p || (sbp->proc() == NFSPROC3_READDIR)) {
    sfs_aid a = sbp->getaid();
    ptr<aclnt> agentclnt = get_agent (a);
    if (agentclnt) {
      sfs_keylist kl;
      if (clnt_stat err = agentclnt->scall (AGENTCTL_DUMPKEYS, NULL, &kl)) {
        warn << "connecting to agent: " << err << "\n";
        _logs.insert(sbp->getaid(), 0);
        return 0;
      }
      for(sfs_keylistelm *e = kl; e; e = e->next) {
	ptr<sfspub> pk = sfscrypt.alloc (e->desc);
	if (!pk)
	  continue;
        strbuf b;
	pk->export_pubkey (b, false);
	str pk_raw = b;
        bigint v = compute_sha1_hash (pk_raw.cstr (), pk_raw.len ());
        for (unsigned i=0; i< _view->nlogs (); i++) {
          if (v == _view->head (i) && _view->latest (i) == _view->head (i)) {
            ptr<log> l = New refcounted<log>
	      (_view->head (i),
	       wrap (this, &cfsrwcd::got_log, a),
	       wrap (this, &cfsrwcd::sign_log_head, a, _view->head (i)));
            _logs.insert (a,l);
            return 0;
          }
        }
      }
    }
    _logs.insert (sbp->getaid (), 0);
  }
  return 0;
}

void
cfsrwcd::got_log (sfs_aid a, log_status status)
{
  if (status)
    _logs.remove (a);
}

ptr<aclnt>
cfsrwcd::get_agent (sfs_aid aid)
{
  ptr<aclnt> c = cahash[aid];
  if (c)
    return c;

  // get a connection to the user's agent; there may be a better way
  // to do this that does not involve a fork...

  int fds[2];
  int cfd = -1;

  if (socketpair (AF_UNIX, SOCK_STREAM, 0, fds) < 0)
    fatal("socketpair: %m\n");

  pid_t pid = fork();
  if (pid) { // parent
    close (fds[1]);
    char c;
    readfd(fds[0], &c, 1, &cfd);
    int status;
    wait(&status);
  }
  else {
    close (fds[0]);
    setuid(aid);

    cfd = suidgetfd_required("agent");
    if (cfd < 0) {
      warn << "suidgetfd failed\n";
      exit(1);
    }
    if (writefd(fds[1], "", 1, cfd) < 0) {
      warn << "writefd failed\n";
      exit(1);
    }
    exit(0);
  }

  if (cfd < 0) {
    warn << "no file descriptor to sfscd\n";
    return 0;
  }

  ref<axprt_unix> sfscdxprt = axprt_unix::alloc(cfd);
  ref<aclnt> sfscdclnt = aclnt::alloc(sfscdxprt, agent_prog_1);

  int32_t res;
  int afd;
  if (clnt_stat err = sfscdclnt->scall (AGENT_GETAGENT, NULL, &res)) {
    warn << "connect to cdc failed " << err << "\n";
    return 0;
  }
  if ((afd = sfscdxprt->recvfd ()) < 0) {
    warn << "connecting to agent via sfscd: no file descriptor\n";
    return 0;
  }

  ref<aclnt> agentclnt = 
    aclnt::alloc (axprt_stream::alloc (afd), cfs_agentctl_prog_1);
  cahash.insert(aid, agentclnt);

  return agentclnt;
}

void
cfsrwcd::sign_log_head_cb (sfs_aid writer, cfsrw_sign_cb_t cb,
                           cfsa_signhead_res *res, clnt_stat err)
{
  if (!err && !res->status)
    cb (res->resok->k, res->resok->s, LOG_OK);
  else {
    sfs_pubkey2 k;
    sfs_sig2 s;
    cb (k, s, LOGERR_ERR);
    if (err)
      cahash.remove(writer);
  }
  delete res;
}

void
cfsrwcd::sign_log_head (sfs_aid writer, bigint head,
                        const str &data, cfsrw_sign_cb_t cb)
{
  ptr<aclnt> agentclnt = get_agent (writer);
  if (agentclnt) {
    cfsa_signhead_res *res = New cfsa_signhead_res;
    cfsa_signhead_arg a;
    a.keyhash = head;
    a.serialized_head.setsize(data.len());
    memmove(a.serialized_head.base(), data.cstr(), data.len());
    agentclnt->call(CFS_AGENTCTL_SIGNHEAD, &a, res,
	            wrap(this, &cfsrwcd::sign_log_head_cb, writer, cb, res));
  }
  else {
    sfs_pubkey2 k;
    sfs_sig2 s;
    cb (k, s, LOGERR_ERR);
  }
}

bool
cfsrwcd::load_snapshot_and_continue (nfscall *sbp, cfscallp c)
{
  ptr<log> l = c->l;

  if (!l) {
    warn << "no log\n";
    return true;
  }
 
  ptr<snaprec>* ss = 0;
  if ((ss = _snapshots[l->head()]) && (*ss)->set) {
    c->s = (*ss)->image;
    c->stop = (*ss)->version;
    warn << "got snapshot\n";
    return true;
  }
  else if (!ss && l->snapshot () != 0) {
    // need to load snapshot
    ptr<snaprec> s = New refcounted<snaprec>;
    s->image = 0;
    s->set = false;
    s->building = true;
    _snapshots.insert (l->head (), s);
    warn << "loading snapshot " << l->snapshot () << "\n";
    snapshot_operation::load_snapshot 
      (l->snapshot (), wrap(this, &cfsrwcd::load_snapshot, sbp, c));
    return false;
  }

  warn << "no snapshot\n";
  return true;
}

void
cfsrwcd::load_snapshot (nfscall *sbp, cfscallp c,
                        bigint loc, ptr<snapshot_obj> s)
{
  if (!s) {
    warn << "cannot load snapshot, running w/o snapshot\n";
    c->s = 0;
    dispatch_cfsrw (sbp, c);
    return;
  }

  ptr<log> l = c->l;

  ptr<snaprec>* ss = _snapshots[l->head ()];
  assert(ss);

  if (!((*ss)->set)) {
    warn << "snapshot loaded " << loc << "\n";
    (*ss)->building = false;
    (*ss)->location = loc;
    (*ss)->version = s->version ();
    (*ss)->image = s;
    (*ss)->ts = timenow;
    (*ss)->set = true;
  }

  c->s = s;
  c->stop = s->version ();
  dispatch_cfsrw (sbp, c);
}

void
cfsrwcd::build_snapshot (ptr<log> l)
{
  _view->version (wrap (this, &cfsrwcd::build_snapshot_with_v, l));
}

static unsigned
nrecords (const vervec &from, const vervec &to)
{
  unsigned n = 0;
  for (unsigned i=0; i < from.size () && i < to.size (); i++) {
    if (to.seqnum (i) > from.seqnum (i))
      n += (to.seqnum (i) - from.seqnum (i));
  }
  return n;
}

void
cfsrwcd::build_snapshot_with_v (ptr<log> l, vervec v, log_status status)
{
  if (status)
    return;

  bool doit = false;
  if (l) {
    ptr<snaprec> *ss = _snapshots[l->head()];
    if (!ss ||
	((*ss)->set && 
	 ((nrecords((*ss)->version, v) > ss_thresh) ||
	  (!(*ss)->version.equals(v) &&
	   ((timenow - (*ss)->ts) > ss_interval)))))
      doit = true;
  }
  if (!doit)
    return;

  ptr<snaprec> *ss = _snapshots[l->head()];
  ptr<snaprec> s = 0;

  if (ss)
    s = *ss;

  if (s && s->set && !s->building && v.gt(s->version)) {
    assert(s->image);
    s->building = true;

    WARN(1) << "build snapshot from previous.\n";
    snapshot_operation::build_snapshot 
      (s->image, s->location, v, 
       wrap (this, &cfsrwcd::snapshot_created, l));
  }

  else if (!s) {
    s = New refcounted<snaprec>;
    s->image = 0;
    s->set = false;
    s->building = true;

    WARN(1) << "build snapshot from scratch.\n";
    snapshot_operation::build_snapshot 
      (_view->root (), v,
       wrap (this, &cfsrwcd::snapshot_created, l));
    _snapshots.insert(l->head(),s);
  }
}

void 
cfsrwcd::snapshot_created (ptr<log> l, bigint sloc, ptr<snapshot_obj> image)
{
  ptr<snaprec> *ss = _snapshots[l->head()];
  assert(ss);
  assert(!l->readonly());
  ptr<snaprec> s = *ss;
  if (image != 0) {
    warn << "snapshot created " << sloc << "\n";
    s->location = sloc;
    l->snapshot (sloc);
    s->version = image->version();
    s->image = image;
    s->ts = timenow;
    s->building = false;
    s->set = true;
  }
  else {
    if (s->image) {
      warn << "snapshot creation failed, use old one\n";
      s->building = false;
      s->set = true;
    }
    else {
      warn << "snapshot creation failed, use old one\n";
      _snapshots.remove (l->head());
    }
  }
  build_snapshot (l);
}



cfsrwcd::cfsrwcd
(sfsprog &p, bigint v, callback<void, cfsrwcd*, nfs_fh3 *>::ref cb)
  : _prog (p), _vid (v),
    _fc (128, wrap (this, &cfsrwcd::file_cache_gc_remove)),
    _view (New refcounted<view> (wrap(this, &cfsrwcd::got_view, cb), v)),
    _cfsrw (_view)
{
  bigint verf = _cfsrw.verf ();
  mpz_get_rawmag_be (_verf3.base(), NFS3_WRITEVERFSIZE, &verf);

  strbuf s;
  sbigint (s, _vid);
  str ss = armor64A(s);
  _cdir = "/var/tmp/ivy/fcache/" << ss;
  if (mkdir(_cdir.cstr(), 0755) < 0 && errno != EEXIST)
    fatal ("cannot create cache directory %s\n", _cdir.cstr());
  for (int i=0; i<254; i++) {
    str f = _cdir << "/" << i;
    if (mkdir(f.cstr(), 0755) < 0 && errno != EEXIST)
      fatal ("cannot create cache directory %s\n", f.cstr());
  }
}

void
cfsrwcd::got_view
(callback<void, cfsrwcd*, nfs_fh3 *>::ref cb, log_status status)
{
  if (status) {
    warn << "could not fetch view block\n";
    (*cb) (this, NULL);
    return;
  }
  WARN(1) << "got view block. id = " << _vid << "\n";
  
  nfs_fh3 *fh = New nfs_fh3;
  bigint n = _view->root ();
  warn << "root inum " << n << "\n";
  strbuf s;
  sbigint (s, n);
  str ss = armor64A(s);
  warn << "root inum " << ss << "\n";
  cfsfh2nfs (n, fh);
  cb (this, fh);
}
  

void
gotrootfh (ref<nfsserv> ns, sfsserver::fhcb cb, cfsrwcd *server, nfs_fh3 *fh)
{
  if (fh) {
    if (!ns->encodefh (*fh))
      (*cb) (NULL);
    else {
      (*cb) (fh);
      ns->setcb (wrap (server, &cfsrwcd::dispatch, ns));
    }
  }
  else
    (*cb) (NULL);
}

void
cfsrwcd_alloc(sfsprog *prog, ref<nfsserv> ns, int tcpfd,
              sfscd_mountarg *ma, sfsserver::fhcb cb)
{
  if (!ma->cres ||
      (ma->carg.civers == 5 && !sfs_parsepath (ma->carg.ci5->sname))) {
    warn << "request to mount " << ma->carg.ci5->sname << "\n";
  
    rxx path_regexp ("chord:([0-9a-zA-Z+-]*)$");
    if (path_regexp.search (ma->carg.ci5->sname)) {
      bigint view;
      str rootfhstr = path_regexp[1];
      str dearm = dearmor64A(rootfhstr);
      mpz_set_rawmag_be (&view, dearm.cstr(), dearm.len());
      vNew cfsrwcd (*prog, view, wrap (&gotrootfh, ns, cb));
      return;
    }
    
    warn << "malformated filesystem name: "
         << ma->carg.ci5->sname << "\n";
    (*cb)(NULL);
    return;
  }
  fatal << "Ivy must run in a connectinfo version 5 SFS environment\n";
}

int
main (int args, char **argv) 
{
  setprogname (argv[0]);
  
  sfsconst_init ();
  random_init_file (sfsdir << "/random_seed");

  if (mkdir("/var/tmp/ivy", 0755) < 0 && errno != EEXIST) {
    warn << "cannot create directory /var/tmp/ivy\n";
    return -1;
  }
  if (mkdir("/var/tmp/ivy/fcache", 0755) < 0 && errno != EEXIST) {
    warn << "cannot create directory /var/tmp/ivy/fcache\n";
    return -1;
  }
  ns::init ();
  if (snapshot_operation::init () < 0) {
    warn << "cannot create snapshot related directories\n";
    return -1;
  }

  warn << "cfsrwcd version " << VERSION
       << " running under PID " << getpid () << "\n";
  sfsconst_init ();
  if (ptr<axprt_unix> x = axprt_unix_stdin ())
    vNew sfsprog (x, &cfsrwcd_alloc, false, true);
  else
    fatal ("Couldn't allocate unix transport.\n");
  sigcb(SIGTERM, wrap(exit, 0)); // for profiling.
  amain();
}

