/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "ranges.h"
#include "fcache.h"
#include "cfsrwcd.h"

struct write_obj {
  static const unsigned int PARALLEL_WRITES = 32;
  typedef callback<void,bool>::ref cb_t;

  cb_t cb;
  cfsrwcd& srv;
  cfsrw& c;
  cfscallp cfscall_ptr;
  file_cache *fe;
  nfs_fh3 fh;
  bigint file;
  uint64 start;
  uint64 size;
  uint64 written;
  unsigned int outstanding_writes;
  bool callback;
  unsigned int blocks;

  void
  write_reply (cfscallp c, cfsrw_status status) {
    blocks++;
    // only make sense to build once per PARALLEL_WRITES blocks
    // anyways, because that's when our log head is likely to be
    // flushed to DHash. we try to build at a slightly faster pace
    // since building snapshot is slower than speed at which we can
    // write to DHash.
    if (blocks == PARALLEL_WRITES/2) {
      srv.build_snapshot (cfscall_ptr->l);
      blocks = 0;
    }
    outstanding_writes--;
    if (!callback && !status) {
      do_write();
      ok ();
    }
    else
      fail();
  }

  void cfsrw_write (uint64 off, uint32 cnt) {
    ptr<aiobuf> buf = cfsrwcd::a->bufalloc (cnt);
    if (!buf) {
      cfsrwcd::a->bufwait
	(wrap (this, &write_obj::cfsrw_write, off, cnt));
      return;
    }
    outstanding_writes++;
    fe->afh->read (off, buf,
	           wrap (this, &write_obj::cfsrw_write_read, off, cnt));
  }

  void cfsrw_write_read (uint64 off, uint32 cnt, ptr<aiobuf> buf,
                         ssize_t sz, int err)
  {
    if (err || (unsigned)sz != cnt) {
      outstanding_writes--;
      if (err)
        warn << "flush_cache: read failed: " << err << "\n";
      else 
	warn << "flush_cache: short read: got "
	     << sz << " wanted " << cnt << "\n";
      fail ();
      return;
    }

    cfscallp cc = New refcounted<cfscall>
      (CFSRWPROC_WRITE, wrap (this, &write_obj::write_reply));
    cfsrw_write_arg *arg = static_cast<cfsrw_write_arg *> (cc->arg ());
    arg->file = file;
    arg->offset = off;
    arg->count = cnt;
    arg->data.setsize (cnt);
    memmove (arg->data.base (), buf->base (), cnt);
    cc->l = cfscall_ptr->l;
    cc->start = cfscall_ptr->start;
    cc->s = cfscall_ptr->s;
    cc->stop = cfscall_ptr->stop;
    c.dispatch (cc);
  }

  bool do_write () {
    if (written < size) {
      unsigned s = size-written;
      s = s > cfsrwcd::blocksize ? cfsrwcd::blocksize : s;
      cfsrw_write (start+written, s);
      written += s;
    }
    if (written < size)
      return false;
    return true;
  }

  static void file_closed (int) { }

  void fail () {
    if (!callback) {
      fe->afh->close (wrap (&write_obj::file_closed));
      fe->afh = 0;
      callback = true;
      cb (false);
    }
    if (outstanding_writes == 0)
      delete this;
  }

  void ok () {
    if (outstanding_writes == 0) {
      if (!callback) {
        srv.build_snapshot (cfscall_ptr->l);
        fe->afh->close (wrap (&write_obj::file_closed));
        fe->afh = 0;
        callback = true;
	cb (true);
      }
      delete this;
    }
  }

  write_obj (str fn, file_cache *fe, uint64 start, uint64 size,
             cfsrwcd& srv, cfsrw& c, vervec v, write_obj::cb_t cb,
	     nfscall *sbp)
    : cb(cb), srv(srv), c(c), fe(fe), fh(fe->fh),
      start(start), size(size), written(0),
      outstanding_writes(0), callback(false)
  {
    assert (fe->afh);
    file = nfs2cfsfh (&fh);
    cfscall_ptr = srv.get_cfscall (sbp, v);
    blocks = 0;

    for (unsigned i=0; i<PARALLEL_WRITES; i++)
      do_write ();
    ok ();
  }

  ~write_obj() {}
};

void
cfsrwcd::flush_cache (file_cache *e, nfscall *sbp)
{
  assert (e->is_dirty());
  e->flush ();
  str fn = fh2fn (e->fh);
  _view->version (wrap (this, &cfsrwcd::flush_cache_with_v, e, sbp));
}

void
cfsrwcd::flush_cache_with_v (file_cache *e, nfscall *sbp,
                             vervec v, log_status status)
{
  if (status) {
    e->error ();
    sbp->error (get_nfsstat3 (status));
    return;
  }

  uint64 start = e->mstart;
  uint64 size = e->mend - e->mstart;

  if (sbp->proc () == NFSPROC3_COMMIT) {
    commit3args *a = sbp->template getarg<commit3args> ();
    start = a->offset;
    size = a->count;
  }

  if (start <= e->mstart)
    e->mstart = start + size;
  if (start + size >= e->mend)
    e->mend = start;
  if (e->mstart >= e->mend) {
    e->mstart = 0;
    e->mend = 0;
  }

  vNew write_obj (fh2fn (e->fh), e, start, size, *this, _cfsrw, v,
                  wrap(this, &cfsrwcd::flush_done, e, sbp), sbp);
}

