/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "definc.h"
#include "log.h"
#include "ns.h"
#include "pack.h"
#include "rxx.h"
#include <time.h>

void
log::init ()
{
  _locked = false;
  _release_cb = 0;
  _new_snapshot = 0;
  _snapshot = 0;
  if (sign)
    _readonly = false;
  else
    _readonly = true;
  _head_record = 0;
  _stores = _failed_stores = 0;
  _refreshed = false;
  _loading = false;
}

log::log (bigint head, cb_logstat_t cb, cfsrw_sign_t s)
  : _head (head), sign(s)
{
  init ();
  ns::retrieve_loghead (_head, wrap(mkref(this), &log::load_head_cb, cb));
}

void
log::load_head (recordp e)
{
  assert(!_refreshed);
  if (e) {
    if (_appends_e.size() > 0)
      _refreshed = true;
    _head_record = e;
    head_log_record* h = e->template record<head_log_record>();
    assert(h);
    _last_id = h->prev ();
    _last_seq = h->seqnum ()-1;
    _root = h->root ();
    // load snapshot location; make sure we don't clobber a snapshot
    // we just created
    if (h->snapshot () != 0 && (_snapshot == 0 || _new_snapshot == 0)) {
      _snapshot = h->snapshot ();
      _new_snapshot = 0;
    }
  }
}

void
log::load_head_cb (cb_logstat_t cb, recordp e)
{
  load_head (e);
  if (e)
    cb (LOG_OK);
  else
    cb (LOGERR_NOLOGHEAD);
}

static void
call_callback (cb_logstat_t cb)
{
  cb (LOG_OK);
}

void
log::append (const vervec& v, recordp e, cb_logstat_t cb)
{
  assert(!_readonly);

  vervec real_v = v;
  for (unsigned i=0; i < real_v.size (); i++) {
    if (real_v.head (i) == _head && real_v.seqnum (i) < _last_seq) {
       real_v.add (real_v.head (i), _last_id, _last_seq);
      break;
    }
  }

  e->set_version(real_v);

  if (_locked) {
    // if log object is locked, we do all inserts asynchronously. on
    // release, if there are errors, we don't insert the log-head and
    // report error to user.
    timecb (tsnow, wrap (call_callback, cb));

    // don't ever load a new log-head if we are appending to the log
    // while the log object is locked.
    _loading = true;
    _refreshed = true;
    
    _appends_e.push_back (e);
    _appends_cb.push_back (0);
  }
  else {
    _appends_e.push_back(e);
    _appends_cb.push_back(cb);
  }

  if (!_loading) {
    _loading = true;
    // need to load a fresh head record
    ns::retrieve_loghead 
      (_head, wrap(mkref(this), &log::load_head_cb,
		   wrap(mkref(this), &log::do_append)));
  }
  else if (_refreshed)
    do_append (LOG_OK);
}

void
log::do_append (log_status status)
{
  bool abort = false;
  if (status)
    abort = true;

  for(unsigned i=0; i<_appends_e.size(); i++) {
    recordp e = _appends_e[i];

    if (!abort) {
      if (e->version().conflicts(_head_record->version())) {
        // oops, another writer, in a more up-to-date view perhaps, has
        // written a head of log. that means this view is out of date
        // and should be made read-only!
        abort = true;
        _readonly = true;
        _head_record = 0;
      }
      else {
        vervec v = e->version();
        _head_record->set_version(v);
      }
    }
    _store_cbs.push_back(_appends_cb[i]);

    if (!abort) {
      e->set_prev (_head_record->prev ());
      e->set_head (_head);
      e->set_seqnum (_head_record->seqnum());
      _stores++;
  
      str data = e->serialize();
      bigint id = compute_sha1_hash (data.cstr(), data.len());
      _head_record->set_prev (id);
      _head_record->set_seqnum (_head_record->seqnum ()+1);
      _last_id = id;
      _last_seq = e->seqnum ();
      store_record
        (id, data, 
         wrap(mkref(this), &log::store_head_after_append), e);
    }
  }
  _appends_e.clear();
  _appends_cb.clear();

  if (abort) {
    _stores++;
    store_head_after_append (LOGERR_ERR);
  }
}


void
log::store_record_cb (cb_logstat_t cb, ns_stat status, bigint cid)
{
  if (status != NS_OK)
    cb (LOGERR_INSERT);
  else
    cb (LOG_OK);
}

void
log::store_head (recordp e, const str data, cb_logstat_t cb,
                 sfs_pubkey2 k, sfs_sig2 s, log_status status)
{
  if (!status)
    ns::insert_loghead (_head, k, s, data.cstr(), data.len(),
	                wrap(mkref(this), &log::store_record_cb, cb), e);
  else
    store_record_cb (cb, NS_ERR, 0);
}

void
log::store_record (bigint key, const str &data, cb_logstat_t cb, recordp e)
{
  assert(e);
  if (key == _head) {
    assert (sign);
    sign (data, wrap(mkref(this), &log::store_head, e, data, cb));
    return;
  }
  ns::insert_record (key, data.cstr(), data.len(),
                     wrap (mkref(this), &log::store_record_cb, cb), e);
}

void
log::head_stored (ptr<vec<cb_logstat_t> > cbs, ptr<head_log_record> h,
                  log_status status)
{
  if (!status && cbs) {
    for (unsigned i=0; i<cbs->size (); i++) {
      if ((*cbs)[i])
	(*cbs)[i] (LOG_OK);
    }
  }
  else {
    if (cbs) {
      for (unsigned i=0; i<cbs->size (); i++) {
        if ((*cbs)[i])
          (*cbs)[i] (LOGERR_INSERT);
      }
    }
  }
  if (_release_cb) {
    _release_cb (status);
    _release_cb = 0;
  }
  if (_appends_e.size() > 0) {
    assert (_loading = true);
    _refreshed = true;
    do_append (LOG_OK);
  }
  else
    _loading = false;
}

void 
log::snapshot (bigint s)
{
  _snapshot = s;
  _new_snapshot = s;
}

void
log::release (cb_logstat_t cb)
{
  _locked = false;
  _release_cb = cb;
  _stores++;
  store_head_after_append (LOG_OK);
}

void
log::sync (cb_logstat_t cb)
{
  _release_cb = cb;
  _stores++;
  store_head_after_append (LOG_OK);
}

void
log::store_head_after_append (log_status status)
{
  _stores--;
  if (status)
    _failed_stores++;
  if (_stores == 0) {
    if (_failed_stores == 0) {
      _head_record->set_time();
      head_log_record* h = _head_record->template record<head_log_record> ();
      if (_new_snapshot != 0) {
	h->snapshot (_new_snapshot);
	_new_snapshot = 0;
      }
      ptr<vec<cb_logstat_t> > cbs = New refcounted<vec<cb_logstat_t> >();
      for (unsigned i=0; i<_store_cbs.size(); i++)
	cbs->push_back(_store_cbs[i]);
      _store_cbs.clear();
      ptr<head_log_record> e = New refcounted<head_log_record>(*h);
      
      if (_locked && _release_cb == 0) {
	// log is locked, and there is no explict sync request. return
	// to caller w/o writing log-head, user will explicitly write
	// the log-head later
	timecb
	  (tsnow, wrap (mkref(this), &log::head_stored, cbs, e, LOG_OK));
      }
      else {
        str data = _head_record->serialize();
        store_record
	  (_head, data, wrap(mkref(this), &log::head_stored, cbs, e), e);
      }
      // leave loading to true and refreshed to false: if a request
      // comes in while DHash is storing the head block, the request
      // is saved (because _loading is true), and we won't try to load
      // a new head block until after the old head block is saved.
      _loading = true;
      _refreshed = false;
    }
    else {
      for (unsigned i=0; i<_store_cbs.size(); i++) {
	if (_store_cbs[i])
	  _store_cbs[i] (LOGERR_INSERT);
      }
      _store_cbs.clear();
      // start over, a new request would first grab the version of the
      // head block in DHash
      _loading = false;
      _refreshed = false;
      if (_release_cb) {
	_release_cb (LOGERR_INSERT);
	_release_cb = 0;
	_failed_stores = 0;
      }
    }
    if (!_locked)
      // don't clear _failed_stores counter if the log object is
      // locked, so that on release we can report errors back to user
      _failed_stores = 0;
  }
}

