/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "definc.h"
#include "ns.h"
#include "logiter.h"
#include "view.h"

bool
logiter::check_and_load_record (unsigned i, recordp e)
{
  recordp olde = _entries[i];
  _entries[i] = e;
  _valid[i] = true;

  if (e == 0 || (olde && e->seqnum () >= olde->seqnum ())) {
    _valid[i] = false;
    if (!e) {
      iterinfo info (*this, 0, LOGERR_NORECORD, !_comparable, true, true);
      _cb (&info);
      if (info.finished == true)
	return false;
      else
	return true;
    }
    else {
      iterinfo info (*this, e, LOGERR_CORRUPTLOG, !_comparable, true, true);
      _cb (&info);
      return false;
    }
  }
  if (olde && !olde->template record<head_log_record> () &&
      (olde->version ()).conflicts (e->version())) {
    iterinfo info (*this, e, LOGERR_CONFLICT, !_comparable, true, true);
    _cb (&info);
    return false;
  }
  if (!_stop_at.empty () && e->seqnum () <= _stop_at[i])
    _valid[i] = false;
  return true;
}

void
logiter::check_ordering ()
{
  // if there are no ordering between any two records, they are
  // concurrent updates due to concurrent operations, or stale log
  // head being used, or partition

  _comparable = true;
  for (unsigned i=0; i<_entries.size(); i++) {
    if (_valid[i] && !_entries[i]->template record<end_log_record>()) {
      for (unsigned j=0; j<_entries.size(); j++) {
	if (_valid[j] && !_entries[j]->template record<end_log_record>() &&
	    i != j) {
          if ((_entries[i]->version ()).equals(_entries[j]->version ()) ||
	      (_entries[i]->version ()).conflicts(_entries[j]->version ())) {
	    warn << "Concurrent version vectors, between log records in "
	         << _entries[i]->head () << " and "
		 << _entries[j]->head () << "\n";
	    _comparable = false;
	    return;
	  }
	}
      }
    }
  }
}

void
logiter::got_version (vervec vv, log_status status)
{
  if (status) {
    iterinfo i (*this, 0, status, false, true, true);
    _cb (&i);
    return;
  }
  
  if (vv.size () == 0) {
    iterinfo i (*this, 0, LOG_OK, false, true, true);
    _cb (&i);
    return;
  }

  _start = vv;
  _nloaded = 0;
  _load_err = 0;
  for (unsigned i=0; i < vv.size (); i++) {
    _valid.push_back (false);
    _entries.push_back ((recordp)0);
    _id.push_back (vv.latest (i));
    if (!_stop.empty()) {
      bool found = false;
      for (unsigned j=0; j < _stop.size() && !found; j++) {
        if (_stop.head(j) == vv.head (i)) {
	  _stop_at.push_back (_stop.seqnum(j));
	  found = true;
        }
      }
      if (!found)
	_stop_at.push_back(0);
    }
    ns::retrieve_record (vv.latest (i),
	                 wrap(mkref(this), &logiter::load_record, i));
  }
}

void
logiter::load_record (unsigned i, recordp e)
{
  _nloaded++;
  if (!check_and_load_record(i, e)) {
    _load_err = 1;
    return;
  }
  if (_nloaded == _start.size () && !_load_err)
    iter ();
}

void
logiter::got_view (log_status status)
{
  if (status) {
    iterinfo i (*this, 0, status, false, true, true);
    _cb (&i);
    return;
  }
  _view->version (wrap (mkref(this), &logiter::got_version));
}

logiter::logiter (cbliter cb, vervec vv)
  : _cb(cb)
{
  _comparable = true;
  _outoforder = false;
  _start = vv;
  timecb (tsnow, wrap (mkref(this), &logiter::got_version, vv, LOG_OK));
}

logiter::logiter (cbliter cb, vervec vv, vervec stop)
  : _cb(cb)
{
  _stop = stop;
  _comparable = true;
  _outoforder = false;
  _start = vv;
  timecb (tsnow, wrap (mkref(this), &logiter::got_version, vv, LOG_OK));
}

logiter::logiter (cbliter cb, bigint view_addr, vervec stop)
  : _cb(cb)
{
  _stop = stop;
  _comparable = true;
  _outoforder = false;
  _view = New refcounted<view>
    (wrap (mkref(this), &logiter::got_view), view_addr);
}

logiter::logiter (cbliter cb, bigint view_addr)
  : _cb(cb)
{
  _comparable = true;
  _outoforder = false;
  _view = New refcounted<view>
    (wrap (mkref(this), &logiter::got_view), view_addr);
}

static void
prefetch_record (int n, ptr<log_record> e)
{
  if (n < 4 && e && e->prev () != 0 && !ns::has_record (e->prev ()))
    ns::retrieve_record (e->prev (), wrap(prefetch_record, n+1));
}

void
logiter::iter_cb (unsigned vi, ptr<log_record> e)
{
  if (!check_and_load_record(vi, e))
    return;
  // prefetch log record into record cache
  if (e->prev () != 0 && !ns::has_record (e->prev ()))
    ns::retrieve_record (e->prev (), wrap(prefetch_record, 1));
  iter ();
}

void
logiter::iter ()
{
  bool got_hi = false;
  unsigned int hi;

  check_ordering ();
  
  // iterinfo (logiter &i, recordp e, log_status status,
  //           bool concurrent, bool finished, bool eol)

  // view block, version vector, and the _entries vector here are both
  // sorted by log head id, from smallest to largest.  we iterate the
  // logs in reverse order, from largest to smallest instead (see L*
  // paper).

  for (unsigned i = _entries.size (); i > 0;) {
    i--;
    if (_valid[i] &&
	(got_hi == false ||
	 _entries[i]->version().gt (_entries[hi]->version ()))) {
      hi = i;
      got_hi = true;
    }
  }

  if (got_hi == false) { // didn't find anything
    iterinfo i (*this, 0, LOG_OK, false, true, true);
    _cb (&i);
    return;
  }

  // now check if there is a log record from a log marked invalid that
  // can be ordered ahead of the chosen log record. if so, set
  // _outoforder to true.

  for (unsigned i = _entries.size (); i > 0;) {
    i--;
    if (!_valid[i] && _entries [i] &&
	(!_entries[i]->template record<end_log_record> ()) &&
	(!_entries[hi]->template record<end_log_record> ())) {
      if (_entries[i]->version().gt (_entries[hi]->version ())) {
	warn << "out of order iteration of log records\n";
	_outoforder = true;
	break;
      }
    }
  }
 
  iterinfo i (*this, _entries[hi], 
              _outoforder ? LOGERR_OUTOFORDER : LOG_OK,
	      !_comparable, false, false);
  _cb (&i);
  if (i.finished)
    return;

  if (i.eol || _entries[hi]->template record<end_log_record> ()) {
    _valid[hi] = false;
    iter ();
  }
  else {
    _id[hi] = _entries[hi]->prev ();
    ns::retrieve_record 
      (_entries[hi]->prev (), wrap(mkref(this), &logiter::iter_cb, hi));
  }
}

