/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef LOGITER_H
#define LOGITER_H

#include "definc.h"
#include "vec.h"
#include "record.h"
#include "vervec.h"

class logiter;
class iterinfo;

typedef callback<void, iterinfo *>::ref cbliter;

// logiter combines log records from per-user logs into one virtual
// log. logiter constructors take a callback function, a view id or a
// view_instance object. the callback is called for every log record,
// in sequence, from the most recent to the least recent.
//
// logiter passes to each callback a pointer to an iterinfo object.
// the iterinfo object contains a reference to the logiter object, a
// pointer to the current log record, status, and several flags.
// logiter frees the iterinfo object when the callback returns. thus,
// DO NOT pass around the iterinfo object.
// 
// the finish flag is set to true if the logiter is done. if the
// callback sets the finish flag to be true, upon return from the
// callback, logiter terminates.
//
// the eol flag is set to true if the logiter is done loading and
// processing log records for the log that the current log record is a
// part of. if the callback sets the eol flag to be true, upon return
// from the callback, logiter will no longer load and process log
// records from that log.
//
// the concurrent flag is set to true by logiter if any two of the log
// records it currently keeps in memory have concurrent version
// vectors. logiter keeps one log record from each log at any time.
//
// the status field reports errors. see LI_ERR definitions below. if
// logiter returns LI_ERR_NREC to the callback, the callback can elect
// to contine to scan other logs by setting the finish flag in the
// iterinfo object to false. subsequently, logiter ignores the
// corrupted log. by default, upon LI_ERR_NREC, iterinfo.finished is
// set to true.

class view;

class logiter : public virtual refcount { 
private:
  cbliter _cb;
  vervec _start;
  vervec _stop;
  ptr<view> _view;
  unsigned _nloaded;
  int _load_err;

  vec<ptr<log_record> > _entries;
  vec<bool> _valid;
  vec<bigint> _id;
  vec<uint> _stop_at;
  bool _comparable;
  bool _outoforder;

  void got_version (vervec vv, log_status status);
  void got_view (log_status status);
  void iter_cb (unsigned vi, recordp e);
  void iter ();

  void load_record (unsigned i, recordp e);
  bool check_and_load_record (unsigned i, recordp e);
  void check_ordering ();

public:
  logiter (cbliter cb, vervec vv);
  logiter (cbliter cb, vervec vv, vervec stop);
  logiter (cbliter cb, bigint view_addr);
  logiter (cbliter cb, bigint view_addr, vervec stop);
  ~logiter () {}

  vervec version () const { return _start; }
  unsigned nlogs () const { return _start.size (); }
};

class iterinfo {
public:
  logiter& iter;
  recordp record;
  log_status status;
  bool concurrent;
  bool finished;
  bool eol;
  iterinfo(logiter &i, recordp e, log_status s, bool c, bool f, bool l)
    : iter (i), record (e), status (s), concurrent (c),
      finished (f), eol (l) {}
  iterinfo(const iterinfo& i)
    : iter (i.iter), record (i.record), status (i.status),
      concurrent (i.concurrent), finished (i.finished), eol (i.eol) {}
};

#endif // __CFSRW_LOG_H

