/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef RECORD_H
#define RECORD_H

#include "definc.h"
#include "vervec.h"
#include "pack.h"

class head_log_record;
class end_log_record;

class log_record 
{
private:
  bigint   _prev;
  bigint   _head;
  uint     _seqnum;
  vervec   _version;
  int      _record_type;
  unsigned _ts;

protected:
  void serialize_meta (strbuf &sb) const;
  virtual const std::type_info& cxxtype () const { return typeid(*this); }
  
  friend class log;
  void set_time ();

public:
  log_record (int type = 0);
  virtual ~log_record () {}
  
  int record_type ()       const { return _record_type; }
  virtual str serialize () const = 0;
  unsigned timestamp ()    const { return _ts; }

  bigint hash () const;
  bigint prev () const { return _prev; }
  bigint head () const { return _head; }
  uint seqnum () const { return _seqnum; }
  vervec version () const { return _version; }
  vervec version_inclusive () const {
    vervec v = _version;
    for (unsigned i=0; i<v.size (); i++)
      if (v.head (i) == head ()) {
	v.add (v.head (i), hash (), seqnum ());
	break;
      }
    return v;
  }
  
  void set_prev (bigint id) { _prev = id; }
  void set_head (bigint id) { _head = id; }
  void set_seqnum (uint n)  { _seqnum = n; }
  void set_version (const vervec &v) { _version = v; }

  template<class T> T* record () {
    if (typeid (T) != cxxtype () &&
	// not sure why cxxtype() can return a refcounted type, since
	// supposedly the this pointer is just a plain log_record ptr.
	typeid (refcounted<T>) != cxxtype ())
      return 0;
    return static_cast<T*> (this);
  }

  static ptr<log_record> dup (ptr<log_record>);
  static ptr<log_record> parse (str);

  static const int T_END  = 0;
  static const int T_HEAD = 1;
};

class end_log_record : public log_record
{
public:
  end_log_record ();
  end_log_record (const end_log_record& r) { *this = r; }
  end_log_record (xin *);
  str serialize () const;
};

class head_log_record : public log_record
{
private:
  bigint _root;
  bigint _snapshot;
  
public:
  head_log_record ();
  head_log_record (const head_log_record& h) { *this = h; }
  head_log_record (bigint root);
  head_log_record (xin *);

  bigint root () const { return _root; }
  bigint snapshot () const { return _snapshot; }
  void snapshot (bigint s) { _snapshot = s; }
  str serialize() const;
};

typedef ptr<log_record> recordp;
typedef callback<void, recordp>::ref cb_recordp;

#endif

