/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef VERVEC_H
#define VERVEC_H

#include "definc.h"
#include "pack.h"

class vervec
{
private:
  vec<bigint> _head;
  vec<bigint> _latest;
  vec<uint> _seqnum;

public:
  vervec() { clear(); }
  void clear () { _head.clear(); _latest.clear(); _seqnum.clear(); }
  bool empty () const { return _head.size() == 0; }

  bigint head (int i) const { return _head[i]; }
  bigint latest (int i) const { return _latest[i]; }
  uint seqnum (int i) const { return _seqnum[i]; }
  unsigned size () const { return _head.size(); }
  
  void serialize (strbuf &sb) const {
    sint(sb, _head.size());
    for (unsigned i=0; i<_head.size(); i++) {
      sbigint (sb, _head[i]);
      sbigint (sb, _latest[i]);
      sint (sb, _seqnum[i]);
    }
  }

  bool ge (const vervec& v) const {
    return (equals(v) || gt(v));
  }

  // v0 > v1 iff
  // (forall i, v0[i] >= v1[i]) & (exists j, v0[j] > v1[j])
  bool gt (const vervec& v) const {
    bool bigger = false;
    unsigned i;
    unsigned j;
    for (i=0, j=0; i<_head.size();) {
      if (j < v._head.size()) {
	if (_head[i] == v._head[j]) { // user in both vectors
	  if (_seqnum[i] < v._seqnum[j])
	    return false;
	  if (_seqnum[i] > v._seqnum[j])
	    bigger = true;
          i++;
	  j++;
	}
	else if (_head[i] < v._head[j]) { // user in my vector, but not hers
	  if (_seqnum[i] > 0)
	    bigger = true;
	  i++;
	}
	else { // user in her vector, but not mine
	  if (v._seqnum[j] > 0)
	    return false;
	  j++;
	}
      }
      else { // rest of users in _head are in my vector, but not hers
	if (_seqnum[i] > 0) {
	  bigger = true;
	  break;
	}
	i++;
      }
    }
    while (j < v._head.size()) { // rest of users in hers, but not mine
      if (v._seqnum[j] > 0)
	return false;
      j++;
    }
    return bigger;
  }

  // v0 conflicts with v1 iff 
  // (exists i, v0[i] > v1[i]) & (exists j, v0[j] < v1[j])
  bool conflicts (const vervec& v) const {
    bool bigger = false;
    bool smaller = false;
    unsigned i;
    unsigned j;
    for (i=0, j=0; i<_head.size() && !(bigger && smaller);) {
      if (j < v._head.size()) {
	if (_head[i] == v._head[j]) { // user in both vectors
	  if (_seqnum[i] < v._seqnum[j])
	    smaller = true;
	  else if (_seqnum[i] > v._seqnum[j])
	    bigger = true;
          i++;
	  j++;
	}
	else if (_head[i] < v._head[j]) { // user in my vector, but not hers
	  if (_seqnum[i] > 0)
	    bigger = true;
	  i++;
	}
	else { // user in her vector, but not mine
	  if (v._seqnum[j] > 0)
	    smaller = true;
	  j++;
	}
      }
      else { // rest of users in _head are in my vector, but not hers
	if (_seqnum[i] > 0) {
	  bigger = true;
	  break;
	}
	i++;
      }
    }
    while (j < v._head.size()) { // rest of users in hers, but not mine
      if (v._seqnum[j] > 0) {
	smaller = true;
	break;
      }
      j++;
    }
    if (bigger && smaller)
      return true;
    return false;
  }

  // v0 equals v1 iff (forall i, v0[i] == v1[i])
  bool equals (const vervec& v) const {
    unsigned i;
    unsigned j;
    for (i=0, j=0; i<_head.size();) {
      if (j < v._head.size()) {
        if (_head[i] == v._head[j]) { // both vector contain same user
	  if (_seqnum[i] != v._seqnum[j])
	    return false;
	  j++;
	  i++;
        }
        else if (_head[i] < v._head[j]) { // user in mine, but not hers
          if (_seqnum[i] != 0)
	    return false;
	  i++;
        }
        else { // in hers, but not in mine
	  if (v._seqnum[j] != 0)
	    return false;
          j++;
        }
      }
      else { // in mine, not hers
	if (_seqnum[i] != 0)
	  return false;
	i++;
      }
    }
    while (j < v._head.size()) { // rest of users in hers, but not mine
      if (v._seqnum[j] != 0)
	return false;
      j++;
    }
    return true;
  }

  // v0 equals v1 iff (forall i, (v0[i] == v1[i] || i == h))
  bool equals_except (const vervec& v, bigint h) const {
    unsigned i;
    unsigned j;
    for (i=0, j=0; i<_head.size();) {
      if (j < v._head.size()) {
        if (_head[i] == v._head[j]) { // both vector contain same user
	  if (_seqnum[i] != v._seqnum[j] && _head[i] != h)
	    return false;
	  j++;
	  i++;
        }
        else if (_head[i] < v._head[j]) { // user in mine, but not hers
          if (_seqnum[i] != 0 && _head[i] != h)
	    return false;
	  i++;
        }
        else { // in hers, but not in mine
	  if (v._seqnum[j] != 0 && v._head[j] != h)
	    return false;
          j++;
        }
      }
      else { // in mine, not hers
	if (_seqnum[i] != 0 && _head[i] != h)
	  return false;
	i++;
      }
    }
    while (j < v._head.size()) { // rest of users in hers, but not mine
      if (v._seqnum[j] != 0 && v._head[j] != h)
	return false;
      j++;
    }
    return true;
  }

  vervec &operator= (const vervec &v) {
    clear();
    for (unsigned i=0; i<v._head.size(); i++) {
      _head.push_back(v._head[i]);
      _latest.push_back(v._latest[i]);
      _seqnum.push_back(v._seqnum[i]);
    }
    return *this;
  }

  void add (bigint h, bigint l, uint seq)
  {
    // make sure heads are in the same order as they are in the view:
    // which is from smallest to largest as defined by the > operator
    // on bigint
    bool inorder = true;
    unsigned i = 0;
    for (i=0; i<_head.size(); i++) {
      if (_head[i] == h) {
	_latest[i] = l;
	_seqnum[i] = seq;
	return;
      }
      else if (_head[i] > h) {
        inorder = false;
	break;
      }
    }
    if (inorder) {
      _head.push_back(h);
      _latest.push_back(l);
      _seqnum.push_back(seq);
    }
    else {
      warn << "out of order vector construction, please tell benjie\n";
      vec<bigint> nh;
      vec<bigint> nl;
      vec<uint> ns;
      bool added = false;
      for(unsigned j=0; j<_head.size(); j++) {
	if (!added && _head[j] > h) {
          added = true;
	  nh.push_back(h);
	  nl.push_back(l);
	  ns.push_back(seq);
	}
	nh.push_back(_head[j]);
	nl.push_back(_latest[j]);
	ns.push_back(_seqnum[j]);
      }
      _head = nh;
      _latest = nl;
      _seqnum = ns;
    }
  }

  static vervec merge_smaller (vervec &v1, vervec &v2) {
    vervec v;
    for (unsigned i=0; i<v1._head.size(); i++) {
      v._head.push_back(v1._head[i]);
      v._latest.push_back(v1._latest[i]);
      v._seqnum.push_back(v1._seqnum[i]);
    }
    for (unsigned i=0; i<v2._head.size(); i++) {
      bool skip = false;
      for (unsigned j=0; j<v1._head.size(); j++) {
	if (v1._head[j] == v2._head[i]) {
	  skip = true;
          if (v1._seqnum[j] > v2._seqnum[i]) {
	    for (unsigned k=0; k<v._head.size(); k++) {
	      if (v._head[k] == v1._head[j]) {
                v._latest[k] = v2._latest[i];
                v._seqnum[k] = v2._seqnum[i];
		break;
	      }
	    }
	  }
	  break;
	}
      }
      if (!skip) {
        v._head.push_back(v2._head[i]);
        v._latest.push_back(v2._latest[i]);
        v._seqnum.push_back(v2._seqnum[i]);
      }
    }
    return v;
  }

  static int parse (str s, vervec &version)
  {
    xin in (s);
    int r = parse (&in, version);
    return r;
  }
  
  static int parse (xin *in, vervec &version)
  {
    int vsize;
    version.clear ();
    if (!in->rint (vsize))
      return -1;
    for (int i=0; i<vsize; i++) {
      bigint h;
      bigint l;
      int s;
      if (!in->rbigint (h))
	return -1;
      if (!in->rbigint (l))
	return -1;
      if (!in->rint (s))
	return -1;
      version.add (h, l, s);
    }
    return 0;
  }

};

#endif

