/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "definc.h"
#include "ns.h"
#include "record.h"
#include "log.h"
#include "view.h"

// view constructor

view_constructor::view_constructor (const view &v)
{
  root (v.root ());
  for (unsigned i = 0; i < v.nlogs (); i++)
    add_log (v.head (i), v.latest (i));
}

bigint
view_constructor::store (callback<void, ns_stat, bigint>::ref cb)
{
  str data = serialize ();
  bigint id = compute_sha1_hash (data.cstr (), data.len ());
  ns::insert_block (data.cstr (), data.len (), cb);
  return id;
}

void
view_constructor::sort ()
  // slow but simple sort
{
  vec<bigint> h;
  vec<bigint> l;
  vec<bool> done;
  bool found = true;
  for (unsigned i=0; i<_head.size(); i++)
    done.push_back(false);

  while (found) {
    found = false;
    bigint sv;
    unsigned s = 0;
    for (unsigned i=0; i<_head.size(); i++) {
      if (!done[i]) {
	if (!found || _head[i] < sv) {
          found = true;
	  sv = _head[i];
	  s = i;
	}
      }
    }
    if (found) {
      h.push_back(_head[s]);
      l.push_back(_latest[s]);
      done[s] = true;
    }
  }

  _head = h;
  _latest = l;
}

str
view_constructor::serialize ()
{
  sort (); // paranoia
  strbuf sb;
  sbigint (sb, _root);
  sint (sb, _head.size ());
  for (unsigned i=0; i<_head.size (); i++) {
    sbigint (sb, _head[i]);
    sbigint (sb, _latest[i]);
  }
  return sb;
}

void
view_constructor::add_log (bigint h, bigint b)
{
  for (unsigned i=0; i<_head.size (); i++) {
    if (_head[i] == h) {
      _latest[i] = b;
      return;
    }
  }
  _head.push_back (h);
  _latest.push_back (b);
  sort ();
}

void
view_constructor::parse (xin *in)
{
  int n;
  in->rbigint (_root);
  in->rint (n);
  for (int i=0; i<n; i++) {
    bigint h, b;
    in->rbigint (h);
    in->rbigint (b);
    add_log (h, b);
  }
}

// view object

view::view (callback<void, log_status>::ref cb, bigint addr)
{
  ptr<view_constructor> vc = New refcounted<view_constructor>;
  ns::retrieve_block (addr, wrap(this, &view::got_view, vc, cb));
}

view::view (ptr<view_constructor> c)
{
  assert (c);
  take (c);
}

void
view::take (ptr<view_constructor> c)
{
  assert (c);
  _root = c->_root;
  for (unsigned i = 0; i < c->_head.size (); i++) {
    _head.push_back (c->_head[i]);
    _latest.push_back (c->_latest[i]);
  }
}

void
view::got_view (ptr<view_constructor> vc, 
                callback<void, log_status>::ref cb, ptr<ns_block> blk)
{
  if (blk == 0) {
    warn << "could not fetch view block\n";
    cb (LOGERR_NOVIEW);
    return;
  }

  str ins (blk->data, blk->len);
  xin *in = New xin (ins);
  vc->parse (in);
  delete in;
  take (vc);
  cb (LOG_OK);
}

void
view::version (callback<void, vervec, log_status>::ref cb)
{
  ptr<view_state> v = New refcounted<view_state> ();
  for (unsigned i=0; i<_head.size (); i++) {
    v->latest.push_back (_latest[i]);
    v->seqnum.push_back (0);
    if (_head[i] == _latest[i])
      // need to load loghead first
      ns::retrieve_loghead
	(_head[i], wrap(this, &view::load_record, true, v, i, cb));
    else
      ns::retrieve_record
	(_latest[i], wrap(this, &view::load_record, false, v, i, cb));
  }
  load_done (v, cb);
}

void
view::load_record (bool ishead, ptr<view_state> v, unsigned i,
                   callback<void, vervec, log_status>::ref cb, recordp e)
{
  if (!e) {
    v->nrecords++;
    v->error = ishead ? LOGERR_NOLOGHEAD : LOGERR_NORECORD;
  }
  else if (ishead) {
    if (e->template record<head_log_record>()) {
      v->latest[i] = e->prev ();
      ns::retrieve_record
	(v->latest[i], wrap(this, &view::load_record, false, v, i, cb));
    }
    else {
      v->nrecords++;
      v->error = LOGERR_NOTLOGHEAD;
    }
  }
  else {
    v->nrecords++;
    v->seqnum[i] = e->seqnum ();
  }
  load_done (v, cb);
}

void
view::load_done (ptr<view_state> v,
                 callback<void, vervec, log_status>::ref cb)
{
  assert (v->nrecords <= _head.size ());
  if (v->nrecords != _head.size ())
    return;

  vervec vv;
  if (!v->error) {
    for (unsigned i=0; i<_head.size(); i++)
      vv.add (_head[i], v->latest[i], v->seqnum[i]);
  }
  cb (vv, v->error);
}

