/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "pastwatch.h"

extern void write_file (str, bigint, callback<void, cfsrw_status>::ref);

static void
sattr_reply (str path, fhcb_t fhcb, bigint fh,
             cfscallp c, cfsrw_status status)
{
  pastwatch_status ("N", path);

  if (status) {
    warn << "cannot set file attribute\n";
    fhcb (fh, -1);
  }
  else
    fhcb (fh, 0);
}

static void
write_file_done (str path, time_t mt, fhcb_t fhcb, bigint fh,
                 cfsrw_status err)
{
  if (err) {
    warn << "cannot write file into repository\n";
    fhcb (fh, -1);
    return;
  }

  cfscallp c = New refcounted<cfscall>
    (CFSRWPROC_SETATTR, wrap (sattr_reply, path, fhcb, fh));
  cfsrw_setattr_arg *arg = static_cast<cfsrw_setattr_arg *> (c->arg ());
  arg->object = fh;
  arg->new_attributes.valid_mtime = 1;
  arg->new_attributes.mtime = mt;
  cfsrw_call (c);
}

static void
create_reply (str path, time_t mt, fhcb_t fhcb,
              cfscallp c, cfsrw_status status)
{
  if (status) {
    bigint blah;
    fhcb (blah, -1);
    return;
  }

  cfsrw_create_res *res = static_cast<cfsrw_create_res *> (c->res ());
  write_file (path, res->object,
              wrap (write_file_done, path, mt, fhcb, res->object));
}

static void
mkdir_reply (str path, time_t mt, fhcb_t fhcb,
             cfscallp c, cfsrw_status status)
{
  if (status) {
    bigint blah;
    fhcb (blah, -1);
    return;
  }
  
  cfsrw_create_res *res = static_cast<cfsrw_create_res *> (c->res ());
  cfscallp cc = New refcounted<cfscall>
    (CFSRWPROC_SETATTR, wrap (sattr_reply, path, fhcb, res->object));
  cfsrw_setattr_arg *arg = static_cast<cfsrw_setattr_arg *> (cc->arg ());
  arg->object = res->object;
  arg->new_attributes.valid_mtime = 1;
  arg->new_attributes.mtime = mt;
  cfsrw_call (cc);
}

static void
symlink_reply (str path, time_t mt, fhcb_t fhcb,
               cfscallp c, cfsrw_status status)
{
  if (status) {
    bigint blah;
    fhcb (blah, -1);
    return;
  }

  cfsrw_symlink_res *res = static_cast<cfsrw_symlink_res *> (c->res ());
  cfscallp cc = New refcounted<cfscall>
    (CFSRWPROC_SETATTR, wrap (sattr_reply, path, fhcb, res->object));
  cfsrw_setattr_arg *arg = static_cast<cfsrw_setattr_arg *> (cc->arg ());
  arg->object = res->object;
  arg->new_attributes.valid_mtime = 1;
  arg->new_attributes.mtime = mt;
  cfsrw_call (cc);
}

void
checkin (const pde_t& e, fhcb_t fhcb)
{
  struct stat sb;
  
  if (int err = lstat (e.path, &sb)) {
    perror ("lstat");
    bigint blah;
    fhcb (blah, err);
    return;
  }

  switch (e.type) {
    case CFSRW_DIR:
      {
	cfscallp c = New refcounted<cfscall>
	  (CFSRWPROC_MKDIR, wrap (mkdir_reply, e.path, sb.st_mtime, fhcb));
	cfsrw_create_arg *arg = static_cast<cfsrw_create_arg *> (c->arg ());
	arg->dir = e.dirfh;
	arg->name = e.name;
	arg->attributes.valid_size = 1;
	arg->attributes.size = sb.st_size;
	arg->attributes.valid_mode = 1;
	arg->attributes.mode = sb.st_mode;
	arg->attributes.valid_mtime = 1;
	arg->attributes.mtime = sb.st_mtime;
	cfsrw_call (c);
      }
      break;
    case CFSRW_REG:
      {
	cfscallp c = New refcounted<cfscall>
	  (CFSRWPROC_CREATE, wrap (create_reply, e.path, sb.st_mtime, fhcb));
	cfsrw_create_arg *arg = static_cast<cfsrw_create_arg *> (c->arg ());
	arg->dir = e.dirfh;
	arg->name = e.name;
	arg->attributes.valid_size = 1;
	arg->attributes.size = sb.st_size;
	arg->attributes.valid_mode = 1;
	arg->attributes.mode = sb.st_mode;
	arg->attributes.valid_mtime = 1;
	arg->attributes.mtime = sb.st_mtime;
	cfsrw_call (c);
      }
      break;
    case CFSRW_LNK:
      {
	char buf[256];
	int r;
	if ((r = readlink (e.path, buf, 256)) < 0) {
          perror ("readlink");
	  bigint blah;
          fhcb (blah, r);
	  return;
	}
	str link (buf, r);

	cfscallp c = New refcounted<cfscall>
	  (CFSRWPROC_SYMLINK, wrap (symlink_reply, e.path, sb.st_mtime, fhcb));
	cfsrw_symlink_arg *arg = static_cast<cfsrw_symlink_arg *> (c->arg ());
	arg->dir = e.dirfh;
	arg->name = e.name;
	arg->linkto = link;
	arg->attributes.valid_size = 1;
	arg->attributes.size = sb.st_size;
	arg->attributes.valid_mode = 1;
	arg->attributes.mode = sb.st_mode;
	arg->attributes.valid_mtime = 1;
	arg->attributes.mtime = sb.st_mtime;
	cfsrw_call (c);
      }
      break;
    default:
      assert (0);
  }
}

