/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include "pastwatch.h"
#include "vdb.h"
#include "cfsrw_wrapper.h"

extern void
read_file (int fd, bigint fh, callback<void, cfsrw_status>::ref cb);

class checkout_object : public virtual refcount
{
private:
  struct checkout_dirent_t {
    bigint fh;
    str path;
    str cfspath;
  };
  typedef ptr<checkout_dirent_t> cdep_t;

  int work;
  int error;

  void
  read_reply (cdep_t cde, vervec v, int fd, cbi cb, cfsrw_status status)
  {
    work--;
  
    if (status) {
      error = -1;
      if (status) {
        close (fd);
        warn << "cannot download " << cde->path << "\n";
      }
      if (work == 0)
        cb (error);
      return;
    }
    
    close (fd);
    pastwatch_status ("U", cde->path);
    vdb_schedule_add (cde->path, pfe_t (cde->fh, v));
  
    if (work == 0)
      cb (error);
  }

  void
  readlink_reply (cdep_t cde, vervec v, cbi cb, cfscallp c,
                  cfsrw_status status)
  {
    work--;
    if (status) {
      error = -1;
      warn << "cannot download " << cde->path << "\n";
      if (work == 0)
        cb (error);
      return;
    }
    cfsrw_readlink_arg *arg = static_cast<cfsrw_readlink_arg *> (c->arg ());
    cfsrw_readlink_res *res = static_cast<cfsrw_readlink_res *> (c->res ());
    
    int r = symlink (res->path.cstr (), cde->path);
    if (r < 0) {
      warn << "cannot create symbolic link " << cde->path << "\n";
      error = -1;
    }
    else {
      pastwatch_status ("S", cde->path);
      vdb_schedule_add (cde->path, pfe_t (arg->file, v));
    }
  
    if (work == 0)
      cb (error);
  }
  
  void
  getattr_reply (cdep_t cde, cbi cb,
                 cfscallp c, cfsrw_status status)
  {
    work--;
    if (status) {
      error = -1;
      warn << "cannot get attribute for " << cde->path << "\n";
      if (work == 0)
        cb (error);
      return;
    }
    cfsrw_getattr_arg *arg = static_cast<cfsrw_getattr_arg *> (c->arg ());
    cfsrw_getattr_res *res = static_cast<cfsrw_getattr_res *> (c->res ());
  
    if (res->attributes.type == CFSRW_REG) {
      int fd = open
        (cde->path, O_CREAT | O_TRUNC | O_WRONLY, res->attributes.mode);
      if (fd < 0) {
        error = -1;
        warn << "cannot open " << cde->path << " for writing\n";
        if (work == 0)
          cb (error);
        return;
      }
      work++;
      read_file (fd, cde->fh,
	         wrap (mkref(this), &checkout_object::read_reply,
		       cde, res->attributes.mver, fd, cb));
    }
  
    else if (res->attributes.type == CFSRW_LNK) {
      work++;
      cfsrw_readlink
        (cde->fh, wrap (mkref(this), &checkout_object::readlink_reply,
			cde, res->attributes.mver, cb));
    }
  
    else if (res->attributes.type == CFSRW_DIR) {
      if (mkdir (cde->path, res->attributes.mode) < 0) {
        error = -1;
        warn << "cannot create directory " << cde->path << "\n";
        if (work == 0)
          cb (error);
        return;
      }
  
      vdb_schedule_mkdir (cde->path, cde->cfspath);
  
      work++;
      cfsrw_readdir
	(cde->fh, 0, wrap (mkref(this), &checkout_object::readdir_reply,
			   cde, cb));
      
      pastwatch_status ("D", cde->path);
      vdb_schedule_add (cde->path, pfe_t (arg->object, res->attributes.mver));
    }
  
    if (work == 0)
      cb (error);
  }
  
  void
  readdir_reply (cdep_t cde, cbi cb,
                 cfscallp c, cfsrw_status status)
  {
    work--;
    if (status) {
      error = -1;
      warn << "cannot checkout directory " << cde->path << "\n";
      if (work == 0)
        cb (error);
      return;
    }
  
    cfsrw_readdir_res *res = static_cast<cfsrw_readdir_res *> (c->res ());
    rpc_ptr<cfs_rpc_dirent> de = res->reply.entries;
    uint64 cookie = 0;
  
    while (de) {
      strbuf newpath;
      newpath << cde->path << "/" << de->name;
      
      cdep_t new_cde = New refcounted<checkout_dirent_t>;
      new_cde->path = newpath;
      new_cde->cfspath = make_cfspath (cde->cfspath, de->name);
      new_cde->fh = de->fh;
  
      work++;
      cfsrw_getattr 
	(de->fh, wrap (mkref(this), &checkout_object::getattr_reply,
		       new_cde, cb));
      cookie = de->cookie;
      de = de->nextentry;
    }
  
    if (!res->reply.eof) {
      work++;
      cfsrw_readdir
        (cde->fh, cookie,
	 wrap (mkref(this), &checkout_object::readdir_reply, cde, cb));
    }
    else if (work == 0)
      cb (error);
  }
 
public:
  checkout_object (bigint root, str path, str cfspath, cbi cb) 
    : work (0), error (0)
  {
    cdep_t cde = New refcounted<checkout_dirent_t>;
    cde->fh = root;
    cde->path = path;
    cde->cfspath = cfspath;
    work++;
    cfsrw_getattr
      (root, wrap (mkref(this), &checkout_object::getattr_reply, cde, cb));
  }

};

void
checkout (bigint root, str path, str cfspath, cbi cb)
{
  ptr<checkout_object> a =
    New refcounted<checkout_object> (root, path, cfspath, cb);
}

