/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <errno.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "pastwatch.h"
#include "update.h"
#include "vdb.h"
#include "cfsrw_wrapper.h"

extern void update (bigint root, str path, cbi cb);
extern void write_file (str, bigint, callback<void, cfsrw_status>::ref);
extern void checkin (const pde_t& e, fhcb_t fhcb);
extern int remove_dirops (vec<str> &);
extern bool arg_debugmode;

class commit_object : public virtual refcount
{
  bigint root;
  int work;
  int error;
  ptr<update_object_spec> up;
  vec<str> commited;

  void done (cbi cb, int error)
  {
    remove_dirops (commited);
    cb (error);
  }

  void
  commit_getattr_reply (unsigned int idx, cbi cb,
                        cfscallp c, cfsrw_status status)
  {
    work--;
    if (status) {
      warn << "cannot get file attribute from repository after commit\n";
      error = -1;
      if (work == 0)
        done (cb, error);
      return;
    }
    
    cfsrw_getattr_arg *arg = static_cast<cfsrw_getattr_arg *> (c->arg ());
    cfsrw_getattr_res *res = static_cast<cfsrw_getattr_res *> (c->res ());
    pde_t *de = &up->to_commit [idx];
    vdb_schedule_add (de->path, pfe_t (arg->object, res->attributes.mver));

    if (work == 0)
      done (cb, error);
  }

  void
  getattr (bigint fh, cfsrw_cb cb)
  {
    cfscallp c = New refcounted<cfscall> (CFSRWPROC_GETATTR, cb);
    cfsrw_getattr_arg *arg = static_cast<cfsrw_getattr_arg *> (c->arg ());
    arg->object = fh;
    // this could be an expensive cfsrw_call; there is no snapshot yet
    // that includes log records we just appended, so this call may
    // involve scanning a bunch of log records.
    cfsrw_call_latest (c);
  }
  
  void
  write_reply (unsigned int idx, cbi cb, cfsrw_status status)
  {
    work--;
    if (status) {
      warn << "cannot write file to repository\n";
      error = -1;
      if (work == 0)
        done (cb, error);
      return;
    }
      
    pde_t *de = &up->to_commit [idx];
    pastwatch_status ("Committing", de->path);
  
    work++;
    getattr
      (de->fh,
       wrap (mkref(this), &commit_object::commit_getattr_reply, idx, cb));
  }
  
  void
  symlink_reply (unsigned int idx, cbi cb, cfscallp c, cfsrw_status status)
  {
    work--;
    if (status) {
      warn << "cannot create symlink in repository\n";
      error = -1;
      if (work == 0)
        done (cb, error);
      return;
    }
      
    cfsrw_symlink_res *res = static_cast<cfsrw_symlink_res *> (c->res ());
    work++;
    getattr
      (res->object,
       wrap (mkref(this), &commit_object::commit_getattr_reply, idx, cb));
  }
  
  void
  sattr_reply (unsigned int idx, cbi cb, cfscallp c, cfsrw_status status)
  {
    work--;
    if (status) {
      warn << "cannot set file attribute in repository for commit\n";
      error = -1;
      if (work == 0)
        done (cb, error);
      return;
    }
  
    pde_t *de = &up->to_commit [idx];
    work++;
    write_file
      (de->path, de->fh,
       wrap (mkref(this), &commit_object::write_reply, idx, cb));
  }
  
  void
  unlink_reply (unsigned int idx, cbi cb, cfscallp cc, cfsrw_status status)
  {
    work--;
    if (status) {
      warn << "cannot unlink symlink in repository for commit\n";
      error = -1;
      if (work == 0)
        done (cb, error);
      return;
    }
  
    pde_t *de = &up->to_commit [idx];
  
    char buf[256];
    int r;
    if ((r = readlink (de->path, buf, 256)) < 0) {
      perror ("readlink");
      error = -1;
    }
    else {
      str link (buf, r);
      cfscallp c = New refcounted<cfscall>
        (CFSRWPROC_SYMLINK,
	 wrap (mkref(this), &commit_object::symlink_reply, idx, cb));
      cfsrw_symlink_arg *arg = static_cast<cfsrw_symlink_arg *> (c->arg ());
      arg->dir = de->dirfh;
      arg->name = de->name;
      arg->linkto = link;
      arg->attributes.valid_size = 1;
      arg->attributes.size = link.len ();
  
      work++;
      cfsrw_call (c);
      pastwatch_status ("Committing", de->path);
    }
  
    if (work == 0)
      done (cb, error);
  }

  void do_commit (cbi cb)
  {
    for (unsigned i=0; i<up->to_commit.size (); i++) {
      pde_t *de = &up->to_commit [i];
     
      if (de->type == CFSRW_LNK) {
	cfscallp c = New refcounted<cfscall>
	  (CFSRWPROC_UNLINK,
	   wrap (mkref(this), &commit_object::unlink_reply, i, cb));
	cfsrw_unlink_arg *arg = static_cast<cfsrw_unlink_arg *> (c->arg ());

	arg->object = de->fh;
	arg->type = de->type;
        arg->dir = de->dirfh;
        arg->name = de->name;
          
	work++;
        cfsrw_call (c);
      }
      else if (de->type == CFSRW_REG) {
	cfscallp c = New refcounted<cfscall>
	  (CFSRWPROC_SETATTR,
	   wrap (mkref(this), &commit_object::sattr_reply, i, cb));
        cfsrw_setattr_arg *arg =
	  static_cast<cfsrw_setattr_arg *> (c->arg ());

        arg->object = de->fh;
        arg->new_attributes.valid_size = 1;
	arg->new_attributes.size = 0;
	  
	work++;
	cfsrw_call (c);
      }
      else
	assert (0);
    }
  }

  void remove_unlink_reply (int i, cbi cb, cfscallp cc, cfsrw_status status)
  {
    work--;
    if (status) {
      warn << "cannot lookup " << up->to_remove [i].path << " in repository\n";
      error = -1;
      if (work == 0)
        done (cb, error);
      return;
    }
    pastwatch_status ("Removing", up->to_remove [i].path);
    commited.push_back (up->to_remove [i].path);
    vdb_schedule_remove (up->to_remove [i].path);
    if (work == 0)
      done (cb, error);
  }

  void remove_lookup_reply (int i, cbi cb, cfscallp cc, cfsrw_status status)
  {
    work--;
    if (status) {
      warn << "cannot lookup " << up->to_remove [i].path << " in repository\n";
      error = -1;
      if (work == 0)
        done (cb, error);
      return;
    }
	
    cfsrw_lookup_arg *arg = static_cast<cfsrw_lookup_arg *> (cc->arg ());
    cfsrw_lookup_res *res = static_cast<cfsrw_lookup_res *> (cc->res ());

    cfscallp c = New refcounted<cfscall>
      (CFSRWPROC_UNLINK,
       wrap (mkref(this), &commit_object::remove_unlink_reply, i, cb));
    cfsrw_unlink_arg *a = static_cast<cfsrw_unlink_arg *> (c->arg ());
    a->object = res->object;
    a->type = up->to_remove [i].type;
    a->dir = arg->dir;
    a->name = arg->name;
    work++;
    cfsrw_call (c);
  }

  void do_remove (cbi cb)
  {
    for (unsigned i=0; i<up->to_remove.size (); i++) {
      work++;
      cfsrw_lookup
	(root, up->to_remove [i].cfspath,
	 wrap (mkref(this), &commit_object::remove_lookup_reply, i, cb));
    }
  }
 
  void add_getattr_reply (int i, cbi cb, cfscallp c, cfsrw_status status)
  {
    work--;
    if (status) {
      warn << "cannot get file attribute from repository after add\n";
      error = -1;
      if (work == 0)
        done (cb, error);
      return;
    }
    
    cfsrw_getattr_arg *arg = static_cast<cfsrw_getattr_arg *> (c->arg ());
    cfsrw_getattr_res *res = static_cast<cfsrw_getattr_res *> (c->res ());
    pde_t *de = &up->to_add [i];
    vdb_schedule_add 
      (de->path, pfe_t (arg->object, res->attributes.mver));

    if (de->type == CFSRW_DIR)
      vdb_schedule_mkdir (de->path, de->cfspath);

    if (work == 0)
      done (cb, error);
  }
  
  void checkin_reply (int i, cbi cb, bigint fh, int status)
  {
    work--;
    if (status) {
      warn << "cannot checkin " << up->to_add [i].path << " to repository\n";
      error = -1;
      if (work == 0)
        done (cb, error);
      return;
    }

    work++;
    commited.push_back (up->to_add [i].path);
    getattr (fh, wrap (mkref(this), &commit_object::add_getattr_reply, i, cb));
  }

  void do_add (cbi cb)
  {
    for (unsigned i=0; i<up->to_add.size (); i++) {
      pde_t e;
      e.name  = up->to_add [i].name;
      e.type  = up->to_add [i].type;
      e.path  = up->to_add [i].path;
      e.dirfh = up->to_add [i].dirfh;
      work++;
      checkin (e, wrap (mkref(this), &commit_object::checkin_reply, i, cb));
    }
  }

  void
  update_done (str path, cbi cb, int err)
  {
    if (err) {
      warn << "please resolve error or conflicts first\n";
      done (cb, error);
    }
    else {
      warn << "committing in " << path << "\n";
      if (arg_debugmode) {
        warn << "should commit " << up->to_commit.size () << " items\n";
        warn << "should add " << up->to_add.size () << " items\n";
        warn << "should remove " << up->to_remove.size () << " items\n";
      }
  
      work++;
      do_add (cb);
      do_remove (cb);
      do_commit (cb);
      work--;
      if (work == 0)
        done (cb, error);
    }
  }

public:
  commit_object (bigint root, str path, cbi cb)
    : root (root), work (0), error (0)
  {
    warn << "checking for updates and conflicts in " << path << "\n";
    up = update_object_spec::alloc
      (root, path, wrap (mkref(this), &commit_object::update_done, path, cb),
       true, false);
  }
};

void
commit (bigint root, str path, cbi cb)
{
  ptr<commit_object> a = New refcounted<commit_object> (root, path, cb);
}

