/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

/* this file calls the GNU diff and diff3 utilities. */

#include <string.h>
#include <stdio.h>

#define DIFF  "diff"
#define DIFF3 "diff3"

#include "pastwatch.h"

int
diff (const char *f1, const char *f2, const char *args, bool print)
{
  int n = strlen (DIFF)+1+strlen (args)+1+strlen (f1)+1+strlen (f2)+1;
  char cmd[n];
  snprintf (cmd, n, "%s %s %s %s", DIFF, args, f1, f2);
  // printf ("%s\n", cmd);

  int r = 0;
  FILE *fp = popen (cmd, "r");
  char buf[1024];
  while (fgets (buf, 1024, fp)) {
    if (print)
      fprintf (stdout, "%s", buf);
    r += strlen (buf);
  }

  if (feof (fp))
    return r;
  return -1;
}

int
diff (const char *f1, const char *f2)
{
  return diff (f1, f2, "-c", true);
}

int
isdiff (const char *f1, const char *f2)
{
  return diff (f1, f2, "-c", false);
}


//
// merge () uses diff3 to merge two files from an older version.
// returns 0 if successful, -1 if error, 1 if there are conflicts.
//

int
merge (const char *fb, const char *tfb,
       const char *fw, const char *tfw,
       const char *fr, const char *tfr,
       bool check_only)
{
  int n = strlen ("-E -am") +1+3+strlen (tfw)+1+3+
          strlen(tfb)+1+3+strlen(tfr)+1;
  char args[n];
  snprintf (args, n, "-E -am -L %s -L %s -L %s", tfw, tfb, tfr);

  n = strlen (DIFF3)+1+strlen (args)+1+
      strlen (fw)+1+strlen (fb)+1+strlen (fr)+1;
  char cmd[n];
  snprintf (cmd, n, "%s %s %s %s %s", DIFF3, args, fw, fb, fr);
  
  // warn << cmd << "\n";

  FILE *fp = popen (cmd, "r");

  char buf[512];
  str s = pastwatch_tmpfile ();
  int fd = open (s, O_CREAT | O_TRUNC | O_RDWR, 0600);
 
  while (fgets (buf, 512, fp))
    write (fd, buf, strlen (buf));
  lseek (fd, 0, SEEK_SET);

  int error = -1;

  if (feof (fp)) {
    int conflict = pclose (fp);
    // warn << "conflict? " << conflict << "\n";

    if (conflict == 0)
      error = 0;
    else if (conflict > 0)
      error = 1;
    else
      error = -1;

    if (error >= 0 && !check_only) {
      int wfd = open (fw, O_TRUNC | O_WRONLY);
      if (wfd) {
        int r;
        while ((r = read (fd, buf, 512)) > 0) {
	  if ((r = write (wfd, buf, r)) < 0)
	    break;
        }
        if (r < 0) {
	  warn << "read/write error while merging\n";
	  error = -1;
	}
        close (wfd);
      }
    }
  }

  close (fd);
  unlink (s);
  return error;
}

/* code stolen from cvs, modified by benjie */

/*
 * Copyright (c) 1992, Brian Berliner and Jeff Polk
 * Copyright (c) 1989-1992, Brian Berliner
 * 
 * You may distribute under the terms of the GNU General Public License as
 * specified in the README file that comes with the CVS source distribution.
 * 
 * Various useful functions for the CVS support code.
 */

#define DIFF3_MERGE_PAT_1 "<<<<<<< "
#define DIFF3_MERGE_PAT_2 "=======\n"
#define DIFF3_MERGE_PAT_3 ">>>>>>> "

int
isconf (const char *path)
{
  FILE *fp;
  char line [2048];
  int result = 0;

  fp = fopen (path, "r");
  if (fp == NULL)
    return -1;

  while (fgets (line, 2048, fp)) {
    if (strncmp (line, DIFF3_MERGE_PAT_1, sizeof DIFF3_MERGE_PAT_1 - 1) == 0 ||
        strncmp (line, DIFF3_MERGE_PAT_2, sizeof DIFF3_MERGE_PAT_2 - 1) == 0 ||
        strncmp (line, DIFF3_MERGE_PAT_3, sizeof DIFF3_MERGE_PAT_3 - 1) == 0) {
      result = 1;
      goto out;
    }
    while (line [strlen (line)-1] != '\n') {
      warn << "long line\n";
      if (!fgets (line, 2048, fp))
	break;
    }
  }
  if (ferror (fp))
    return -1;

out:
  fclose (fp);
  return result;
}

