/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <errno.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "pastwatch.h"
#include "vdb.h"
#include "path.h"

extern bool arg_force;
extern bool arg_debugmode;

static bool
dir_empty (str path)
{
  DIR *dir = opendir (path);
  if (!dir)
    return false;
  struct dirent *e;
  while ((e = readdir (dir))) {
    if (strcmp (e->d_name, ".") && strcmp (e->d_name, "..") &&
        strcmp (e->d_name, PASTWATCH_REF)) {
      closedir (dir);
      return false;
    }
  }
  closedir (dir);
  return true;
}

static void
skip_lines (FILE *fp, unsigned nlines)
{
  char buf [256];
  for (unsigned i=0; i<nlines; i++) {
    fgets (buf, 256, fp);
    while (buf [strlen (buf)-1] != '\n' && fgets (buf, 256, fp));
  }
}

int
dirop (int command, str path)
{
  struct stat sb;
  int err = lstat (path, &sb);

  if (err && command != cmd_remove && command != cmd_clear) {
    perror ("cannot stat file or directory");
    return 1;
  }

  // check if the file or directory to be removed is still around; if
  // -f is given, remove the file or directory
  if (command == cmd_remove && !err) {
    if (arg_force) {
      if (sb.st_mode & S_IFDIR) {
	if (dir_empty (path)) {
	  if (remove_dir (path)) {
	    warn << "cannot remove directory " << path
	         << ": directory may be non-empty!\n";
	    return 1;
	  }
	}
	else {
	  warn << path << ": directory not empty, cannot remove.\n";
	  return 1;
	}
      }
      else if (unlink (path)) {
	perror ("unlink");
        warn << "cannot remove " << path << " from working directory.\n";
	return 1;
      }
    }
    else {
      warn << "please remove " << path << " from working directory first.\n";
      return 1;
    }
  }

  // check if directory already been added
  if (command == cmd_add && (sb.st_mode & S_IFDIR)) {
    str d = path << "/" << PASTWATCH_REF;
    if (lstat (d, &sb) == 0) {
      warn << path << " contains " << PASTWATCH_REF << ", already added?\n";
      return 1;
    }
  }

  str dir = get_parent_dir (path);
  str name;
  if (dir == ".")
    name = path;
  else if (dir == "")
    name = "";
  else
    name = substr (path, dir.len ()+1);
 
  // for "clear", just remove the entry from db
  if (command == cmd_clear) {
    vdb_remove (path);
    return 0;
  }
  
  bool exists = vdb_exists (path);

  // check if file already exists in repository or not
  if (!exists && command == cmd_remove) {
    warn << "cannot find " << path 
         << " in database, is it in the repository?\n";
    return 1;
  }
  else if (exists && command == cmd_add) {
    warn << path << " in database, already exists in repository?\n";
    return 1;
  }
 
  str c;
  if (command == cmd_add)
    c = "A";
  else if (command == cmd_remove)
    c = "R";
  else
    assert (0);
  str msg = c << " " << name << "\n";

  str p = get_entries_file (dir);

  int fd = open (p, O_RDWR, 0644);
  if (fd < 0) {
    perror ("cannot create or open op file");
    return 1;
  }

  FILE *fp = fdopen (fd, "r+");
  skip_lines (fp, 1);

  char buf [2+name.len ()+2]; // 2+ is for "A ", +2 is for \n and \0
  while (fgets (buf, 2+name.len ()+2, fp)) {
    if (strlen (buf) < 3 || 
	(buf [0] != 'A' && buf [0] != 'R') || buf [1] != ' ') {
      warn << "corrupted op file, cannot schedule operation.\n";
      fclose (fp);
      return 1;
    }
    if (buf [strlen (buf)-1] == '\n' && strlen (buf)-3 == name.len () &&
        memcmp (buf+2, name.cstr (), name.len ()) == 0) {
      warn << path << " already scheduled for add or remove.\n";
      fclose (fp);
      return 1;
    }
    while (buf [strlen (buf)-1] != '\n' && fgets (buf, 2+name.len ()+2, fp));
  }

  lseek (fd, 0, SEEK_END);
  if (write (fd, msg.cstr (), msg.len ()) < 0 || fclose (fp) < 0) {
    perror ("cannot write to op file");
    return 1;
  }
  return 0;
}

int
get_dirops (str path, vec<str>& afn, vec<str>& rfn, vec<str>& xfn)
{
  str p = get_entries_file (get_dir (path));

  int fd = open (p, O_RDONLY, 0644);
  if (fd < 0) {
    warn << "cannot open " << p << "\n";
    return -1;
  }

  FILE *fp = fdopen (fd, "r");
  skip_lines (fp, 1);

  char buf [256];
  while (fgets (buf, 256, fp)) {
    vec<str> *vp = (buf[0] == 'A') ? &afn : ((buf[0] == 'R') ? &rfn : &xfn);
    // assert (vp != &xfn);  // for now, we don't support rename"
    str s (buf, strlen (buf));
    s = substr (s, 2);
    while (buf [strlen (buf)-1] != '\n' && fgets (buf, 256, fp)) {
      strbuf b = s << str (buf, strlen (buf));
      s = b; 
    }
    s = substr (s, 0, s.len ()-1);
    vp->push_back (s);
  }
  fclose (fp);
  return 0;
}

int
set_dirops (str path, vec<str>& afn, vec<str>& rfn, vec<str>& xfn)
{
  str p = get_entries_file (get_dir (path));

  int fd = open (p, O_RDWR, 0644);
  if (fd < 0) {
    warn << "cannot open " << p << "\n";
    return -1;
  }
  
  FILE *fp = fdopen (fd, "r+");
  skip_lines (fp, 1);
  long offset = ftell (fp);
  lseek (fd, offset, SEEK_SET);
  
  for (unsigned i=0; i < afn.size(); i++) {
    str msg = "A " << afn [i] << "\n";
    if (write (fd, msg.cstr (), msg.len ()) < 0) {
      perror ("directory operation file");
      close (fd);
      return -1;
    }
  }

  for (unsigned i=0; i < rfn.size(); i++) {
    str msg = "R " << rfn [i] << "\n";
    if (write (fd, msg.cstr (), msg.len ()) < 0) {
      perror ("directory operation file");
      close (fd);
      return -1;
    }
  }

#if 0
  for (unsigned i=0; i < xfn.size(); i++) {
    str msg = "X " << xfn [i] << "\n";
    if (write (fd, msg.cstr (), msg.len ()) < 0) {
      perror ("directory operation file");
      close (fd);
      return -1;
    }
  }
#endif
  
  offset = lseek (fd, 0, SEEK_CUR);
  if (ftruncate (fd, offset)) {
    warn << "cannot truncate " << p << ": may be left in corrupted state.\n";
    close (fd);
    return -1;
  }

  close (fd);
  return 0;
}

int
remove_dirops (vec<str> &v)
{
  for (unsigned i=0; i<v.size (); i++) {
    str p = v [i];
    vec<str> av;
    vec<str> rv;
    vec<str> xv;
    str d = get_parent_dir (p);
    if (arg_debugmode)
      warn << "removing dirop " << p << ", try dir " << d << "\n";
    if (get_dirops (d, av, rv, xv))
      return -1;
    vec<str> nav;
    vec<str> nrv;
    vec<str> nxv;
    str n;
    if (d.len ()+1 < p.len ())
      n = substr (p, d.len ()+1);
    else
      n = p;
    bool removed = false;
    for (unsigned j=0; j<av.size () && !removed; j++) {
      if (n != av [j])
	nav.push_back (av [j]);
      else
	removed = true;
    }
    for (unsigned j=0; j<rv.size () && !removed; j++) {
      if (n != rv [j])
	nrv.push_back (rv [j]);
      else
	removed = true;
    }
    for (unsigned j=0; j<xv.size () && !removed; j++) {
      if (n != xv [j])
	nxv.push_back (xv [j]);
      else
	removed = true;
    }
    if (set_dirops (d, nav, nrv, nxv))
      return -1;
  }
  return 0;
}

int
get_cfspath (str path, str &cfspath)
{
  struct stat sb;
  
  str refpath;
  str name;

  if (lstat (path, &sb) == 0 && (sb.st_mode & S_IFDIR)) {
    name = "";
    refpath = get_entries_file (path);
  }
  else {
    str s = get_dir (path);
    refpath = get_entries_file (s);

    if (s == ".")
      name = path;
    else if (s == "")
      name = "";
    else
      name = substr (path, s.len ()+1);
  }

  char buf [1024];
  int fd = open (refpath, O_RDONLY);
  if (fd < 0) {
    warn << "cannot open " << refpath << "\n";
    return -1;
  }
  FILE *fp = fdopen (fd, "r");
  if (!fgets (buf, 1024, fp)) {
    warn << "cannot read " << refpath << "\n";
    fclose (fp);
    return -1;
  }

  assert (buf [strlen (buf)-1] == '\n');
  fclose (fp);

  if (strlen (buf) == 1)
    cfspath = name;
  else {
    str c = str (buf, strlen (buf)-1);
    strbuf b;
    if (name == "")
      b << c;
    else
      b << c << "/" << name;
    cfspath = b;
  }
  return 0;
}

