/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "parseopt.h"

#include "definc.h"
#include "view.h"
#include "ns.h"
#include "log.h"
#include "cfsrw.h"
#include "snapshot_spec.h"

#include "pastwatch.h"
#include "vdb.h"
#include "path.h"

int cfsrw_verbose = ENV("PASTWATCH_VERBOSE",0);

extern void checkin  (const pde_t& e, fhcb_t fhcb);
extern void checkout (bigint root, str path, str cfspath, cbi cb);
extern void update   (bigint root, str path, cbi cb);
extern void commit   (bigint root, str path, cbi cb);
extern int dirop (int command, str path);

extern void snapshot_find (ptr<cfscall> c, vervec start, cb_snapshot);
extern void snapshot_add  (bigint addr, vervec version);

bigint view_id;
bigint loghead_id;

static ptr<cfsrw> cfsrw_srv;
static ptr<view> repository;
static vervec current_version;
static ptr<sfspriv> sk;
static ptr<log> mylog;
static ptr<snapshot_obj> last_snapshot = 0;
static bigint last_snapshot_location;
static vervec last_snapshot_starts;
static str arg_path;
static int command;

static bool log_ok = false;
static bool view_ok = false;

const int cmd_checkin = 0;
const int cmd_checkout = 1;
const int cmd_update = 2;
const int cmd_commit = 3;
const int cmd_add = 4;
const int cmd_remove = 5;
const int cmd_rename = 6;
const int cmd_clear = 7;

str  arg_keystr;
str  arg_viewstr;
bool arg_updatedir = false;
bool arg_checkonly = false;
bool arg_debugmode = false;
bool arg_force = false;


//
// util functions
//

static inline bool
repositoryop (int command)
{
  if (command == cmd_checkin || command == cmd_checkout ||
      command == cmd_update || command == cmd_commit)
    return true;
  return false;
}

static inline bool
constop (int command)
{
  if (command == cmd_checkout || command == cmd_update)
    return true;
  return false;
}

static inline str
cmd2str (int cmd)
{
  str s;
  switch (cmd) {
    case cmd_checkout:
      s = "checkout";
      break;
    case cmd_update:
      s = "update";
      break;
    case cmd_checkin:
      s = "checkin";
      break;
    case cmd_commit:
      s = "commit";
      break;
    default:
      assert (0);
      break;
  }
  return s;
}

static vervec
intermediate_repository_version ()
{
  vervec v = current_version;
  v.add (loghead_id, mylog->last_id (), mylog->last_seq ());
  return v;
}

static unsigned
nrecords (const vervec &from, const vervec &to)
{
  unsigned n = 0;
  for (unsigned i=0; i < from.size () && i < to.size (); i++) {
    if (to.seqnum (i) > from.seqnum (i))
      n += (to.seqnum (i) - from.seqnum (i));
  }
  return n;
}


//
// exported
//

void
cfsrw_call (ptr<cfscall> c)
{
  c->l = mylog;
  c->s = last_snapshot;
  c->start = current_version;
  c->stop = last_snapshot_starts;
  cfsrw_srv->dispatch (c);
}


void
cfsrw_call_with_snapshot (ptr<cfscall> c, vervec start,
                          bigint addr, ptr<snapshot_obj> s)
{
  c->l = mylog;
  c->start = start;
  
  if (s) {
    if (arg_debugmode)
      warn << "using snapshot " << addr << "\n";
    assert (start.gt (s->version ()));
    c->s = s;
    c->stop = s->version ();
  } 
  else
    c->s = 0;
  cfsrw_srv->dispatch (c);
}

void
cfsrw_call_at (ptr<cfscall> c, vervec start)
{
  if (last_snapshot_starts.gt (start)) {
    snapshot_find (c, start, wrap (cfsrw_call_with_snapshot, c, start));
    return;
  }
  else {
    c->l = mylog;
    c->s = last_snapshot;
    c->start = start;
    c->stop = last_snapshot_starts;
    cfsrw_srv->dispatch (c);
  }
}

void
cfsrw_call_latest (ptr<cfscall> c)
{
  vervec v = intermediate_repository_version ();
  c->l = mylog;
  c->s = last_snapshot;
  c->start = v;
  c->stop = last_snapshot_starts;
  cfsrw_srv->dispatch (c);
}


//
// driver procedures in reverse order; main is at the end of the file,
// and the last procedure to run is below.
//

static void
unlock_repository (int status)
{
  // release mutual exclusion on repository
  exit (status);
}

static void
update_local (log_status status)
{
  if (status)
    warn << "warning: " 
         << "writing log-head failed, new snapshot location not written.\n";
  
  // update current version, but only take into account of our updates
  current_version = intermediate_repository_version ();
  if (command == cmd_checkout || command == cmd_commit ||
      command == cmd_update)
    vdb_process ();
  unlock_repository (0);
}

static void
log_release ()
{
  mylog->release (wrap (update_local));
}

static void
snapshot_created (callback<void,void>::ref cb,
                  bigint loc, ptr<snapshot_obj> s)
{
  if (s) {
    warn << "new snapshot: " << loc << "\n";
    last_snapshot_location = loc;
    last_snapshot = s;
    last_snapshot_starts = s->version ();
    mylog->snapshot (loc);
    snapshot_add (loc, s->version ());
  }
  else
    warn << "warning: cannot build snapshot after changes.\n";
  cb ();
}

static void
build_snapshot_after (log_status status)
{
  if (status) {
    warn << "writing log-head failed, all changes aborted.\n";
    unlock_repository (1);
    return;
  }
  
  vervec v = intermediate_repository_version ();
  if (last_snapshot) {
    if (arg_debugmode)
      warn << "build snapshot from previous snapshot.\n";
    snapshot_operation::build_snapshot
      (last_snapshot, last_snapshot_location, v,
       wrap (snapshot_created, wrap (log_release)));
  }
  else {
    if (arg_debugmode)
      warn << "build snapshot from scratch.\n";
    snapshot_operation::build_snapshot
      (repository->root (), v,
       wrap (snapshot_created, wrap (log_release)));
  }
}

static void
sync_log ()
{
  mylog->sync (wrap (build_snapshot_after));
}

static void
command_done (int error)
{
  if (error) {
    if (error < 0)
      warn << "error while running " << cmd2str (command) << "\n";
    update_local (LOG_OK);
    return;
  }
  
  if (constop (command))
    log_release ();
  else
    sync_log ();
}

static void
dispatch ()
{
  vervec v = current_version;
  for (unsigned i=0; i<v.size (); i++)
    v.add (v.head (i), v.latest (i), 0);

  bool empty = (nrecords (v, current_version) < 4);

  switch (command) {
    case cmd_checkin:
      if (empty)
        recurse (arg_path, wrap (checkin),
	         wrap (command_done), repository->root ());
      else {
        warn << "cannot checkin to existing repository, try update instead\n";
        unlock_repository (1);
	return;
      }
      break;
    case cmd_update:
      if (empty) {
	warn << "cannot checkout from empty repository\n";
	unlock_repository (1);
	return;
      }
      update (repository->root (), arg_path, wrap (command_done));
      break;
    case cmd_checkout:
      if (empty) {
	warn << "cannot checkout from empty repository\n";
	unlock_repository (1);
	return;
      }
      checkout (repository->root (), arg_path, "", wrap (command_done));
      break;
    case cmd_commit:
      if (empty) {
	warn << "cannot commit to empty repository\n";
	unlock_repository (1);
	return;
      }
      commit (repository->root (), arg_path, wrap (command_done));
      break;
    default:
      assert (0);
  }
}

static void
lock_repository ()
{
  // lock repository

  dispatch ();
}

static void
build_snapshot_before ()
{
  vervec v = current_version;
  if (view_ok && log_ok) {
    if (last_snapshot && nrecords (last_snapshot_starts, v) > 10) {
      if (arg_debugmode)
	warn << "build snapshot from previous snapshot.\n";
      snapshot_operation::build_snapshot
        (last_snapshot, last_snapshot_location, v,
	 wrap (snapshot_created, wrap (lock_repository)));
    }
    else if (!last_snapshot) {
      if (arg_debugmode)
	warn << "build snapshot from scratch.\n";
      snapshot_operation::build_snapshot
        (repository->root (), v,
	 wrap (snapshot_created, wrap (lock_repository)));
    }
    else
      lock_repository ();
  }
}

static void
got_repository_version (vervec v, log_status status)
{
  if (status) {
    warn << "cannot compute version of the repository\n";
    exit (1);
  }
  current_version = v;
  view_ok = true;
  build_snapshot_before ();
}

static void
load_snapshot (bigint loc, ptr<snapshot_obj> s)
{
  if (!s)
    last_snapshot = 0;
  else {
    warnx << "    snapshot: " << loc << "\n";
    last_snapshot_location = loc;
    last_snapshot = s;
    last_snapshot_starts = s->version ();
  }
  log_ok = true;
  build_snapshot_before ();
}
  
static void
got_log (log_status status)
{
  if (status) {
    warn << "cannot load log head\n";
    exit (1);
  }

  mylog->lock ();
  if (mylog->snapshot () != 0)
    snapshot_operation::load_snapshot
      (mylog->snapshot (), wrap (load_snapshot));
  else {
    log_ok = true;
    build_snapshot_before ();
  }
}

static void
got_view (log_status status)
{
  if (status) {
    warn << "cannot fetch view object\n";
    exit (1);
  }

  cfsrw_srv = New refcounted<cfsrw> (repository);
  repository->version (wrap (got_repository_version));
}

static void
sign_log_head (const str &data, cfsrw_sign_cb_t cb)
{
  sfs_sig2 s;
  sk->sign (&s, data);
  sfs_pubkey2 pk;
  sk->export_pubkey (&pk);
  cb (pk, s, LOG_OK);
}

static void
usage ()
{
  warn << "usage: " << progname << " [-fund] [-s keyfile] [-v view] cmd dir\n";
  exit (1);
}

int
main (int argc, char **argv)
{
  char *sk_file = getenv ("PASTWATCH_IDENTITY");
  char *viewstr = getenv ("PASTWATCH_VIEW");
  setprogname (argv[0]);

  int ch;
  while ((ch = getopt (argc, argv, "funds:v:")) != -1)
    switch (ch) {
    case 'f':
      arg_force = true;
      break;
    case 'u':
      arg_updatedir = true;
      break;
    case 'd':
      arg_debugmode = true;
      break;
    case 'n':
      arg_checkonly = true;
      break;
    case 'v':
      viewstr = optarg;
      break;
    case 's':
      sk_file = optarg;
      break;
    case '?':
    default:
      usage ();
    }
  argc -= optind;
  argv += optind;

  command = -1;
  if (strncmp (argv[0], "checkin", 7) == 0)
    command = cmd_checkin;
  else if (strncmp (argv[0], "checkout", 8) == 0)
    command = cmd_checkout;
  else if (strncmp (argv[0], "update", 6) == 0)
    command = cmd_update;
  else if (strncmp (argv[0], "commit", 6) == 0)
    command = cmd_commit;
  else if (strncmp (argv[0], "clear", 5) == 0)
    command = cmd_clear;
  else if (strncmp (argv[0], "add", 3) == 0)
    command = cmd_add;
  else if (strncmp (argv[0], "remove", 6) == 0)
    command = cmd_remove;
  // else if (strncmp (argv[0], "rename", 6) == 0)
    // command = cmd_rename;

  if ((argc != 2) || !sk_file || !viewstr || command == -1)
    usage ();

  arg_path = argv[1];
  arg_path = sanitize_path (arg_path);

  struct stat sb;
  if (lstat (arg_path, &sb) < 0 &&
      command != cmd_clear &&
      command != cmd_remove &&
      command != cmd_checkout) {
    warn << arg_path << ": cannot stat file or directory\n";
    return 1;
  }

  str dearm = dearmor64A (viewstr);
  mpz_set_rawmag_be (&view_id, dearm.cstr(), dearm.len());

  str key = file2wstr (sk_file);
  if (!key) {
    warn << sk_file << ": " << strerror (errno) << "\n";
    return 1;
  }
  else if (!(sk = sfscrypt.alloc_priv (key, SFS_SIGN))) {
    warn << "could not decode " << sk_file << "\n";
    warn << key << "\n";
    return 1;
  }

  ns::init ();
  if (snapshot_operation::init () < 0) {
    warn << "cannot create snapshot related directories\n";
    return 1;
  }
  random_init ();

  strbuf b;
  sk->export_pubkey (b, false);
  str pk_raw = b;
  loghead_id = compute_sha1_hash (pk_raw.cstr (), pk_raw.len ());
  
  strbuf bb;
  bb << loghead_id;
  arg_keystr = bb;
  arg_viewstr = str (viewstr, strlen (viewstr));

  if (vdb_init ())
    return 1;

  warnx << "        user: " << arg_keystr << "\n";
  warnx << "  repository: " << arg_viewstr << "\n";
  
  if (!repositoryop (command))
    return dirop (command, arg_path);

  mylog = New refcounted<log>
    (loghead_id, wrap (got_log), wrap (sign_log_head));
  repository = New refcounted<view> (wrap (got_view), view_id);
  amain();
}

