/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "pastwatch.h"

#define CONCURRENT_READS 16

struct read_obj {
  typedef callback<void,cfsrw_status>::ref cb_t;

  int fd;
  bigint fh;
  cb_t cb;

  uint64 size;
  uint64 off;
  unsigned int reads;
  int error;

  bool old_version;
  vervec start;

  void done () 
  {
    if (reads == 0) {
      if (error)
	cb (CFSRWERR_IO);
      else
	cb (CFSRW_OK);
      delete this;
    }
  }

  void
  read_reply (uint64 off, uint64 cnt, cfscallp c, cfsrw_status status)
  {
    reads --;

    if (status || error) {
      error = 1;
      done ();
      return;
    }

    cfsrw_read_res *res = static_cast<cfsrw_read_res *> (c->res ());
      
    if (res->count != cnt) {
      warn << "asked for " << off << "+" << cnt
	   << ", got " << res->count << "\n";
      error = 1;
      done ();
      return;
    }

    pwrite (fd, res->data.base (), res->count, off);
    do_read ();
    done ();
  }

  bool do_read() 
  {
    if (off >= size)
      return true;

    uint64 cnt = size - off;
    cnt = cnt > 8192 ? 8192 : cnt;
    reads ++;

    cfscallp c = New refcounted<cfscall>
      (CFSRWPROC_READ, wrap (this, &read_obj::read_reply, off, cnt));
    cfsrw_read_arg *arg = static_cast<cfsrw_read_arg *> (c->arg ());
    arg->file = fh;
    arg->offset = off;
    arg->count = cnt;

    off += cnt;
    if (old_version)
      cfsrw_call_at (c, start);
    else
      cfsrw_call (c);

    return false;
  }

  void get_attr (cfscallp c, cfsrw_status status)
  {
    if (!status) {
      cfsrw_getattr_res *res = static_cast<cfsrw_getattr_res *> (c->res ());
      size = res->attributes.size;
      off = 0;
      bool eof = false;
      for (int i = 0; i < CONCURRENT_READS && !eof; i++)
        eof = do_read ();
    }
    else
      error = 1;
    done ();
  }

  void init ()
  {
    if (ftruncate (fd, 0) < 0) {
      error = 1;
      done ();
      return;
    }
    cfscallp c = New refcounted<cfscall>
      (CFSRWPROC_GETATTR, wrap (this, &read_obj::get_attr));
    cfsrw_getattr_arg *arg = static_cast<cfsrw_getattr_arg *> (c->arg ());
    arg->object = fh;
    if (old_version)
      cfsrw_call_at (c, start);
    else
      cfsrw_call (c);
  }
  
  read_obj (int fd, bigint fh, read_obj::cb_t cb)
    : fd(fd), fh(fh), cb(cb), reads(0), error(0), old_version (false)
  {
    init ();
  }
  
  read_obj (int fd, bigint fh, vervec start, read_obj::cb_t cb)
    : fd(fd), fh(fh), cb(cb), reads(0), error(0),
      old_version (true), start(start)
  {
    init ();
  }

  ~read_obj() {}
};

void
read_file (int fd, bigint fh, callback<void, cfsrw_status>::ref cb)
{
  vNew read_obj (fd, fh, cb);
}

void
read_file (int fd, bigint fh, vervec start,
           callback<void, cfsrw_status>::ref cb)
{
  vNew read_obj (fd, fh, start, cb);
}

