/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <errno.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "cfsrw_wrapper.h"
#include "pastwatch.h"
#include "update.h"
#include "path.h"
#include "vdb.h"

extern void checkout (bigint fh, str path, str cfspath, cbi cb);
extern void read_file (int fd, bigint fh, callback<void, cfsrw_status>::ref);
extern void read_file (int fd, bigint fh, vervec,
                       callback<void, cfsrw_status>::ref);
extern int get_dirops (str path, vec<str> &, vec<str> &, vec<str> &);
extern int set_dirops (str path, vec<str> &, vec<str> &, vec<str> &);

extern int diff   (const char *f1, const char *f2);
extern int isdiff (const char *f1, const char *f2);
extern int merge  (const char *fb, const char *tb,
                   const char *fw, const char *tfw,
                   const char *fr, const char *tfr, bool check_only);
extern int isconf (const char *path);

class update_object : public virtual refcount, public update_object_spec
{
private:
  struct ude_t {  // update directory entry
    str name;
    str path;
    str cfspath;
    cfsrw_ftype type;
    mtime_t mtime;
    bool db;
    bigint fh;
    vervec last_version;
    mtime_t last_mtime;
    bool repository;
    vervec cur_version;
    bigint dirfh;
    bool dirempty;
    pde_t get_pde () {
      pde_t e;
      e.name = name;
      e.path = path;
      e.cfspath = cfspath;
      e.type = type;
      e.fh = fh;
      e.dirfh = dirfh;
      return e;
    }
  };
  typedef vec<ude_t> udev_t;

  struct upst_t {
    str path;
    str cfspath;
    bigint dirfh;
    udev_t local_dir;
    udev_t cfs_dir;
  };
  typedef ptr<upst_t> upstp_t;

  int work;
  int error;
  bool requires_update;
  
  cbi final_cb;
  bool checkonly;
  bool updatedir;

  vec<str> dirs;
  vec<str> dirs_to_remove;

  void
  update_done (int err)
  {
    for (unsigned i=0; i<dirs_to_remove.size (); i++) {
      if (remove_dir (dirs_to_remove [i]))
	warn << "cannot remove directory " << dirs_to_remove [i] 
	     << ": directory may be non-empty!\n";
    }
    final_cb (err);
  }

  //
  // update_next_directory () gets called whenever a directory update
  // is done.  it checks if there are other directories on the dirs
  // vector.  this is how we handle recursion.
  //

  void
  update_next_directory (upstp_t upstp, cbi cb)
  {
    if (upstp)
      dirs.pop_front ();
    if (dirs.size ()) {
      str path = dirs[0];
      str cfspath;
      if (get_cfspath (path, cfspath)) {
        cb (-1);
        return;
      }
      update_directory (path, cfspath, cb);
    }
    else if (checkonly && requires_update)
      cb (1); // 1 means requires update (-1 is error, 0 is okay)
    else
      cb (0);
  }
  
  
  //
  // compare individual file or symbolic link in working directory
  // against repository.  for files, uses diff and diff3 to get diffs
  // and merge diffs.
  //
  
  void
  compare_file_have_base (str base, int tfd, str current, ude_t de,
                          cbi cb, cfsrw_status status)
  {
    close (tfd);
  
    if (status) {
      unlink (current);
      unlink (base);
      cb (-1);
      return;
    }
  
    bool conflict = false;
    bool domerge = false;
  
    if (!(de.last_version).equals (de.cur_version)) { 
      // changed in repository

      domerge = true;
      int c = merge (base.cstr (), "base",
                     de.path.cstr (), "working",
	             current.cstr (), "current", checkonly);
      if (c < 0) {
        warn << de.path << ": cannot merge changes from repository\n";
        unlink (current);
        unlink (base);
        cb (-1);
        return;
      }
      else if (c)
        conflict = true;
    }
  
    if (!conflict) {
      int r = isdiff (de.path.cstr (), current.cstr ());
      if (r < 0) {
        warn << de.path << ": cannot compare against repository\n";
        unlink (current);
        unlink (base);
        cb (-1);
        return;
      }
      if (r) {
        pastwatch_status ("M", de.path);
        to_commit.push_back (de.get_pde ());
        if (!checkonly)
          // file changed, preserve old mtime until commit
	  vdb_schedule_add
	    (de.path, pfe_t (de.fh, de.cur_version, de.last_mtime));
      }
      else { // either content did not change or merge downloaded the changes
        if (domerge)
          pastwatch_status ("P", de.path);
        if (!checkonly)
	  vdb_schedule_add
	    (de.path, pfe_t (de.fh, de.cur_version, de.mtime));
      }
    }
    else {
      pastwatch_status ("C", de.path);
      if (!checkonly)
        // conflict, but changes are merged.  this means we can store
        // the new version vector from repository, but leave the old
        // mtime so that we know we still need to compare the file again
        // on the next update.
        vdb_schedule_add
	  (de.path, pfe_t (de.fh, de.cur_version, de.last_mtime));
    }
  
    unlink (current);
    unlink (base);
    cb (conflict);
  }
  
  void
  compare_file_have_current (str tfn, int tfd, ude_t de,
                             cbi cb, cfsrw_status status)
  {
    close (tfd);
  
    if (status) {
      unlink (tfn);
      cb (-1);
      return;
    }
  
    int r = isdiff (de.path.cstr (), tfn.cstr ());
    if (r < 0) {
      warn << de.path << ": cannot compare against repository\n";
      unlink (tfn);
      cb (-1);
      return;
    }
    if (r) {
      str s = pastwatch_tmpfile ();
      int fd = open (s, O_CREAT | O_TRUNC | O_WRONLY, 0600);
      read_file 
        (fd, de.fh, de.last_version,
         wrap (mkref(this), &update_object::compare_file_have_base,
	       s, fd, tfn, de, cb));
    }
    else {
      if (!checkonly)
        vdb_schedule_add (de.path, pfe_t (de.fh, de.cur_version));
      cb (0);
    }
  }
  
  void
  compare_file (ude_t de, cbi cb)
  {
    int r = isconf (de.path);
    if (r < 0) {
      warn << "cannot open file " << de.path << "\n";
      cb (-1);
      return;
    }
    if (r) {
      pastwatch_status ("C", de.path);
      if (!checkonly)
        // still conflict, don't merge.  keep old version and mtime.
        vdb_schedule_add
	  (de.path, pfe_t (de.fh, de.last_version, de.last_mtime));
      cb (-1);
      return;
    }
  
    str s = pastwatch_tmpfile ();
    int fd = open (s, O_CREAT | O_TRUNC | O_WRONLY, 0600);
    read_file 
      (fd, de.fh,
       wrap (mkref(this), &update_object::compare_file_have_current,
	     s, fd, de, cb));
  }
  
  void
  readlink_reply (ude_t de, cbi cb, cfscallp c, cfsrw_status status)
  {
    char buf [1024];
    int r = readlink (de.path, buf, 1024);
  
    if (status || r < 0) {
      warn << de.path << ": cannot read link value from cfs or local dir.\n";
      cb (-1);
      return;
    }
  
    bool modified = true;
    if ((de.last_mtime).sec == (de.mtime).sec &&
        (de.last_mtime).nsec == (de.mtime).nsec)
      modified = false;
  
    str l (buf, r);
    cfsrw_readlink_res *res = static_cast<cfsrw_readlink_res *> (c->res ());
  
    if (l != res->path) {
      if (!modified) {  // just download the link from repository
        if (!checkonly) {
          r = unlink (de.path);
          if (r < 0) {
            warn << de.path << ": cannot remove to create new symlink.\n";
            cb (-1);
            return;
          }
          r = symlink (res->path, de.path);
          if (r < 0) {
            warn << de.path << ": cannot create symbolic link.\n";
            cb (-1);
            return;
          }
          vdb_schedule_add (de.path, pfe_t (de.fh, de.cur_version));
        }
        pastwatch_status ("S", de.path);
      }
      else { // modified
        if (!(de.last_version).equals (de.cur_version)) { 
	  // changed in repository
          
	  pastwatch_status ("C", de.path);
	  // add the old version to preserve conflict until it is
	  // resolved. preserve mtime until commit.
          if (!checkonly)
	    vdb_schedule_add
	      (de.path, pfe_t (de.fh, de.last_version, de.last_mtime));
        }
        else {
          pastwatch_status ("M", de.path);
          to_commit.push_back (de.get_pde ());
	  // preserve mtime until commit.
          if (!checkonly)
	    vdb_schedule_add
	      (de.path, pfe_t (de.fh, de.cur_version, de.last_mtime));
        }
      }
    }
    else if (!checkonly)
      vdb_schedule_add (de.path, pfe_t (de.fh, de.cur_version));
    cb (0);
  }
  
  void
  update_file_with_attr (ude_t de, cbi cb)
  {
    assert (de.db);
    assert (de.repository);
  
    if ((de.last_version).gt (de.cur_version)) {
      warn << de.path
           << ": error: working version greater than repository version.\n";
      warn << de.path
           << ": wv = " << de.last_version.seqnum (0)
	   << ", rv = " << de.cur_version.seqnum (0) << "\n";
      cb (-1);
      return;
    }
 
    if (de.type == CFSRW_LNK || de.type == CFSRW_REG) {
      // regular file and symlinks
  
      bool modified = true;
      if ((de.last_mtime).sec == (de.mtime).sec &&
	  (de.last_mtime).nsec == (de.mtime).nsec)
        modified = false;
  
      if (!(de.last_version).equals (de.cur_version)) { 
	// changed in repository

        if (arg_debugmode)
	  warn << de.path << ": version changed in repository, download\n";
 
	if (de.type == CFSRW_LNK)
	  cfsrw_readlink
	    (de.fh,
	     wrap (mkref(this), &update_object::readlink_reply, de, cb));
        else 
          compare_file (de, cb);
      }
      else { // did not change in repository
  
        if (!modified) {
          if (!checkonly)
	    vdb_schedule_add
	      (de.path, pfe_t (de.fh, de.cur_version, de.mtime));
	  cb (0);
        }
        else { // modified locally, compare with repository
	  if (de.type == CFSRW_LNK)
	    cfsrw_readlink
	      (de.fh,
	       wrap (mkref(this), &update_object::readlink_reply, de, cb));
	  else
            compare_file (de, cb);
        }
      }
    }
    else { 
      // don't need to handle directory entry here; the entry was
      // pushed onto the dirs vector
      if (!checkonly)
        vdb_schedule_add (de.path, pfe_t (de.fh, de.cur_version));
      cb (0);
    }
  }
  
  
  //
  // compare_directory () compares working directory against the
  // corresponding directory in the repository. it checks for
  // conflicting directory operations, verify scheduled add and
  // remove, perform directory operations done in the repository, and
  // reports files not checked in.  after directory entries are
  // synchronized, it performs pairwise content comparison on each
  // file.
  //
  
  void
  checkout_done (upstp_t upstp, cbi cb, int err)
  {
    work --;
    if (err || error) {
      error = -1;
      if (work == 0) {
        cb (error);
      }
      return;
    }
    if (work == 0)
      update_next_directory (upstp, cb);
  }
  
  void
  update_dir_done (upstp_t upstp, cbi cb, int err)
  {
    work --;
    if (err || error) {
      error = -1;
      if (work == 0) {
        cb (error);
      }
      return;
    }
    if (work == 0)
      update_next_directory (upstp, cb);
  }
  
  void
  compare_directory (upstp_t upstp, cbi cb)
  {
    vec <bool> in_cfs;
    in_cfs.setsize (upstp->cfs_dir.size ());
    for (unsigned i=0; i<upstp->cfs_dir.size (); i++)
      in_cfs [i] = false;
 
    bool has_conflict = false;
    vec <str> local_remove;
    vec <str> local_rename_old;
    vec <str> local_rename_new;
    vec <int> local_unknown;
    vec <str> local_add_name;
    vec <cfsrw_ftype> local_add_type;
    vec <bigint> local_add_fh;
   
    // get list of scheduled directory operations (e.g. add, remove,
    // and rename)

    vec <str> addv;
    vec <str> removev;
    vec <str> renamev;
    vec <str> naddv;
    vec <str> nremovev;
    vec <str> nrenamev;

    if (get_dirops (upstp->path, addv, removev, renamev)) {
      warn << "cannot get list of scheduled directory operations!\n";
      cb (-1);
      return;
    }

    // check if a file that was scheduled for add disappeared
    vec<str> v;
    for (unsigned i=0; i<addv.size (); i++) {
      str path = upstp->path << "/" << addv [i];
      bool found = false;
      for (unsigned j=0; j<upstp->local_dir.size () && !found; j++) {
	if (addv [i] == upstp->local_dir [j].name)
	  found = true;
      }
      if (!found)
        warn << path << " scheduled for add, but disappeared.\n";
      else
	v.push_back (addv [i]);
    }
    addv = v;

    //
    // first, iterate through files in local working directory
    //
    for (unsigned i=0; i<upstp->local_dir.size (); i++) {
      ude_t *de = &(upstp->local_dir [i]);
      str path = de->path;

      // check if a file that was scheduled for removal re-appeared.
      vec<str> v;
      for (unsigned j=0; j<removev.size (); j++) {
        if (de->name == removev [j])
          warn << path << " scheduled for removal, but re-appeared.\n";
	else
	  v.push_back (removev [j]);
      }
      removev = v;
 
      bool badtype = false;
      bool modified = true;
      bool fh_exists = false;
      bool name_exists = false;
      unsigned cfs_idx = 0;
      unsigned cfs_name_idx = 0;
      str rep_name;
     
      // check if file has been modified
      if (de->mtime.sec == de->last_mtime.sec && 
	  de->mtime.nsec == de->last_mtime.nsec)
        modified = false;
  
      // check if file name and fh exist in repository; and if fh does
      // exist, what the name of the file is
      for (unsigned j=0; j<upstp->cfs_dir.size (); j++) {
        ude_t *cde = &(upstp->cfs_dir [j]);

        if (de->db && cde->fh == de->fh) { // fh exists
	  // check if type matches
          if (de->type != cde->type) {
            pastwatch_status 
	      ("C", path, "file type does not match one in repository");
	    has_conflict = true;
	    badtype = true;
            in_cfs [j] = true;
	    break;
          }
	  else {
            fh_exists = true;
	    rep_name = cde->name;
	    cfs_idx = j;
	  }
        }
        if (cde->name == de->name) {
          name_exists = true;
	  cfs_name_idx = j;
        }
      }
  
      //
      // between modified, fh_exists, and name_exists, there are 8
      // cases, some of which can be collapsed.
      //
      if (!badtype && fh_exists) {
        if (rep_name == de->name) {
	  // not modified, fh exists, and name is the same
	  if (arg_debugmode) {
	    if (modified)
	      warn << path << ": modified, fh exists, name same\n";
	    else
	      warn << path << ": not modified, fh exists, name same\n";
	  }
  
          in_cfs [cfs_idx] = true;
        }
        else {
	  // not modified, fh exists, but under a different name
	  if (arg_debugmode) {
	    if (modified)
	      warn << path << ": modified, fh exists, diff name\n";
	    else
	      warn << path << ": not modified, fh exists, diff name\n";
	  }
  
	  strbuf b;
	  b << "-> " << rep_name;
	  pastwatch_status ("X", path, b);
  
	  local_rename_old.push_back (de->name);
	  local_rename_new.push_back (rep_name);
	  de->name = rep_name;
          de->path = upstp->path << "/" << rep_name;
          de->cfspath = make_cfspath (upstp->cfspath, rep_name);
  
          in_cfs [cfs_idx] = true;
        }
      }

      else if (!badtype && !modified && !fh_exists) {
        // not modified, fh no longer exists in the same directory
        if (arg_debugmode)
	  warn << path << ": not modified, fh removed from dir\n";
  
        if (de->repository) {
	  // file in some other dir, signal a conflict
	  pastwatch_status ("C", path, "file moved to another directory?");
	  has_conflict = true;
        }
        else if (name_exists) {
          // file no longer in repository
	  pastwatch_status 
	    ("", path, "removed then re-appeared. re-run update.");
          local_remove.push_back (de->name);
        }
        else {
          // file no longer in repository
	  pastwatch_status ("", path, "no longer relevant.");
          local_remove.push_back (de->name);
        }
      }
  
      else if (!badtype && modified && !fh_exists) {
        if (de->db || name_exists) {
          // modified, but fh no longer exists in dir, this is a conflict
	  if (de->db) {
	    if (de->repository)
	      pastwatch_status ("C", path, "moved to another directory?");
	    else
	      pastwatch_status ("C", path, "modified, but no longer relevant.");
	  }
	  else
	    pastwatch_status ("C", path,
	                      "someone else already created this file.");
	  has_conflict = true;
        }
        else
	  local_unknown.push_back (i);
      }
    }
  
    //
    // now, find any file in cfs dir only and mark them for download,
    // but ignore files we want to remove
    //
    for (unsigned i=0; i<upstp->cfs_dir.size (); i++) {
      if (in_cfs [i])
        continue;
      
      ude_t *cde = &(upstp->cfs_dir [i]);
      str path = cde->path;
      str cfspath = cde->cfspath;
     
      bool should_remove = false;
      for (unsigned j=0; j<removev.size (); j++) {
        if (cde->name == removev [j]) {
	  // directory can only be removed if they are empty
	  if (cde->type == CFSRW_DIR && !cde->dirempty) {
	    str path = upstp->path << "/" << cde->name;
            warn << path 
	         << ": directory not empty in repository, cannot remove.\n";
	    has_conflict = true;
          }
	  else {
	    should_remove = true;
	    to_remove.push_back (cde->get_pde ());
	    nremovev.push_back (cde->name);
	    pastwatch_status ("R", path);
	  }
	  break;
	}
      }
      if (!should_remove) {
        local_add_name.push_back (cde->name);
        local_add_fh.push_back (cde->fh);
        local_add_type.push_back (cde->type);
      }
    }

    //
    // check for files that were removed already in the repository;
    // use vdb_schedule_remove to remove their db entries from working
    // directory
    //
    for (unsigned i=0; i<removev.size (); i++) {
      bool found = false;
      for (unsigned j=0; j<nremovev.size () && !found; j++) {
        if (removev [i] == nremovev [j])
	  found = true;
      }
      if (!found) {
	str path = upstp->path << "/" << removev[i];
	if (arg_debugmode)
	  warn << path << " has already been removed from repository.\n";
        vdb_schedule_remove (path);
      }
    }

    //
    // check files that may need to be added
    //
    for (unsigned i=0; i<local_unknown.size (); i++) {
      ude_t *de = &(upstp->local_dir [local_unknown [i]]);
      bool should_add = false;
      for (unsigned j=0; j<addv.size (); j++) {
        if (addv [j] == de->name) {
          // this file needs to be added
	  should_add = true;
	  to_add.push_back (de->get_pde ());
          naddv.push_back (de->name);
	  pastwatch_status ("A", de->path);
	  break;
	}
      }
      if (!should_add)
	pastwatch_status ("?", de->path);
    }

    if (set_dirops (upstp->path, naddv, nremovev, nrenamev)) {
      warn << "cannot write to directory operations file, abort.\n";
      cb (-1);
      return;
    }
  
    if (has_conflict) {
      warn << "please resolve above conflicts first!\n";
      cb (-1);
      return;
    }
  
    if (local_remove.size () || local_rename_old.size () || 
        local_add_name.size ())
      requires_update = true;
  
    if (!checkonly) { // actually update the local working directory
	
      // do remove before rename, so rename won't fail
      for (unsigned i=0; i<local_remove.size (); i++) {
        str path = upstp->path << "/" << local_remove [i];
        struct stat sb;
        if (lstat (path, &sb)) {
          strbuf b;
          b << "cannot remove " << path;
          perror (str (b).cstr ());
          cb (-1);
          return;
	}
	if ((sb.st_mode & S_IFDIR) == S_IFDIR)
	  dirs_to_remove.push_back (path);
	else if (unlink (path)) {
          strbuf b;
          b << "cannot remove " << path;
          perror (str (b).cstr ());
          cb (-1);
          return;
        }
        vdb_schedule_remove (path);
      }
      local_remove.clear ();
    
      // rename. do this in multiple passes.  in each pass, if target
      // exists, skip, and do it again on a later pass.  the idea is
      // that we don't remove a target, with rename, that needs to be
      // renamed later.
    
      vec <bool> local_rename_done;
      local_rename_done.setsize (local_rename_old.size ());
      for (unsigned i=0; i<local_rename_done.size (); i++)
        local_rename_done [i] = false;
    
      bool changed = true;
      while (changed) {
        changed = false;
        for (unsigned i=0; i<local_rename_old.size (); i++) {
          str oldf = upstp->path << "/" << local_rename_old [i];
          str newf = upstp->path << "/" << local_rename_new [i];
          struct stat sb;
          int r = lstat (newf, &sb);
          if (r < 0 && errno == ENOENT) {
	    // target does not exist, can rename
	    changed = true;
	    if (rename (oldf, newf)) {
	      strbuf b;
	      b << "cannot rename " << oldf << " to " << newf;
	      perror (str (b).cstr ());
	      cb (-1);
	      return;
	    }
	    local_remove.push_back (oldf);
	    // don't need to vdb_schedule_add newf; all renamed
	    // entries have the target entry marked in the in_cfs
	    // vector, so the code that pairwise compare files will do
	    // the vdb_schedule_add.
	    local_rename_done [i] = true;
          }
          else if (r < 0) {
	    strbuf b;
	    b << "cannot stat " << newf;
	    perror (str (b).cstr ());
	    cb (-1);
	    return;
          }
        }
      }
    
      for (unsigned i=0; i<local_remove.size (); i++) {
        str s = local_remove [i];
        assert (s != "");
        str path = upstp->path << "/" << s;
        vdb_schedule_remove (path);
      }
  
      for (unsigned i=0; i<local_rename_done.size (); i++) {
        if (!local_rename_done [i]) {
          warn << "rename " << local_rename_old [i] << " -> "
	       << local_rename_new [i] << " skipped!\n";
          assert (0);
        }
      }
    }
 
    // below this point, need to make sure work==0 before we call cb,
    // since we are running with multiple "threads"
    work++;
  
    // download new files
    for (unsigned i=0; i<local_add_name.size (); i++) {
      str path = upstp->path << "/" << local_add_name [i];
      str cfspath = make_cfspath (upstp->cfspath, local_add_name [i]);
      if (!checkonly && (local_add_type [i] != CFSRW_DIR || updatedir)) {
        work++;
        checkout (local_add_fh [i], path, cfspath,
	          wrap (mkref(this), &update_object::checkout_done, upstp, cb));
      }
      else if (checkonly) {
	if (local_add_type [i] != CFSRW_DIR)
	  pastwatch_status ("U", path);
	else if (updatedir)
	  pastwatch_status ("D", path);
      }
    }
  
    // do pairwise file update
    for (unsigned i=0; i<upstp->cfs_dir.size (); i++) {
      if (in_cfs [i]) {
        ude_t *cde = &(upstp->cfs_dir [i]);
        unsigned j;
        for (j=0; j<upstp->local_dir.size (); j++) {
          if (cde->fh == upstp->local_dir [j].fh)
	    break;
        }
        assert (j < upstp->local_dir.size ());
        ude_t de = upstp->local_dir [j];
        assert (de.name == cde->name);
  
        if (arg_debugmode)
	  warn << "updating " << de.path << "\n";
  
        work++;
        update_file_with_attr
	  (de,
	   wrap (mkref(this), &update_object::update_dir_done, upstp, cb));
      }
    }
  
    work--;
    if (work == 0)
      update_next_directory (upstp, cb);
  }
  
  
  //
  // readdir_reply () gathers directory entries from the repository.
  // for each entry, also calls cfsrw_getattr to gather its attribute
  // (reply is handled by cde_getattr_reply ()), and if the entry is a
  // directory, calls readdir to see if it is empty (reply is handled
  // by cde_readdir_reply ()).  after gathering directories entries,
  // calls compare_directory () to compare the working directory
  // against the repository.
  //
  
  void
  cde_readdir_reply (upstp_t upstp, unsigned idx, cbi cb,
	             cfscallp c, cfsrw_status status)
  {
    work--;
  
    if (status || error) {
      error = -1;
      if (status)
        warn << upstp->path << ": cannot read cfs directory\n";
      if (work == 0)
        cb (error);
      return;
    }
  
    ude_t *de = &(upstp->cfs_dir [idx]);
    cfsrw_readdir_res *res = static_cast<cfsrw_readdir_res *> (c->res ());
    de->dirempty = (res->entries == 0);

    if (work == 0)
      compare_directory (upstp, cb);
  }
  
  void
  cde_getattr_reply (upstp_t upstp, unsigned idx, cbi cb,
                     cfscallp c, cfsrw_status status)
  {
    work--;
    ude_t *de = &(upstp->cfs_dir [idx]);
  
    if (status || error) {
      error = -1;
      if (status)
        warn << de->cfspath << ": cannot get attribute\n";
      if (work == 0)
        cb (error);
      return;
    }
  
    cfsrw_getattr_res *res = static_cast<cfsrw_getattr_res *> (c->res ());
    de->type = res->attributes.type;
    de->repository = true;
    de->cur_version = res->attributes.mver;
    de->dirempty = false;

    if (de->type == CFSRW_DIR) {
      work++;
      cfsrw_readdir
	(de->fh, 0,
	 wrap (mkref(this), &update_object::cde_readdir_reply, upstp, idx, cb));
    }
  
    if (work == 0)
      compare_directory (upstp, cb);
  }
  
  void
  readdir_reply (upstp_t upstp, cbi cb,
	         cfscallp c, cfsrw_status status)
  {
    work--;
  
    if (status || error) {
      error = -1;
      if (status)
        warn << upstp->path << ": cannot read cfs directory\n";
      if (work == 0)
        cb (error);
      return;
    }
  
    cfsrw_readdir_res *res = static_cast<cfsrw_readdir_res *> (c->res ());
    rpc_ptr<cfs_rpc_dirent> e = res->reply.entries;
    uint64 cookie = 0;
  
    while (e) {
      ude_t de;
      de.name = e->name;
      de.path = upstp->path << "/" << e->name;
      de.cfspath = make_cfspath (upstp->cfspath, e->name);
      de.repository = true;
      de.fh = e->fh;
      de.dirfh = upstp->dirfh;
      upstp->cfs_dir.push_back (de);
      work++;
      cfsrw_getattr
        (e->fh, wrap (mkref(this), &update_object::cde_getattr_reply,
		      upstp, upstp->cfs_dir.size ()-1, cb));
      cookie = e->cookie;
      e = e->nextentry;
    }
  
    if (!res->reply.eof) {
      work++;
      cfsrw_readdir
        (upstp->dirfh, cookie,
	 wrap (mkref(this), &update_object::readdir_reply, upstp, cb));
    }
    else if (work == 0)
      compare_directory (upstp, cb);
  }
  
  
  //
  // update_directory () starts updating a directory by gathering all
  // directory entries in the local working directory. for each entry
  // that has a local cfs fh, also calls cfsrw_getattr to see if it is
  // still in the repository (reply is handled by localdir_getattr ()).
  // after this is done, call cfsrw_readdir to start gather directory
  // entries in the corresponding directory in the repository.
  //
  
  void
  lde_getattr (upstp_t upstp, unsigned idx, cbi cb,
               cfscallp c, cfsrw_status status)
  {
    work--;
    ude_t *de = &(upstp->local_dir [idx]);
  
    cfsrw_getattr_res *res = static_cast<cfsrw_getattr_res *> (c->res ());
    if (status || res->attributes.nlink == 0)
      de->repository = false;
    else {
      de->repository = true;
      de->cur_version = res->attributes.mver;
    }
   
    if (work == 0) {
      work++;
      cfsrw_readdir
	(upstp->dirfh, 0,
	 wrap (mkref(this), &update_object::readdir_reply, upstp, cb));
    }
  }
  
  void
  update_directory (str path, str cfspath, cbi cb)
  {
    ptr<pfe_raw> v = vdb_lookup (path);
    if (!v) {
      warn << "cannot find " << path << " in database, in repository?\n";
      cb (-1);
      return;
    }

    DIR *dir = opendir (path);
    if (dir == 0) {
      perror ("opendir");
      cb (-1);
      return;
    }
  
    upstp_t upstp = New refcounted <upst_t>;
    upstp->cfspath = cfspath;
    upstp->path = path;
    upstp->dirfh = v->fh ();
  
    warn << "updating " << path << "\n";
  
    struct dirent *e;
    while ((e = readdir (dir))) {
      if (e->d_type != DT_DIR && e->d_type != DT_REG && e->d_type != DT_LNK) {
        warn << "skipping " << e->d_name << ": file type "
             << e->d_type << " not supported\n";
        continue;
      }
      if (strcmp (e->d_name, ".") && strcmp (e->d_name, "..") &&
          strcmp (e->d_name, PASTWATCH_REF)) {
        ude_t de;
        de.name = e->d_name;
        de.path = path << "/" << e->d_name;
        de.cfspath = make_cfspath (cfspath, e->d_name);
        de.dirfh = upstp->dirfh;
        de.fh = 0;
  
        ptr<pfe_raw> v = vdb_lookup (de.path);
        de.db = (v != 0);
        if (v) {
	  de.fh = v->fh ();
	  de.last_mtime = v->mtime ();
	  de.last_version = v->version ();
        }
  
        struct stat sb;
        if (lstat (de.path, &sb)) {
	  perror ("lstat");
	  cb (-1);
	  return;
        }
	// for link, both S_IFLNK and S_IFREG are set, so check for
	// S_IFLNK first
	if ((sb.st_mode & S_IFLNK) == S_IFLNK)
	  de.type = CFSRW_LNK;
	else if ((sb.st_mode & S_IFREG) == S_IFREG)
	  de.type = CFSRW_REG;
	else if ((sb.st_mode & S_IFDIR) == S_IFDIR)
	  de.type = CFSRW_DIR;
        de.mtime.sec = sb.st_mtimespec.tv_sec;
        de.mtime.nsec = sb.st_mtimespec.tv_nsec;
        de.repository = false;
  
        upstp->local_dir.push_back (de);
  
        if (de.db) {
	  work++;
          cfsrw_getattr
	    (de.fh, wrap (mkref(this), &update_object::lde_getattr,
			  upstp, upstp->local_dir.size ()-1, cb));
        
	  if (e->d_type == DT_DIR)
	    dirs.push_back (de.path);
        }
      }
    }
  
    if (work == 0) {
      work++;
      cfsrw_readdir
	(upstp->dirfh, 0, wrap (mkref(this), &update_object::readdir_reply,
				upstp, cb));
    }
  }
  
  
  //
  // open db. figure out corresponding path in the repository.  check if
  // file or directory exists in the repository.
  //
  
  void
  file_getattr (str path, str cfspath, cbi cb,
                cfscallp c, cfsrw_status status)
  {
    cfsrw_getattr_res *res = static_cast<cfsrw_getattr_res *> (c->res ());
  
    if (status || res->attributes.nlink == 0) {
      pastwatch_status ("C", path, "removed from repository?");
      cb (-1);
      return;
    }
    
    struct stat sb;
    if (int err = lstat (path, &sb)) {
      perror ("cannot stat file or directory");
      cb (err);
      return;
    }
  
    if ((sb.st_mode & S_IFDIR) == S_IFDIR) {
      update_directory (path, cfspath, cb);
      return;
    }

    ptr<pfe_raw> v = vdb_lookup (path);
    assert (v);
  
    ude_t de;
    de.path = path;
    de.cfspath = cfspath;
    de.mtime.sec = sb.st_mtimespec.tv_sec;
    de.mtime.nsec = sb.st_mtimespec.tv_nsec;

    // for link, both S_IFLNK and S_IFREG are set, so check for
    // S_IFLNK first
    if ((sb.st_mode & S_IFLNK) == S_IFLNK)
      de.type = CFSRW_LNK;
    else if ((sb.st_mode & S_IFREG) == S_IFREG)
      de.type = CFSRW_REG;
    else if ((sb.st_mode & S_IFDIR) == S_IFDIR)
      de.type = CFSRW_DIR;
   
    if (de.type != res->attributes.type) {
      pastwatch_status 
	("C", path, "file type does not match one in repository");
      cb (-1);
      return;
    }
  
    de.db = true;
    de.fh = v->fh ();
    de.last_version = v->version ();
    de.last_mtime = v->mtime ();
    de.repository = true;
    de.cur_version = res->attributes.mver;
  
    update_file_with_attr (de, cb);
  }
  
  void
  lookup_reply (str path, str cfspath, cbi cb,
                cfscallp c, cfsrw_status status)
  {
    if (status) {
      warn << "cannot find " << cfspath << " in repository\n";
      cb (-1);
      return;
    }
  
    ptr<pfe_raw> v = vdb_lookup (path);
    
    if (!v) {
      pastwatch_status ("?", path);
      cb (0);
      return;
    }
  
    cfsrw_lookup_res *res = static_cast<cfsrw_lookup_res *> (c->res ());
    if (res->object != v->fh ()) {
      pastwatch_status 
        ("C", path, "object has been moved, update in parent directory?");
      cb (-1);
      return;
    }

    cfsrw_getattr
      (v->fh (),
       wrap (mkref(this), &update_object::file_getattr, path, cfspath, cb));
  }

public:
  update_object (bigint root, str path, cbi cb, bool checkonly, bool updatedir)
    : work (0), error (0), requires_update (0), final_cb (cb),
      checkonly (checkonly), updatedir (updatedir)
  {
    to_commit.clear ();
    to_remove.clear ();
    to_add.clear ();
  
    str cfspath;
    if (get_cfspath (path, cfspath)) {
      cb (-1);
      return;
    }
    
    struct stat sb;
    if (int err = lstat (path, &sb)) {
      perror ("cannot stat directory");
      cb (-1);
      return;
    }
          
    dirs.clear ();
    if ((sb.st_mode & S_IFDIR) == S_IFDIR)
      dirs.push_back (path);
  
    if (cfspath == "") // update on root dir
      cfsrw_getattr
	(root, wrap (mkref(this), &update_object::file_getattr,
		     path, cfspath, 
		     wrap (mkref(this), &update_object::update_done)));
    else // do lookup to get fh
      cfsrw_lookup
        (root, cfspath,
	 wrap (mkref(this), &update_object::lookup_reply, path, cfspath, 
	       wrap (mkref (this), &update_object::update_done)));
  }
};

void
update (bigint root, str path, cbi cb)
{
  ptr<update_object> a = New refcounted<update_object>
    (root, path, cb, arg_checkonly, arg_updatedir);
}

ptr<update_object_spec>
update_object_spec::alloc (bigint root, str path, cbi cb,
                           bool checkonly, bool updatedir)
{
  ptr<update_object> a = New refcounted<update_object>
    (root, path, cb, checkonly, updatedir);
  return a;
}

