/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <qhash.h>
#include "pastwatch.h"
#include "path.h"
#include "vdb.h"

/*
 * we keep local version information in a db.  this db contains
 * mapping from local file name to cfs file handle.  this file also
 * contains, for each entry, the version vector from the log record
 * that last modifies the object.
 */

static vec<str> add_path;
static vec<pfe_t> add_pfe;
static vec<str> mkdir_path;
static vec<str> mkdir_cfspath;
static vec<str> remove_path;

extern str arg_keystr;
extern str arg_viewstr;

static inline ref<dbrec>
str2dbrec (str s)
{
  return New refcounted<dbrec> (s.cstr (), s.len ());
}

static inline ref<dbrec>
bigint2dbrec (bigint fh)
{
  char buf[sha1::hashsize];
  bzero (buf, sha1::hashsize);
  mpz_get_rawmag_be (buf, sha1::hashsize, &fh);
  return New refcounted<dbrec> (buf, sha1::hashsize);
}

int
vdb_init ()
{
  add_path.clear ();
  add_pfe.clear ();
  mkdir_path.clear ();
  mkdir_cfspath.clear ();
  remove_path.clear ();

  char *home = getenv ("HOME");
  str pastwatch_dir = str (home, strlen (home)) << "/.pastwatch";
  if (mkdir (pastwatch_dir, 0755) && errno != EEXIST) {
    perror ("cannot create .pastwatch in home");
    return -1;
  }
  
  str userdir = pastwatch_dir << "/" << arg_keystr;
  if (mkdir (userdir, 0755) && errno != EEXIST) {
    perror ("cannot create user directory in .pastwatch");
    return -1;
  }

  str userviewdir = userdir << "/" << arg_viewstr;
  if (mkdir (userviewdir, 0755) && errno != EEXIST) {
    perror ("cannot create view directory");
    return -1;
  }

  return 0;
}

void
vdb_schedule_mkdir (str path, str cfspath)
{
  mkdir_path.push_back (path);
  mkdir_cfspath.push_back (cfspath);
}

void
vdb_schedule_remove (str path)
{
  remove_path.push_back (path);
}

void
vdb_schedule_add (str path, struct pfe_t pfe)
{
  add_path.push_back (path);
  add_pfe.push_back (pfe);
}

int
vdb_process ()
{
  for (unsigned i = 0; i < mkdir_path.size (); i++) {
    str p = mkdir_cfspath [i];
    str d = mkdir_path [i] << "/" << PASTWATCH_REF;
    if (mkdir (d, 0755)) {
      warn << "cannot create " << d << "\n";
      return -1;
    }
    str f = get_entries_file (mkdir_path [i]);
    str m = p << "\n";
    int fd = open (f, O_CREAT | O_TRUNC | O_WRONLY, 0644);
    if (fd < 0 || write (fd, m.cstr (), m.len ()) < 0) {
      warn << "cannot write to " << f << "\n";
      return -1;
    }
    close (fd);
  
    str dbp = get_db_file (mkdir_path [i]);
    ptr<dbfe> db = New refcounted <dbfe> ();
    dbOptions opts;
    if (int err = db->opendb((char *)dbp.cstr (), opts)) {
      warn << "cannot create " << dbp << "\n";
      return err;
    }
  }

  for (unsigned i=0; i < remove_path.size (); i++) {
    if (arg_debugmode)
      warn << "removing " << remove_path [i] << " from db\n";
    vdb_remove (remove_path [i]);
  }

  for (unsigned i = 0; i < add_path.size (); i++) {
    if (arg_debugmode)
      warn << "inserting " << add_path [i] << " into db\n";
    vdb_remove (add_path [i]);

    pfe_t pfe = add_pfe [i];
    if (pfe.mtime.sec == 0 && pfe.mtime.nsec == 0) {
      struct stat sb;
      if (int err = lstat (add_path [i], &sb)) {
        perror ("cannot stat file");
	return -1;
      }
      mtime_t m;
      if (sb.st_mtimespec.tv_sec > sb.st_ctimespec.tv_sec ||
	  (sb.st_mtimespec.tv_sec == sb.st_ctimespec.tv_sec &&
	   sb.st_mtimespec.tv_nsec > sb.st_ctimespec.tv_nsec)) {
        m.sec = sb.st_mtimespec.tv_sec;
        m.nsec = sb.st_mtimespec.tv_nsec;
      }
      else {
        m.sec = sb.st_ctimespec.tv_sec;
        m.nsec = sb.st_ctimespec.tv_nsec;
      }
      pfe.mtime = m;
    }
    if (vdb_insert (add_path [i], pfe))
      return -1;
  }

  return 0;
}

qhash<str, ptr<dbfe> > dbfeptrs;

static ptr<dbfe>
opendb (str dir)
{
  ptr<dbfe> *dbptr = dbfeptrs [dir];
  if (dbptr) {
    assert (*dbptr);
    return *dbptr;
  }

  str dbpath = get_db_file (dir);
  ptr<dbfe> db = New refcounted <dbfe> ();
  dbOptions opts;
  if (int err = db->opendb((char *)dbpath.cstr (), opts)) {
    warn << "cannot open " << dbpath << "\n";
    return 0L;
  }

  dbfeptrs.insert (dir, db);
  return db;
}

static void
split_path (str path, str &dir, str &name)
{
  dir = get_parent_dir (path);
  if (dir == ".")
    name = path;
  else if (dir == "")
    name = "";
  else
    name = substr (path, dir.len ()+1);
}

int
vdb_insert (str path, pfe_t pfe)
{
  struct stat sb;
  if (int err = lstat (path, &sb)) {
    warn << "cannot stat " << path << "\n";
    return -1;
  }
  if ((sb.st_mode & S_IFDIR) == S_IFDIR) {
    ptr<dbfe> db = opendb (path);
    if (!db)
      return -1;
    db->insert (str2dbrec (""), pfe.to_dbrec ());
  }

  str dir;
  str name;
  split_path (path, dir, name);

  str p = get_entries_file (dir);
  if (!lstat (p, &sb)) {
    ptr<dbfe> db = opendb (dir);
    if (!db)
      return -1;
    db->insert (str2dbrec (name), pfe.to_dbrec ());
  }
  return 0;
}

int
vdb_remove (str path)
{
  str dir;
  str name;
  split_path (path, dir, name);
  
  ptr<dbfe> db = opendb (dir);
  if (!db)
    return -1;
  db->del (str2dbrec (name));
  return 0;
}

bool
vdb_exists (str path)
{
  str dir;
  str name;
  split_path (path, dir, name);
  
  ptr<dbfe> db = opendb (dir);
  if (!db)
    return false;
  ptr<dbrec> k = str2dbrec (name);
  ptr<dbrec> v = db->lookup (k);
  if (v)
    return true;
  return false;
}

ptr<pfe_raw>
vdb_lookup (str path)
{
  struct stat sb;
  if (int err = lstat (path, &sb)) {
    warn << "cannot stat " << path << "\n";
    return 0L;
  }
  if ((sb.st_mode & S_IFDIR) == S_IFDIR) {
    ptr<dbfe> db = opendb (path);
    if (db) {
      ptr<dbrec> k = str2dbrec ("");
      ptr<dbrec> v = db->lookup (k);
      ptr<pfe_raw> p = 0;
      if (v)
        p = New refcounted<pfe_raw> (v);
      return p;
    }
  }

  str dir;
  str name;
  split_path (path, dir, name);
  
  ptr<dbfe> db = opendb (dir);
  if (!db)
    return 0L;
  
  ptr<dbrec> k = str2dbrec (name);
  ptr<dbrec> v = db->lookup (k);
  ptr<pfe_raw> p = 0;
  if (v)
    p = New refcounted<pfe_raw> (v);
  return p;
}

int
remove_dir (str path)
{
  str f0 = get_entries_file (path);
  unlink (f0);
  str f1 = get_db_file (path);
  unlink (f1);
  str d0 = path << "/" << PASTWATCH_REF;
  rmdir (d0);
  if (rmdir (path))
    return -1;
  return 0;
}

