/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "dbfe.h"
#include "definc.h"

#include "snapshot_spec.h"
#include "cfsrw_rec.h"

#include "snapshot.h"

//
// load and build
//

int
snapshot_operation::init ()
{
  if (mkdir("/var/tmp/cfsrw-scache", 0755) < 0 && errno != EEXIST) {
    warn << "cannot create directory /var/tmp/cfsrw-scache\n";
    return -1;
  }
  for (unsigned i=0; i<240; i++) {
    strbuf s;
    s << "/var/tmp/cfsrw-scache/";
    s << i;
    if (mkdir(str(s).cstr(), 0755) < 0 && errno != EEXIST) {
      warn << "cannot create directory " << str(s) << "\n";
      return -1;
    }
  }
  return 0;
}

inline ptr<snapshot>
obj2impl (ptr<snapshot_obj> s)
{
  ptr<snapshot> i = *((ptr<snapshot>*) &s);
  return i;
}

void
snapshot_operation::load_snapshot
  (bigint location, callback<void, bigint, ptr<snapshot_obj> >::ref cb)
{
  ptr<snapshot> s = New refcounted<snapshot> (location, cb);
}

void
snapshot_operation::build_snapshot
  (ptr<snapshot_obj> prev, bigint loc, vervec start,
   callback<void, bigint, ptr<snapshot_obj> >::ref cb)
{
  ptr<snapshot> ps = obj2impl (prev);
  ptr<snapshot> s = New refcounted<snapshot> (0, start, loc, ps, cb);
}

void
snapshot_operation::build_snapshot
  (bigint root, vervec start,
   callback<void, bigint, ptr<snapshot_obj> >::ref cb)
{
  ptr<snapshot> s = New refcounted<snapshot>
    (root, start, 0, (ptr<snapshot>) 0, cb);
}

//
// misc
//

void 
set_cfsrwattr (cfsrw_sattr *sa, cfs_inode_lnk *lnode)
{
  sa->valid_size = 1;
  sa->size = lnode->attr.size;
  sa->valid_mode = 1;
  sa->mode = lnode->attr.mode;
  sa->valid_uid = 1;
  sa->uid = lnode->attr.uid;
  sa->valid_gid = 1;
  sa->gid = lnode->attr.gid;
}

void 
set_cfsrwattr (cfsrw_sattr *sa, cfs_inode_reg *rnode)
{
  sa->valid_size = 1;
  sa->size = rnode->attr.size;
  sa->valid_mode = 1;
  sa->mode = rnode->attr.mode;
  sa->valid_uid = 1;
  sa->uid = rnode->attr.uid;
  sa->valid_gid = 1;
  sa->gid = rnode->attr.gid;
}

ptr<log_record>
get_link_record (ptr<cfs_data> sdata, bigint dir, str name)
{  
  bigint inum;
  cfsrw_ftype itype = CFSRW_REG;
  cfs_inode *inode = sdata->inode;
  
  if (inode->type == CFSLNK) {
    str dearm = dearmor64A ((*(*inode).lnk).path.cstr ());
    mpz_set_rawmag_be(&inum, dearm.cstr(), dearm.len());
    itype = CFSRW_LNK;
  } 
  else {
    str dearm = dearmor64A ((*(*inode).reg).path.cstr ());
    mpz_set_rawmag_be(&inum, dearm.cstr(), dearm.len());
    if (inode->type == CFSREG || inode->type == CFSREG_EXEC)
      itype = CFSRW_REG;
    else
      itype = CFSRW_DIR;
  }
  ptr<log_record> r =
    New refcounted <link_log_record> (inum, itype, dir, name);
  return r;
}

ptr<log_record>
get_inode_record (ptr<cfs_data> sdata, bigint dir, str name)
{
  bigint inum;
  cfsrw_ftype itype = CFSRW_REG;
  cfs_inode *inode = sdata->inode;
  cfsrw_sattr sa;
  
  if (inode->type == CFSLNK) {
    str dearm = dearmor64A ((*(*inode).lnk).path.cstr ());
    mpz_set_rawmag_be(&inum, dearm.cstr(), dearm.len());
    itype = CFSRW_LNK;
    set_cfsrwattr (&sa, inode->lnk);
  } 
  else {
    str dearm = dearmor64A ((*(*inode).reg).path.cstr ());
    mpz_set_rawmag_be(&inum, dearm.cstr(), dearm.len());
    if (inode->type == CFSREG || inode->type == CFSREG_EXEC)
      itype = CFSRW_REG;
    else
      itype = CFSRW_DIR;
    set_cfsrwattr (&sa, inode->reg);
  }
  ptr<log_record> r = New refcounted <inode_log_record> (itype, inum, sa);
  return r;
}


//
// find_parent
//

static void
find_parent_cb (bigint object, cfsrw_states::cb_fh cb, ptr<cfs_data> data)
{
  if (!data) {
    cb (0, CFSRWERR_IO);
    return;
  }
  if (data->type != CFS_INODE) {
    cb (0, CFSRWERR_IO);
    return;
  }
  cfs_inode *inode = (*data).inode;
  if (inode->type != CFSDIR) {
    cb (0, CFSRWERR_IO);
    return;
  }
  cb (inode->reg->parent, CFSRW_OK);
}

void 
snapshot_operation::find_parent (ptr<snapshot_obj> s, bigint object,
                                 cfsrw_states::cb_fh cb)
{
  ptr<dbrec> q = id2dbrec (object);
  ptr<dbrec> v = obj2impl (s)->fmap_db->lookup (q);
  
  if (!v) {
    cb (0, CFSRWERR_IO);
    return;
  }
  ns_cfsrw::retrieve_cfs_readonly
    (dbrec2id(v), wrap (find_parent_cb, object, cb));
}

//
// lookup
//

struct lookup_obj {
  bigint dir;
  str name;
};

static void 
lookup_cb3 (bigint dir, str name, cfsrw_states::cb_lookup cb,
            ptr<cfs_data> data)
{
  ptr<log_record> r = 0;
  
  if (!data || data->type != CFS_INODE) {
    cb (r, CFSRWERR_IO);
    return;
  }
  r = get_link_record (data, dir, name);
  cb (r, CFSRW_OK);
}

static void 
lookup_cb2 (ptr<snapshot> s, cfs_inode *inode, uint32 blocknum,
            ptr<lookup_obj> object, cfsrw_states::cb_lookup cb,
	    ptr<cfs_data> data)
{
  ptr<log_record> r = 0;
  
  if (!data || data->type != CFS_DIRBLK) {
    cb (r, CFSRWERR_IO);
    return;
  }

  cfs_dirent *e = (*(*data).dir).entries;
  while (e && (e->name != object->name))
    e = e->nextentry;
    
  if (!e) {
    if (inode->reg->direct.size () > blocknum+1) {
      blocknum++;
      bigint dirblock = inode->reg->direct[blocknum];
      WARN (1) << "loading dirblk " << blocknum << ": " << dirblock << "\n";
      ns_cfsrw::retrieve_cfs_readonly
        (dirblock, wrap (lookup_cb2, s, inode, blocknum, object, cb));
      return;
    }
    cb (r, CFSRWERR_NOENT);
  }
  else {
    ptr<dbrec> q = id2dbrec (e->fh);
    ptr<dbrec> v = s->fmap_db->lookup (q);
    if (!v) {
      cb (r, CFSRWERR_IO);
      return;
    }
    ns_cfsrw::retrieve_cfs_readonly
      (dbrec2id(v), wrap (lookup_cb3, object->dir, object->name, cb)); 
  }
}

static void
lookup_cb1 (ptr<snapshot> s, bigint dir, str name,
            cfsrw_states::cb_lookup cb, ptr<cfs_data> data)
{
  ptr<log_record> r = 0;
  
  if (!data || data->type != CFS_INODE ||
      ((*data).inode)->type != CFSDIR) {
    cb (r, CFSRWERR_IO);
    return;
  }

  cfs_inode *inode = (*data).inode;
  
  if (inode->reg->direct.size() > 0) {
    bigint dirblock = inode->reg->direct[0];
    WARN (1) << "loading directory block 0: " << dirblock << "\n";
    ptr<lookup_obj> object = New refcounted<lookup_obj>;
    object->dir = dir;
    object->name = name;
    ns_cfsrw::retrieve_cfs_readonly
      (dirblock, wrap (lookup_cb2, s, inode, 0, object, cb));  
  }
  else {
    // directory is empty
    cb (r, CFSRWERR_NOENT);
    return;
  }
}

void 
snapshot_operation::lookup (ptr<snapshot_obj> s, bigint dir, str name,
                            cfsrw_states::cb_lookup cb)
{
  ptr<dbrec> q = id2dbrec (dir);
  ptr<dbrec> v = obj2impl (s)->fmap_db->lookup (q);
  
  if (!v) {
    cb ((ptr<log_record>) 0, CFSRWERR_IO);
    return;
  }
  ns_cfsrw::retrieve_cfs_readonly
    (dbrec2id(v), wrap (lookup_cb1, obj2impl (s), dir, name, cb));
}

//
// find_inode
//

static void
combine_attr (ptr<cfsrw_states::attr> ca, ptr<cfs_data> sdata)
{
  cfs_inode *inode = sdata->inode;

  if (!ca->truncated) {
    if (inode->type == CFSLNK && 
	(ca->sattr.size < inode->lnk->attr.size || !ca->sattr.valid_size))
      ca->sattr.size = inode->lnk->attr.size;
    else if (inode->type != CFSLNK &&
	(ca->sattr.size < inode->reg->attr.size || !ca->sattr.valid_size))
      ca->sattr.size = inode->reg->attr.size;
  }
  if (inode->type == CFSLNK) {
    cfs_inode_lnk *lnode= inode->lnk;
    ca->nlink += lnode->nlink;
    ca->sattr.mode = ca->sattr.valid_mode ? ca->sattr.mode : lnode->attr.mode;
    ca->sattr.uid = ca->sattr.valid_uid ? ca->sattr.uid : lnode->attr.uid;
    ca->sattr.gid = ca->sattr.valid_gid ? ca->sattr.gid : lnode->attr.gid;
    if (!ca->sattr.valid_mtime || ca->sattr.mtime < lnode->mtime.seconds) {
      ca->sattr.mtime = lnode->mtime.seconds;
      ca->sattr.valid_mtime = true;
    }
    if (ca->ctime < lnode->ctime.seconds)
      ca->ctime = lnode->ctime.seconds;
    if (!ca->mver_set)
      get_mver (ca->mver, lnode->mver);
  }
  else {
    cfs_inode_reg *rnode= inode->reg;
    ca->nlink += rnode->nlink;
    ca->sattr.mode = ca->sattr.valid_mode ? ca->sattr.mode : rnode->attr.mode;
    ca->sattr.uid = ca->sattr.valid_uid ? ca->sattr.uid : rnode->attr.uid;
    ca->sattr.gid = ca->sattr.valid_gid ? ca->sattr.gid : rnode->attr.gid;
    if (!ca->sattr.valid_mtime || ca->sattr.mtime < rnode->mtime.seconds) {
      ca->sattr.mtime = rnode->mtime.seconds;
      ca->sattr.valid_mtime = true;
    }
    if (ca->ctime < rnode->ctime.seconds)
      ca->ctime = rnode->ctime.seconds;
    if (!ca->mver_set)
      get_mver (ca->mver, rnode->mver);
  }
  ca->sattr.valid_size = 1;
  ca->sattr.valid_mode = 1;
  ca->sattr.valid_uid = 1;
  ca->sattr.valid_gid = 1;
  ca->sattr.valid_mtime = 1;
}

static void 
get_inode_cb (bigint inum, ptr<cfsrw_states::attr> ca,
              cfsrw_states::cb_attr cb, ptr<cfs_data> data)
{
  if (!data || data->type != CFS_INODE) {
    if (!data)
      warn << "io, no data\n";
    else
      warn << "no inode\n";
    cb (ca, CFSRWERR_IO);
    return;
  }
  ptr<log_record> r = get_inode_record (data, 0, 0);
  inode_log_record *i = r->template record<inode_log_record> ();
  combine_attr (ca, data);
  ca->type = i->object_type ();
  cb (ca, CFSRW_OK);
}

void 
snapshot_operation::find_inode (ptr<snapshot_obj> s, bigint inum,
                                ptr<cfsrw_states::attr> fa,
				cfsrw_states::cb_attr cb)
{
  ptr<dbrec> q = id2dbrec (inum);
  ptr<dbrec> v = obj2impl (s)->fmap_db->lookup (q);
  
  if (!v) {
    // warn << "stale\n";
    cb (fa, CFSRWERR_STALE);
    return;
  }
  ns_cfsrw::retrieve_cfs_readonly
    (dbrec2id(v), wrap (get_inode_cb, inum, fa, cb));
}

//
// readdir
//

static void
got_dir_block (bigint dir, cfs_inode *inode, uint32 blocknum,
               ptr<cfsrw_states::readdir> rs, cfsrw_states::cb_readdir cb,
	       ptr<cfs_data> data)
{
  if (!data || data->type != CFS_DIRBLK) {
    cb (rs, CFSRWERR_IO);
    return;
  }
  
  bool skip;

  cfs_dirent *e = (*(*data).dir).entries;
  while (e) {
    WARN (2) << "looking at dirent name " << e->name << "\n";
    skip = false;
     
    for (uint i=0; i < rs->changed.size(); i++) {
      unlink_log_record *ue =
	rs->changed[i]->template record<unlink_log_record> ();
      rename_log_record *re =
	rs->changed[i]->template record<rename_log_record> ();
      if ((ue && ue->name () == e->name) ||
	  (re && ((re->from_dir () == dir && re->from_name () == e->name) ||
		  (re->target_exists () && re->target () == e->fh))))
	skip = true;
    }

    if (!skip) {
      // check if there's enough space in RPC reply, &c. this avoids a
      // crash when rmdir checks if a directory is empty, and also
      // perhaps when a directory is too large to fit in a single
      // READDIR reply.
      if (rs && !dup_entry (&(rs->res->reply.entries), e)) {
        if (rs->cookie && rs->n < rs->cookie)
	  rs->n++;
        else {
	  if (rs->count < (e->name.len ()+30)) {
	    cb (rs, CFSRW_OK);
            return;
	  }
	  else
	    rs->count -= (e->name.len ()+30);

          WARN(2) << "including " << e->name << "\n";
          (*rs->direntp).alloc();
	  (*rs->direntp)->fh = e->fh;
          (*rs->direntp)->name = e->name;
          rs->n++;
          (*rs->direntp)->cookie = rs->n;
          rs->direntp = &(*rs->direntp)->nextentry;
        }
      }
    }
    e = e->nextentry;
  }

  if (inode->reg->direct.size() > blocknum+1) {
    blocknum++;
    bigint dirblock = inode->reg->direct[blocknum];
    ns_cfsrw::retrieve_cfs_readonly
      (dirblock, wrap (got_dir_block, dir, inode, blocknum, rs, cb));
    return;
  }

  rs->res->reply.eof = true;
  cb (rs, CFSRW_OK);
}

static void 
got_dir_inode (bigint dir, ptr<cfsrw_states::readdir> rs,
               cfsrw_states::cb_readdir cb, ptr<cfs_data> data)
{
  if (!data || data->type != CFS_INODE) {
    cb (rs, CFSRWERR_IO);
    return;
  }
  
  cfs_inode *inode = (*data).inode;
  if (inode->type != CFSDIR) {
    cb (rs, CFSRWERR_NOTDIR);
    return;
  }

  if (inode->reg->direct.size() > 0) {
    bigint dirblock = inode->reg->direct[0];
    ns_cfsrw::retrieve_cfs_readonly
      (dirblock, wrap (got_dir_block, dir, inode, 0, rs, cb));
  }
  else {
    rs->res->reply.eof = true;
    cb (rs, CFSRW_OK);
    return;
  }
}

void
snapshot_operation::readdir (ptr<snapshot_obj> s, bigint dir,
                             ptr<cfsrw_states::readdir> rs,
			     cfsrw_states::cb_readdir cb)
{
  ptr<dbrec> q = id2dbrec (dir);
  ptr<dbrec> v = obj2impl (s)->fmap_db->lookup (q);
  if (!v) {
    cb (rs, CFSRWERR_NOENT);
    return;
  }
  ns_cfsrw::retrieve_cfs_readonly
    (dbrec2id(v), wrap(got_dir_inode, dir, rs, cb));
}

//
// read
//

struct read_state {
  ptr<snapshot> s;
  int outstanding_fetches;
  int fetch_errors;
  ptr<cfsrw_states::read> rs;
  cfsrw_states::cb_read rcb;

  read_state (ptr<snapshot> s)
    : s(s), outstanding_fetches(0), fetch_errors(0) {}
};

static void
done_fill_buf (ptr<cfsrw_states::read> rs, cfsrw_states::cb_read rcb)
{
  uint64 c = rs->rng->end () - rs->offset;
  if (c > rs->count)
    c = rs->count;
  bool eof = ((rs->offset+c) == rs->rng->end ());

  if (rs->valid_size) {
    if (rs->offset > rs->size) {
      c = 0;
      eof = true;
    }
    else if (rs->offset+c > rs->size) {
      c = rs->size-rs->offset;
      eof = true;
    }
  }
  WARN (2) << "done_fill_buf " << c << " bytes, eof? " << eof << "\n";
  rs->eof = eof;
  rs->rcount = c;
  rcb (rs, CFSRW_OK);
}

static void
check_fill (ptr<read_state> rstate)
{
  if (rstate->outstanding_fetches == 0) {
    // don't need to add 1 to _count here, since we know the size of
    // the file in snapshot
    if (rstate->rs->rng->check (rstate->rs->offset, rstate->rs->count) || 
	rstate->fetch_errors == 0)
      done_fill_buf (rstate->rs, rstate->rcb);
    else
      rstate->rcb (rstate->rs, CFSRWERR_IO);
  }
}

static void 
got_file_block (ptr<read_state> rstate, uint64 gs, uint64 gl,
                ptr<cfs_data> data)
{
  rstate->outstanding_fetches--;
  if (!data || data->type != CFS_FILEBLK) {
    rstate->fetch_errors++;
    check_fill (rstate);
    return;
  }

  // uint64 blocknum = gs / SNAPSHOT_BLOCKSIZE;
  // uint64 blockstart = blocknum * SNAPSHOT_BLOCKSIZE;
  //
  // XXX the following logic is horribly broken... we didn't load up
  // the entire "gl" number of bytes
  //
  uint64 start = (gs % SNAPSHOT_BLOCKSIZE);
  uint64 size = rstate->rs->offset + rstate->rs->count - gs;
  if (size > gl)
    size = gl;
  if (size > SNAPSHOT_BLOCKSIZE - start)
    size = SNAPSHOT_BLOCKSIZE - start;

  uint64 bi = gs - rstate->rs->offset;
  uint64 wi = start;

  memcpy (rstate->rs->buf + bi, (*(*data).data).base() + wi, size);

  rstate->rs->rng->add (gs, size);
  WARN (2) << "fill_buf: filled offset " << gs << " count " << size << "\n";

  check_fill (rstate);
}

static void 
got_sindir_block (ptr<read_state> rstate, uint32 blocknum,
                  uint64 gs, uint64 gl, ptr<cfs_data> data)
{
  rstate->outstanding_fetches--;
  if (!data || data->type != CFS_INDIR) {
    rstate->fetch_errors++;
    check_fill (rstate);
    return;
  }
  bigint block = (*(*data).indir).handles[blocknum];
  rstate->outstanding_fetches++;
  ns_cfsrw::retrieve_cfs_readonly
    (block, wrap(got_file_block, rstate, gs, gl));
}

static void 
got_dindir_block (ptr<read_state> rstate, uint32 blocknum, uint32 blockoffst,
                  uint64 gs, uint64 gl, ptr<cfs_data> data)
{
  rstate->outstanding_fetches--;
  if (!data || data->type != CFS_INDIR) {
    rstate->fetch_errors++;
    check_fill (rstate);
    return;    
  }
  bigint block = (*(*data).indir).handles[blocknum];
  rstate->outstanding_fetches++;
  ns_cfsrw::retrieve_cfs_readonly
    (block, wrap (got_sindir_block, rstate, blockoffst, gs, gl));
}

static void 
got_tindir_block (ptr<read_state> rstate, uint32 tblocknum, uint32 dblocknum,
                  uint64 gs, uint64 gl, ptr<cfs_data> data)
{
  rstate->outstanding_fetches--;
  if (!data || data->type != CFS_INDIR) {
    rstate->fetch_errors++;
    check_fill (rstate);
    return;    
  }
  bigint block = (*(*data).indir).handles[dblocknum];
  rstate->outstanding_fetches++;
  ns_cfsrw::retrieve_cfs_readonly
    (block,
     wrap (got_dindir_block, rstate, dblocknum, dblocknum % NFH, gs, gl));
}

static void
fill_buf (ptr<read_state> rstate, cfs_inode_reg *rnode, uint64 gs, uint64 gl)
{
  // fill buf with data starting from gs to gl or end of block, which
  // ever is smaller. commit to ranges and return.
  
  WARN (2) << "fill_buf: gs " << gs << " gl " << gl 
           << " size " << rnode->attr.size << "\n";
  if (gs >= rnode->attr.size) {
    WARN (2) << "snapshot_operation::fill_buf: gap starts after file ends.\n";
    return;
  }
  bigint n;
  uint32 blocknum = gs / SNAPSHOT_BLOCKSIZE;
  if (blocknum < CFS_NDIR) {
    n = rnode->direct[blocknum];
    rstate->outstanding_fetches++;
    ns_cfsrw::retrieve_cfs_readonly
      (n, wrap(got_file_block, rstate, gs, gl));
  }
  else if ((blocknum-CFS_NDIR) < NFH) {
    n = rnode->indirect;
    rstate->outstanding_fetches++;
    ns_cfsrw::retrieve_cfs_readonly
      (n, wrap (got_sindir_block, rstate, blocknum - CFS_NDIR, gs, gl));
  }
  else if ((blocknum-CFS_NDIR-NFH) < NFH*NFH) {
    n = rnode->double_indirect;
    uint32 dblock = (blocknum-CFS_NDIR-NFH);
    rstate->outstanding_fetches++;
    ns_cfsrw::retrieve_cfs_readonly
      (n, wrap (got_dindir_block, rstate, dblock / NFH, dblock % NFH, gs, gl));
  }
  else if ((blocknum-CFS_NDIR-NFH-NFH*NFH) < NFH*NFH*NFH) {
    n = rnode->triple_indirect;
    uint32 tblock = (blocknum - CFS_NDIR - NFH - NFH*NFH);
    uint32 toffst = tblock % (NFH*NFH);
    rstate->outstanding_fetches++;
    ns_cfsrw::retrieve_cfs_readonly
      (n, wrap (got_tindir_block, rstate, tblock / (NFH*NFH), toffst, gs, gl));
  }
  else
    warn << "fill_buf: read offset way too big! \n";
}

static void
read_file_cb2 (ptr<read_state> rstate, ptr<cfs_data> data)
{
  if (!data || data->type != CFS_INODE) {
    rstate->rcb (rstate->rs, CFSRWERR_IO);
    return;
  }
  
  ptr<cfsrw_states::read> rs = rstate->rs;
  cfsrw_states::cb_read rcb = rstate->rcb;

  cfs_inode *inode = (*data).inode;
  if (inode->type == CFSLNK) {
    cfs_inode_lnk *lnode = inode->lnk;
    if (!rs->rng->check (rs->offset, lnode->dest.len ())) {
      memcpy (rs->buf, lnode->dest.cstr (), lnode->dest.len ());
      rs->rcount = lnode->dest.len ();
      rs->eof = true;
      rcb(rs, CFSRW_OK);
    }
  }
  else if (inode->type == CFSREG) {
    cfs_inode_reg *rnode = inode->reg;
    WARN(2) << "read snapshot " << rs->offset << "+" << rs->count << "\n";
    if (rs->rng->check(rs->offset, rs->count)) {
      // got all the bytes, in here to check if we hit EOF or not
      uint64 c = rs->rng->end() - rs->offset;
      if(c > rs->count)
	c = rs->count;
      bool eof = (rs->offset+c >= rnode->attr.size);
      rs->rcount = c;
      rs->eof = eof;
      rcb(rs, CFSRW_OK);
    }
    else {
      // we only want to read what was asked, so bytes before
      // rs->offset are done, and bytes after rs->offset+rs->count,
      // til end of the version of the file stored in snapshot, are
      // done. also, by marking everything from rs->offset+rs->count
      // to eof as done, later on when we test for eof, we know if we
      // have hit eof or not.
      if (rs->offset > 0)
        rs->rng->add (0, rs->offset);
      if ((rs->offset+rs->count) < rnode->attr.size)
        rs->rng->add
	  (rs->offset+rs->count,rnode->attr.size-(rs->offset+rs->count));
      // if requested is more than file size, mark file size to
      // requested size as done. remember that a setattr log record
      // could have expanded the file size before scanning hit the
      // snapshot.
      if ((rs->offset+rs->count) > rnode->attr.size &&
	  rs->offset < rnode->attr.size)
	rs->rng->add(rnode->attr.size, rs->offset+rs->count);
      rs->rng->gaps (wrap (fill_buf, rstate, rnode));
      check_fill (rstate);
    }
  }
}

void
snapshot_operation::read (ptr<snapshot_obj> s, ptr<cfsrw_states::read> rs,
                          cfsrw_states::cb_read rcb)
{
  ptr<dbrec> q = id2dbrec (rs->object);
  ptr<dbrec> v = obj2impl (s)->fmap_db->lookup (q);

  if (!v) {
    rcb (rs, CFSRWERR_IO);
    return;
  }
  ptr<read_state> rstate = New refcounted<read_state> (obj2impl (s));
  rstate->rs = rs;
  rstate->rcb = rcb;
  ns_cfsrw::retrieve_cfs_readonly
    (dbrec2id(v), wrap (read_file_cb2, rstate));
}

