/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "dbfe.h"
#include "snapshot.h"
#include "cfsrw_rec.h"

#define WARN(x) if (sn_verbose >= (x)) warn
int sn_verbose = 0;

#define SNAPSHOT_DBLKSIZE 256 // number of entries per directory block

aiod* snapshot::a = New aiod (2);


snapshot::snapshot (bigint addr, cb_snapshot cb)
  : id (addr), prev_snapshot (0), prev_image (0), bcb(cb)
{
  insert_failures = 0;
  outstanding_inserts = 0;
  outstanding_treeops = 0;
  inserts = 0;
  fmap_db = 0;
  tree_build_started = false;
  ns::retrieve_block
    (addr, wrap (mkref(this), &snapshot::get_fmap_cb, addr));
}

void 
snapshot::get_fmap_cb (bigint addr, ptr<ns_block> blk) 
{
  if (!blk) {
    warn << "snapshot::get_fmap_cb: can't find snapshot block.\n";
    bcb (0, (ptr<snapshot>)0);
    return;
  }
  parse_fmap (true, str (blk->data, blk->len), addr);
}

void 
snapshot::get_more_fmap_cb (bigint addr, ptr<ns_block> blk) 
{
  if (!blk) {
    warn << "snapshot::get_more_fmap_cb: can't find snapshot block.\n";
    bcb (0, (ptr<snapshot>)0);
    return;
  }
  parse_fmap (false, str (blk->data, blk->len), addr);
}

void
snapshot::parse_fmap (bool first, str ins, bigint addr)
{
  xin *in = New xin (ins);

  if (first) {
    del (); // clear all vectors first
    int n;
    in->rbigint (prev_snapshot);
    in->rbigint (rootdir);
    in->rint (n);
    for (int i=0; i<n; i++) {
      bigint h, l;
      int s;
      in->rbigint (h);
      in->rbigint (l);
      in->rint (s);
      _version.add(h, l, s);
    }
  
    if (!(fmap_db = fmap_createdb ())) {
      warn << "cannot create fmap database.\n";
      bcb (0, (ptr<snapshot>)0);
      return;
    }
    fmap_load_done = false;
  }

  int n;
  int remain;
  bigint next;

  in->rint (n);
  in->rint (remain);
  in->rbigint (next);

  for (int i=0; i<n; i++) {
    bigint inum, chash;
    in->rbigint (inum);
    in->rbigint (chash);

    ptr<dbrec> k = id2dbrec (inum);
    ptr<dbrec> d = id2dbrec (chash);
    if (fmap_db->insert (k, d)) {
      fmap_load_done = true;
      bcb (0, (ptr<snapshot>)0);
      delete in;
      return;
    }
  }
  delete in;

  if (remain > 0) {
    WARN(1) << "load extra fmap page at " << next << "\n";
    ns::retrieve_block
      (next, wrap (mkref(this), &snapshot::get_more_fmap_cb, addr));
  }
  else {
    fmap_load_done = true;
    bcb (addr, mkref(this));
  }
}

snapshot::snapshot
(bigint root, vervec start, bigint prev, ptr<snapshot> image, cb_snapshot cb)
  : bcb(cb)
{
  insert_failures = 0;
  outstanding_inserts = 0;
  outstanding_treeops = 0;
  inserts = 0;
  fmap_db = 0;
  id = 0;
  tree_build_started = false;
   
  if (image)
    rootdir = image->rootdir;
  else
    rootdir = root;

  if (!(fmap_db = fmap_createdb ())) {
    warn << "cannot create fmap database.\n";
    timecb (tsnow, wrap (mkref(this), &snapshot::snapshot_failed));
    return;
  }
  fmap_load_done = false;

  if (image) {
    prev_snapshot = prev;
    prev_image = image;

    ptr<dbEnumeration> it = image->fmap_db->enumerate();
    ptr<dbPair> p;
    while (p = it->nextElement()) {
      ptr<dbrec> v = image->fmap_db->lookup (p->key);
      bigint inum = dbrec2id(p->key);
      bigint chash = dbrec2id(v);
      ptr<dbrec> k = id2dbrec (inum);
      ptr<dbrec> d = id2dbrec (chash);
      if (fmap_db->insert (k, d)) {
	fmap_load_done = true;
	bcb (0, (ptr<snapshot>)0);
	return;
      }
    }
    fmap_load_done = true;
    build_from_prev (image, start);
  }
  else {
    prev_snapshot = 0;
    prev_image = 0;
    build_from_scratch (start);
  }
}

void
snapshot::snapshot_failed ()
{
  bcb (0, (ptr<snapshot>)0);
}

ptr<dbfe>
snapshot::fmap_createdb()
{
  assert(fmap_db == 0);

  ptr<dbfe> db = New refcounted<dbfe>();
  dbOptions opts;
  opts.addOption("opt_async", 1);
  opts.addOption("opt_flag", DB_CREATE | DB_TRUNCATE);

  // don't supply a path: sleepycat DB3 is not async. to avoid
  // deadlock we need to keep the DB in-memory. furthermore, don't
  // want to corrupt the fmap_db of other snapshots.

  if (int err = db->opendb(0L, opts)) {
    warn << "open returned: " << strerror(err) << " (" << err << ")\n";
    return 0;
  }

  return db;
}

void 
snapshot::del ()
{
  active_snodes.deleteall ();
  log_records.clear ();
  if (fmap_db)
    fmap_db = 0;
  _version.clear ();
}

void 
snapshot::build_from_scratch (vervec start)
{
  vervec stop;
  vNew refcounted<logiter>
    (wrap(mkref(this), &snapshot::scan_log_records, start, stop), start, stop);
}

void
snapshot::build_from_prev (ptr<snapshot> ps, vervec start)
{
  vNew refcounted<logiter>
    (wrap (mkref(this), &snapshot::scan_log_records, start, ps->version ()),
     start, ps->version ());
}

void
snapshot::load_earlier_snapshot (vervec start, bigint addr, ptr<snapshot_obj> s)
{
  if (s == 0) {
    // can't load earlier snapshot
    bcb (0, (ptr<snapshot>)0);
    return;
  }

  snapshot_operation::build_snapshot (s, addr, start, bcb);
}

void 
snapshot::scan_log_records (vervec start, vervec stop, iterinfo *li)
{
  if (!li || (li->status != LOG_OK && li->status != LOGERR_OUTOFORDER)) {
    warn << "snapshot: cannot scan logs\n";
    bcb (0, (ptr<snapshot>)0);
    return;
  }

  if (li->status == LOGERR_OUTOFORDER) {
    warn << "snapshot: out of order iteration of log records\n";
    li->finished = true;
    assert (prev_snapshot > 0 && prev_image);
    
    // the correct thing to do here is to skip the prev_snapshot. so
    // we should build a new snapshot from an even earlier snapshot.
    // we should probably build it incrementally (i.e. build a
    // slightly earlier snapshot from even earlier snapshot, then
    // build the current snapshot), may be we will do that later.

    if (prev_image->prev () != 0)
      ptr<snapshot> nps = New refcounted<snapshot>
        (prev_image->prev (),
	 wrap (mkref(this), &snapshot::load_earlier_snapshot, start));
    else
      // build from scratch
      ptr<snapshot> ns = New refcounted<snapshot>
	(rootdir, start, 0, (ptr<snapshot>) 0, bcb);
    return;
  }

  if (!li->finished) {
    bool fini = false;
    if (prev_snapshot > 0) {
      for (uint j=0; j<stop.size(); j++) {
        if (li->record->head () == stop.head(j)) {
          if (li->record->seqnum () == stop.seqnum(j)) {
            li->eol = true;
            fini = true;
          }
          break;
        }
      }
    }
    if (!fini) {
      WARN(2) << "head " << li->record->head () << "\n";
      WARN(2) << "this " << li->record->hash () << "\n";
      WARN(2) << "seqn " << li->record->seqnum () << "\n";
      WARN(2) << "type " << li->record->record_type () << "\n";
      log_records.push_back (li->record);
    }
  }
  
  else if (!tree_build_started) {
    tree_build_started = true;
    warn << "snapshot: " << log_records.size() << " records to process\n";
    _version = li->iter.version ();
    outstanding_treeops = 0;
    inode_stored = 0;
    treeop_failures = 0;
    tree_stored = false;
    fs_tree();
  }
}

void 
snapshot::fs_tree ()
{
  assert (outstanding_treeops == 0);
  if (outstanding_treeops == 0)
    timecb (tsnow, wrap (mkref(this), &snapshot::build_tree));
}

void
snapshot::build_tree ()
{
  // build a filesystem tree one operation at a time: avoid concurrent
  // building so we don't have to use locks! this may be slow...

  if (!log_records.empty() && !treeop_failures) {
    ptr<log_record> e = log_records.pop_back ();
    WARN(2) << "processing log type: " << e->record_type () << "\n";
    switch (e->record_type ()) {
    case cfsrw_log_record::T_INODE:
      {
        inode_log_record *ie = e->template record<inode_log_record> ();
        ptr<cfs_data> sdata = New refcounted<cfs_data> (CFS_INODE);
        setinode (e, sdata->inode);
        update_index (ie->object (), sdata);
	if (outstanding_treeops == 0)
	  fs_tree();
      }
      break;
    case log_record::T_HEAD:
      {
        head_log_record *hle = e->template record<head_log_record>();
        rootdir = hle->root ();
	fs_tree();
      }
      break;
    case cfsrw_log_record::T_LINK:
      {
        link_log_record *le = e->template record<link_log_record>();
        do_link (le->dir (), le->name (), e);
      }
      break;
    case cfsrw_log_record::T_UNLINK:
      {
        unlink_log_record *ue = e->template record<unlink_log_record>();
        do_unlink (0, ue->dir (), ue->name (), e);
      }      
      break;
    case cfsrw_log_record::T_WRITE:
      do_write (e);
      break;
    case cfsrw_log_record::T_ATTR:
      update_attr (e);
      break;
    case cfsrw_log_record::T_RENAME:
      do_rename (e);
      break;

    case log_record::T_END:
    case cfsrw_log_record::T_INTEND:
    default:
      fs_tree();
      break;
    }
  }
  else if (!treeop_failures && outstanding_treeops == 0 && 
           log_records.empty() && !tree_stored) {
    tree_stored = true;
    WARN(2) << "store tree\n";
    store_tree();
  }
  else if (treeop_failures) {
    warn << "snapshot built failed! dropping this snapshot.\n";
    // for debugging
    // fatal << "snapshot failed.\n";
    bcb (0, (ptr<snapshot>)0);
  }
}

void
snapshot::store_tree()
{
  if (outstanding_treeops == 0) {
    if (prev_snapshot == 0) {
      snode *se = active_snodes [rootdir];
      assert (se);
      se->ent->inode->reg->nlink++; // rootdir nlink is always 1
      store (rootdir);
    }
    else if (active_snodes.size() > 0) {
      active_snodes.traverse (wrap (mkref(this), &snapshot::store_new));
      if (outstanding_inserts == 0)
        insert_snapshot ();
    }
    else
      insert_snapshot();
  }
}

void 
snapshot::setinode (ptr<log_record> e, cfs_inode *inode)
{
  inode_log_record *ie = e->template record<inode_log_record>();
  strbuf w;
  sbigint (w, ie->object ());
  WARN(2) << "setinode: inode type: " << ie->object_type ()
          << " inum: " << armor64A(w) << "\n";

  switch (ie->object_type ()) {
  case CFSRW_DIR:    // ignore dir opaque case
  case CFSRW_REG:
    {
      ftypero t =
	(ie->object_type () == CFSRW_DIR) ? CFSDIR : CFSREG;
      inode->set_type (t); 
      rpc_clear (*inode->reg);
      inode->reg->path = armor64A(w);
      inode->reg->nlink = 0;
      
      init_attr (inode->reg);
      update_ctime (inode->reg, ie->timestamp ());
      update_mtime (inode->reg, ie->timestamp ());
      update_mver (inode->reg, e);
      set_attr (inode->reg, ie->attr ());

      inode->reg->direct.setsize (0);
      if (t == CFSREG)
        write_file (e, inode->reg);
    }
    break;
  case CFSRW_LNK:
    {
      inode->set_type (CFSLNK);
      rpc_clear (*inode->lnk);
      inode->lnk->path = armor64A(w);
      inode->lnk->nlink = 0;
  
      init_attr (inode->lnk);
      update_ctime (inode->lnk, ie->timestamp ());
      update_mver (inode->lnk, e);
      set_attr (inode->lnk, ie->attr ());
      inode->lnk->dest = ""; // will be set to a path later by the write logent
    }
    break;
  default:
    warn << "snapshot::setinode: strange inode type.\n";
  }
}

void
snapshot::lookup_fmap (bigint inum, cb_snode cb)
{
  ptr<dbrec> q = id2dbrec(inum);
  ptr<dbrec> v = fmap_db->lookup(q);
  bigint chash = inum;
  if (v)
    chash = dbrec2id (v);
  ns_cfsrw::retrieve_cfs
    (chash, wrap (mkref(this), &snapshot::lookup_fmap_load_block_cb, cb, inum));
}

void
snapshot::lookup_fmap_load_block_cb (cb_snode cb, bigint inum,
                                     ptr<cfs_data> sdata)
{
  if (sdata) {
    snode *e = update_index (inum, sdata);
    cb (e);
  }
  else
    cb ((snode*)0);
}

void 
snapshot::do_link (bigint dir, str name, ptr<log_record> eptr)
{
  outstanding_treeops++;
  WARN(2) << "do_link: dir inum: " << dir << " file name: " << name << "\n";
  snode *d = active_snodes[dir];
  if (!d) {
    lookup_fmap
      (dir, wrap(mkref(this), &snapshot::do_link_got_dir, name, eptr));
  }
  else
    do_link_got_dir(name, eptr, d);
}

void
snapshot::do_link_got_dir (str name, ptr<log_record> eptr, snode *d)
{
  if (!d || !d->ent || d->ent->type != CFS_INODE ||
      d->ent->inode->type != CFSDIR) {
    warn << "snapshot::do_link, no dir snode\n";
    treeop_failures++;
    outstanding_treeops--;
    fs_tree();
    return;
  }

  assert (dirop_record (eptr));
  snode *e = active_snodes[dirop_record (eptr)->object ()];

  if (!e) {
    lookup_fmap
      (dirop_record (eptr)->object (),
       wrap(mkref(this), &snapshot::do_link_got_node, name, eptr, d));
  }
  else
    do_link_got_node(name, eptr, d, e);
}

void
snapshot::do_link_got_node
(str name, ptr<log_record> eptr, snode *d, snode *e)
{
  if (!e || !e->ent || e->ent->type != CFS_INODE) {
    warn << "snapshot::do_link: node does not exist.\n";
    treeop_failures++;
    outstanding_treeops--;
    fs_tree();
    return;
  }

  if (e->ent->inode->type == CFSDIR)
    // linking in a directory, set that directory's ".." reference
    e->ent->inode->reg->parent = d->inum;

  cfs_inode_reg *dnode = d->ent->inode->reg;    
  bigint dirblk;
  
  if (dnode->direct.size () == 0) {
    // create a new block for this directory
    ptr<cfs_data> dir = New refcounted<cfs_data>(CFS_DIRBLK);
    getrnd (&dirblk, CFS_FHSIZE);
    update_index (dirblk, dir);
    dnode->direct.setsize (1);
    dnode->direct[0] = dirblk;
    dir->dir->eof = true;
    dir->dir->count = 0;
    add_dirent (name, eptr, dnode, dir, e);
    outstanding_treeops--;
    fs_tree();
    return;
  }

  else {
    int blk = dnode->direct.size ()-1;
    dirblk = dnode->direct[blk];

    WARN(1) << "do_link: searching dnode->direct[" << blk << "] = "
            << dirblk << "\n";

    snode *dbe = active_snodes[dirblk];
    if (dbe)
      do_link_got_dirblk(name, eptr, dirblk, d, e, dbe);
    else {
      lookup_fmap
	(dirblk,
	 wrap(mkref(this), &snapshot::do_link_got_dirblk, name, eptr,
	      dirblk, d, e));
    }
  }
}

void
snapshot::do_link_got_dirblk
(str name, ptr<log_record> eptr, bigint dirblk,
 snode *d, snode *e, snode *dbe)
{
  if (!dbe || dbe->ent->type != CFS_DIRBLK) {
    warn << "snapshot::do_link: dirblk does not exist.\n";
    treeop_failures++;
    outstanding_treeops--;
    fs_tree();
    return;
  }
  
  ptr<cfs_data> dir = dbe->ent;
  cfs_inode_reg *dnode = d->ent->inode->reg;

  add_dirent (name, eptr, dnode, dir, e);
  outstanding_treeops--;
  fs_tree();
}

void
snapshot::add_dirent (str name, ptr<log_record> e, cfs_inode_reg *dnode,
                      ptr<cfs_data> dir, snode *fe)
{
  // directory size is bytes needed to store directory, so we set that
  // before we insert directory block into dhash, not here.

  if (dir->dir->count == SNAPSHOT_DBLKSIZE) {
    warn << "snapshot directory size limit reached!!!\n";
    treeop_failures++;
    return;
  }

  bigint inum;
  assert (dirop_record (e));
  update_ctime (dnode, dirop_record (e)->timestamp ());
  update_mver (dnode, e);
  inum = dirop_record (e)->object ();

  rpc_ptr<cfs_dirent> direntp;
  direntp.alloc ();
  direntp->fh = inum;
  WARN(2) << "adding to dirent " << name << ", inum " 
          << inum << "\n";
  direntp->name = name;

  if (dir->dir->entries)
    direntp->nextentry = dir->dir->entries;
  dir->dir->entries = direntp;
  dir->dir->count++;
    
  if (dir->dir->count == SNAPSHOT_DBLKSIZE) {
    int blk = dnode->direct.size ();
    if (blk < CFS_NDIR) {
      dir->dir->eof = false;
      // create new directory block
      bigint dirblk;
      ptr<cfs_data> newdir = New refcounted<cfs_data>(CFS_DIRBLK);
      getrnd (&dirblk, CFS_FHSIZE);
      update_index (dirblk, newdir);
      newdir->dir->eof = true;
      newdir->dir->count = 0;
      WARN(1) << "created new dir blk (" << blk << ") " << dirblk << "\n";

      rpc_vec<bigint, CFS_NDIR> dv = dnode->direct;
      dnode->direct.setsize (blk+1);
      for (int j=0; j<blk; j++) dnode->direct[j] = dv[j];
      dnode->direct[blk] = dirblk;
    }
  }

  cfs_inode *inode = (fe->ent)->inode;
  if (inode->type == CFSLNK) {
    inode->lnk->nlink++;
    update_ctime (inode->lnk, e->timestamp ());
    update_mver (inode->lnk, e);
  }
  else {
    inode->reg->nlink++;
    update_ctime (inode->reg, e->timestamp ());
    update_mver (inode->reg, e);
    if (inode->type == CFSDIR) // take into account of ..
      dnode->nlink++;
  }
}

snode * 
snapshot::update_index (bigint inum, ptr<cfs_data> sdata)
{
  snode *se = New snode (inum, sdata);
  active_snodes.insert (se);
  WARN(2) << "snapshot::update_index: inum " << inum << "\n";
  return se;
}

void 
snapshot::do_write (ptr<log_record> e)
{
  outstanding_treeops++;
  write_log_record *we = e->template record<write_log_record>();
  snode *f = active_snodes[we->file()];
  if (f)
    do_write_got_node (e, f);
  else
    lookup_fmap
      (we->file(), wrap(mkref(this), &snapshot::do_write_got_node, e));
}

void
snapshot::do_write_got_node (ptr<log_record> e, snode *f)
{
  if (!f || f->ent->type != CFS_INODE) {
    warn << "snapshot::do_write: file inode does not exist.\n";
    treeop_failures++;
    outstanding_treeops--;
    fs_tree();
    return;
  }

  cfs_inode *inode = (f->ent)->inode;
  if (inode->type == CFSLNK) {
    do_symlink (e, inode->lnk);
    outstanding_treeops--;
    fs_tree();
  }
  else if (inode->type == CFSREG || inode->type == CFSREG_EXEC) {
    outstanding_treeops--;
    write_file (e, inode->reg);
  }
  else {
    warn << "do_write_got_node: wrong file type!\n";
    treeop_failures++;
    outstanding_treeops--;
    fs_tree();
  }
}

void 
snapshot::do_symlink (ptr<log_record> e, cfs_inode_lnk *lnode)
{
  write_log_record *we = e->template record<write_log_record>();
  if (we->count () > MAXPATHLEN) {
    warn << "snapshot::do_symlink: target too large.\n";
    treeop_failures++;
    return;
  }
  lnode->dest = nfspath3 (we->data ());
  update_ctime (lnode, we->timestamp ());
  update_mver (lnode, e);
}

void 
snapshot::write_file (ptr<log_record> e, cfs_inode_reg *rnode)
{
  outstanding_treeops++;
  str fname = get_filename (rnode->path);
  snapshot::a->open (fname, O_WRONLY | O_CREAT, 0666,
                     wrap (mkref(this), &snapshot::write_file_open_cb,
		           e, rnode));
}

void
snapshot::write_file_open_cb (ptr<log_record> e, cfs_inode_reg *rnode,
                              ptr<aiofh> fh, int err)
{
  if (!fh) {
    warn << "write_file: open failed: " << err << "\n";
    treeop_failures++;
    outstanding_treeops--;
    fs_tree();
    return;
  }

  write_log_record *we = e->template record<write_log_record> ();
  inode_log_record *ie = e->template record<inode_log_record> ();
  attr_log_record  *ae = e->template record<attr_log_record> ();

  if (we) {
    ptr<aiobuf> buf = snapshot::a->bufalloc (we->count ());
    if (!buf) {
      // warn << "no buf, bufwait\n";
      snapshot::a->bufwait
	(wrap (mkref(this), &snapshot::write_file_open_cb, e, rnode, fh, 0));
      return;
    }

    // mark region as modified
    if (rnode->mstart == 0 && rnode->mend == 0) {
      rnode->mstart = we->offset ();
      rnode->mend = we->offset () + we->count ();
      WARN(2) << "mstart set to " << rnode->mstart
	      << " and mend set to " << rnode->mend << "\n";
    }
    else {
      if (rnode->mstart > we->offset ())
	rnode->mstart = we->offset ();
      if (rnode->mend < we->offset () + we->count ())
	rnode->mend = we->offset () + we->count ();
      WARN(2) << "mstart set to " << rnode->mstart
	      << " and mend set to " << rnode->mend << "\n";
    }
    // write beyond size of file, mark everything between old size of
    // file to where write started as modified
    if (rnode->attr.size < rnode->mstart)
      rnode->mstart = rnode->attr.size;

    memmove(buf->base (), we->data (), we->count ());
    WARN(2) << "calling aiod::write at offset " << we->offset() << "\n";
    fh->write (we->offset (), buf,
	       wrap (mkref(this), &snapshot::write_file_write_cb,
		     e, rnode, fh));
    return;
  }
  else if (ie || ae) {
    if ((ie && ie->attr ()->valid_size) || (ae && ae->attr ()->valid_size)) {
      uint64 size = ie ? ie->attr ()->size : ae->attr ()->size;

      // mark the whole file as modified in case the setattr increased
      // size of the file
      rnode->mstart = 0;
      rnode->mend = size;
      WARN(2) << "mstart set to " << rnode->mstart
	      << " and mend set to " << rnode->mend << "\n";

      WARN(2) << "aiod::truncate file to " << size << " bytes\n";
      fh->ftrunc (size, wrap (mkref(this), &snapshot::write_file_trunc_cb,
	                      e, rnode, fh));
    }
    else
      write_file_write_cb (e, rnode, fh, 0, 0, 0); // close file
    return;
  }
  assert (0);
}

void
snapshot::write_file_trunc_cb (ptr<log_record> e, cfs_inode_reg *rnode,
                               ptr<aiofh> fh, int err)
{
  if (err) {
    warn << "write_file: trunc failed: " << err << "\n";
    treeop_failures++;
  }
  fh->close (wrap (mkref(this), &snapshot::write_file_close_cb));
}

void
snapshot::write_file_write_cb (ptr<log_record> e, cfs_inode_reg *rnode,
                               ptr<aiofh> fh, ptr<aiobuf> buf,
			       ssize_t sz, int err)
{
  write_log_record *we = e->template record<write_log_record>();
  if (we) {
    if (err || ((unsigned) sz != we->count ())) {
      warn << "write_file: write failed: " << err << "\n";
      treeop_failures++;
    }
    update_mtime (rnode, we->timestamp ());
    update_ctime (rnode, we->timestamp ());
    update_mver (rnode, e);

    if (rnode->attr.size < (we->offset () + we->count ()))
      rnode->attr.size = we->offset () + we->count ();
  }
  fh->close (wrap (mkref(this), &snapshot::write_file_close_cb));
}

void
snapshot::write_file_close_cb (int err)
{
  if (err) {
    warn << "write_file: close failed: " << err << "\n";
    treeop_failures++;
  }
  outstanding_treeops--;
  fs_tree();
}

void 
snapshot::do_unlink(int phase, bigint dir, str name, ptr<log_record> e)
{
  outstanding_treeops++;
  snode *d = active_snodes[dir];
  if (d)
    do_unlink_got_dir (phase, name, e, d);
  else {
    lookup_fmap
      (dir, wrap(mkref(this), &snapshot::do_unlink_got_dir, phase, name, e));
  }
}

void
snapshot::do_unlink_got_dir (int phase, str name, ptr<log_record> e, snode* d)
{
  if (!d || d->ent->inode->type != CFSDIR) {
    warn << "snapshot::do_unlink: no CFSDIR entry.\n";
    treeop_failures++;
    outstanding_treeops--;
    fs_tree();
    return;
  }
  cfs_inode_reg *dnode = d->ent->inode->reg;
  if (dnode->direct.size () == 0) {
    done_unlink(phase, e, true);
    return;
  } 
  
  bigint dirblk = dnode->direct[0];
  snode *dbe = active_snodes[dirblk];
  if (dbe)
    do_unlink_got_dirblk (phase, name, e, 0, d, dbe);
  else
    lookup_fmap
      (dirblk, wrap(mkref(this), &snapshot::do_unlink_got_dirblk,
		    phase, name, e, 0, d));
}

void
snapshot::do_unlink_got_dirblk (int phase, str name, ptr<log_record> e,
                                int blknum, snode *d, snode *dbe)
{
  if (!dbe || dbe->ent->type != CFS_DIRBLK) {
    warn << "snapshot::do_unlink: dirblk does not exist.\n";
    treeop_failures++;
    outstanding_treeops--;
    fs_tree();
    return;
  }
  cfs_data *dir = dbe->ent;
  remove_dirent (phase, name, e, d, dir, blknum);
}

void 
snapshot::remove_dirent (int phase, str name, ptr<log_record> e,
                         snode *d, cfs_data *dir, uint32 blknum)
{
  cfs_inode_reg *dnode = d->ent->inode->reg;
  assert(dirop_record (e));
  update_ctime (dnode, e->timestamp ());
  update_mver (dnode, e);

  bigint fh;
  cfs_dirent* de = dir->dir->entries;
  if (de && (de->name == name)) {
    fh = de->fh;
    dir->dir->entries = dir->dir->entries->nextentry;
  }
  else {
    while (de && de->nextentry && (de->nextentry->name != name))
      de = de->nextentry;
    if (!de || !de->nextentry) {
      if (blknum == dnode->direct.size ()-1)
        done_unlink(phase, e, true);
      else {
        blknum++;
	bigint dirblk = dnode->direct[blknum];
        snode *dbe = active_snodes[dirblk];
        if (dbe)
          do_unlink_got_dirblk (phase, name, e, blknum, d, dbe);
        else
          lookup_fmap
            (dirblk, wrap(mkref(this), &snapshot::do_unlink_got_dirblk,
		          phase, name, e, blknum, d));
      }
      return;
    }
    fh = de->nextentry->fh;
    de->nextentry = de->nextentry->nextentry;
  }
  assert(dir->dir->count > 0);
  dir->dir->count--;

  snode *ie = active_snodes[fh];
  if (ie)
    remove_dirent_got_fh (phase, fh, e, dnode, ie);
  else {
    lookup_fmap
      (fh, wrap(mkref(this), &snapshot::remove_dirent_got_fh,
		phase, fh, e, dnode));
  }
}

void
snapshot::remove_dirent_got_fh (int phase, bigint fh, ptr<log_record> e,
                                cfs_inode_reg *dnode, snode *ie)
{
  if (!ie || ie->ent->type != CFS_INODE) {
    warn << "snapshot::remove_dirent: entry is not type INODE.\n";
    treeop_failures++;
  }
  else {
    cfs_inode *inode = (ie->ent)->inode;
    if (inode->type == CFSLNK) {
      (*(*inode).lnk).nlink--;
      update_ctime (inode->lnk, e->timestamp ());
      update_mver (inode->lnk, e);
    }
    else {
      (*(*inode).reg).nlink--;
      update_ctime (inode->reg, e->timestamp ());
      update_mver (inode->reg, e);
      if (inode->type == CFSDIR) // take into account of ..
        dnode->nlink--;
    }
  }
  done_unlink(phase, e);
}

void
snapshot::done_unlink(int phase, ptr<log_record> e, bool error)
{
  outstanding_treeops--;
  rename_log_record *re = e->template record<rename_log_record>();  
  if (re) {
    if (phase == 0)
      do_unlink (1, re->from_dir(), re->from_name(), e);
    else
      do_link (re->dir(), re->name (), e);
  }
  else {
    if (error) {
      warn << "unlink failed\n";
      treeop_failures++;
    }
    fs_tree();
  }
}

void 
snapshot::update_attr (ptr<log_record> eptr)
{
  outstanding_treeops++;
  attr_log_record *ae = eptr->template record<attr_log_record>();
  snode *e = active_snodes[ae->object ()];
  if (e)
    update_attr_got_node (eptr, e);
  else {
    lookup_fmap
      (ae->object(), wrap(mkref(this), &snapshot::update_attr_got_node, eptr));
  }
}

void
snapshot::update_attr_got_node (ptr<log_record> eptr, snode *e)
{
  attr_log_record *ae = eptr->template record<attr_log_record>();
  if (!e || e->ent->type != CFS_INODE) {
    warn << "snapshot::update_attr: type is not INODE.\n";
    treeop_failures++;
    outstanding_treeops--;
    fs_tree ();
  }
  else {
    if (e->ent->inode->type == CFSLNK) {
      update_ctime (e->ent->inode->lnk, ae->timestamp ());
      update_mver (e->ent->inode->lnk, eptr);
      set_attr (e->ent->inode->lnk, ae->attr ());
      outstanding_treeops--;
      fs_tree ();
    }
    else {
      update_ctime (e->ent->inode->reg, ae->timestamp ());
      if(ae->attr()->valid_size)
        update_mtime (e->ent->inode->reg, ae->timestamp ());
      update_mver (e->ent->inode->reg, eptr);
      set_attr (e->ent->inode->reg, ae->attr ());
      if (e->ent->inode->type == CFSREG) {
        outstanding_treeops--;
        write_file (eptr, e->ent->inode->reg);
      }
      else {
        outstanding_treeops--;
        fs_tree ();
      }
    }
  }
}

void 
snapshot::do_rename (ptr<log_record> e)
{
  rename_log_record *re = e->template record<rename_log_record>();  
  do_unlink (0, re->dir(), re->name(), e);
}

