/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef CFSRW_SNAPSHOT_H
#define CFSRW_SNAPSHOT_H

#include "aiod.h"
#include "vec.h"
#include "ihash.h"
#include "definc.h"

#include "record.h"
#include "view.h"
#include "logiter.h"
#include "pack.h"
#include "vervec.h"

#include "ns_cfsrw.h"
#include "cfsrw.h"
#include "cfs.h"
#include "ranges.h"
#include "snapshot_spec.h"


#define SNAPSHOT_BLOCKSIZE 8192
#define NFH ((SNAPSHOT_BLOCKSIZE-100) / (sha1::hashsize*2))
#define FMAP_ENTRIES_PER_PAGE NFH

#define WARN(x) if (sn_verbose >= (x)) warn
extern int sn_verbose;

typedef struct snode {
  bigint inum;
  ptr<cfs_data> ent;
  ihash_entry<snode> slink;
  snode (bigint, ptr<cfs_data>);
  ~snode ();
} snode;

inline
snode::snode (bigint i, ptr<cfs_data> e) : 
  inum (i), ent (e)
{
}

inline
snode::~snode ()
{
  ent = 0;
}

class snapshot;

typedef callback<void, snode*>::ptr cb_snode;

typedef ihash<bigint, snode, &snode::inum, &snode::slink> snode_hash_t;

class xin;
class dbfe;
class dbrec;

class snapshot : public virtual refcount, public snapshot_obj {
private:
  friend class snapshot_operation;
  static aiod* a;

  bigint id; // content hash of this snapshot.
  bigint prev_snapshot;
  ptr<snapshot> prev_image;

  vervec _version;
  bigint rootdir;

  bool tree_build_started;
  bool tree_stored;
  int inode_stored;
  int outstanding_treeops;
  int treeop_failures;

  int outstanding_inserts;
  int insert_failures;
  int inserts;

  vec< ptr<log_record> > log_records; // total order log entries
  snode_hash_t active_snodes;  // intermediate fs tree in cfs format
  
  cb_snapshot bcb;
  
  void got_view (cfsrw_status status);
  void build_from_scratch (vervec);
  void build_from_prev (ptr<snapshot>, vervec);
  void scan_log_records (vervec, vervec, iterinfo *);

  void setinode (ptr<log_record>, cfs_inode *);
  
  void do_link (bigint, str, ptr<log_record>);
  void do_link_got_dir (str, ptr<log_record>, snode*);
  void do_link_got_node (str, ptr<log_record>, snode*, snode*);
  void do_link_got_dirblk (str, ptr<log_record>, bigint,
                           snode*, snode*, snode*);
  void add_dirent (str, ptr<log_record>, cfs_inode_reg*,
                   ptr<cfs_data>, snode*);

  void do_write (ptr<log_record>);
  void do_write_got_node (ptr<log_record>, snode*);
  void do_symlink (ptr<log_record>, cfs_inode_lnk *);
  void write_file (ptr<log_record>, cfs_inode_reg *);
  
  void write_file_open_cb (ptr<log_record> e, cfs_inode_reg *rnode,
                           ptr<aiofh> fh, int err);
  void write_file_write_cb (ptr<log_record> e, cfs_inode_reg *rnode,
                            ptr<aiofh> fh, ptr<aiobuf> buf,
			    ssize_t sz, int err);
  void write_file_trunc_cb (ptr<log_record> e, cfs_inode_reg *rnode,
                            ptr<aiofh> fh, int err);
  void write_file_close_cb (int err);

  void do_unlink (int, bigint, str, ptr<log_record>);
  void do_unlink_got_dir (int, str, ptr<log_record>, snode*);
  void do_unlink_got_dirblk (int, str, ptr<log_record>, int, snode*, snode*);
  void remove_dirent (int, str, ptr<log_record>, snode *,
                      cfs_data *, uint32);
  void remove_dirent_got_fh (int, bigint, ptr<log_record>,
                             cfs_inode_reg *, snode *);
  void done_unlink (int, ptr<log_record>, bool err = false);
  
  void update_attr (ptr<log_record>);
  void update_attr_got_node (ptr<log_record>, snode*);

  void do_rename (ptr<log_record>);

  snode *update_index (bigint, ptr<cfs_data>);
  
  void fs_tree ();
  void build_tree ();
  void store_tree ();

  void store (bigint);
  void store_new (snode *);
  bigint store_one_fmap_block (str);

  void insert_and_addtofmap (char *, size_t, bigint, ptr<cfs_data> d = 0);
  void insert (bigint, char *, size_t, ptr<cfs_data> d = 0);

  void insert_cb (bool, bigint, ns_stat, bigint);
  void insert_snapshot ();
  void insert_fmapblock_cb (ns_stat, bigint);
  void store_inode (bigint, cfs_inode *);

  void store_file (bigint, cfs_inode *);
  void store_directory (bigint, cfs_inode *);
  void store_tind (vec<bigint> *fhs, bigint *ti_fh,
                   uint32 off, uint32& blocknum);
  void store_dind (vec<bigint> *fhs, bigint *di_fh,
                   uint32 off, uint32& blocknum);
  void store_sind (vec<bigint> *fhs, bigint *si_fh,
                   uint32 off, uint32& blocknum);
  void store_file_close (bigint inum, cfs_inode *inode,
                         vec<bigint>* fhs, ptr<aiofh> fd);
  void store_file_close_cb (int);
  void store_file_block (cfs_inode *inode, uint32 blocknum,
                         const char *block, ssize_t size, bigint *fh);
  void store_file_read_block (bigint inum, cfs_inode *inode,
                              uint32 blocknum, vec<bigint>* fhs,
			      ptr<aiofh> fd, ptr<aiobuf> buf,
			      ssize_t sz, int err);
  void store_file_open_cb (bigint inum, cfs_inode *inode,
                           ptr<aiofh> fd, int err);

  void get_fmap_cb (bigint, ptr<ns_block> blk);
  void get_more_fmap_cb (bigint, ptr<ns_block> blk);
  void del ();

  void lookup_fmap (bigint, cb_snode);
  void lookup_fmap_load_block_cb (cb_snode, bigint, ptr<cfs_data>);
  void load_earlier_snapshot (vervec, bigint, ptr<snapshot_obj>);

  bool fmap_load_done;
  int fmap_block_inserts;
  int fmap_block_insert_failures;

  void snapshot_failed ();
  
  ptr<dbfe> fmap_createdb ();
  
public:
  // load snapshot from DHash
  snapshot (bigint loc, cb_snapshot cb);
  // creates new snapshot from an old image. if image is 0, create
  // from scratch
  snapshot (bigint root, vervec start,
            bigint prev, ptr<snapshot> image, cb_snapshot cb);

  ptr<dbfe> fmap_db; // file map database

  ~snapshot () { del (); }
  bigint location () const { return id; };
  bigint prev () const { return prev_snapshot; }
  vervec version () const { return _version; }

  static int init ();

private:
  void parse_fmap (bool, str, bigint);
  void serialize_and_store ();
};

inline void
get_mver (vervec &version, rpc_bytes<RPC_INFINITY> &v)
{
  str s (v.base (), v.size ());
  vervec::parse (s, version);
}

inline void 
update_mtime (cfs_inode_reg *inode, unsigned ts)
{
  inode->mtime.seconds = ts;
  inode->mtime.nseconds = 0;
}

inline void 
update_ctime (cfs_inode_reg *inode, unsigned ts)
{
  inode->ctime.seconds = ts;
  inode->ctime.nseconds = 0;
}

inline void
update_mver (cfs_inode_reg *inode, recordp r)
{
  vervec v = r->version_inclusive ();
  strbuf b;
  v.serialize (b);
  str s = b;
  inode->mver = s;
}

inline void 
update_mtime (cfs_inode_lnk *inode, unsigned ts)
{
  inode->mtime.seconds = ts;
  inode->mtime.nseconds = 0;
}

inline void 
update_ctime (cfs_inode_lnk *inode, unsigned ts)
{
  inode->ctime.seconds = ts;
  inode->ctime.nseconds = 0;
}

inline void
update_mver (cfs_inode_lnk *inode, recordp r)
{
  vervec v = r->version_inclusive ();
  strbuf b;
  v.serialize (b);
  str s = b;
  inode->mver = s;
}

inline void
init_attr (cfs_inode_reg *inode)
{
  inode->mtime.seconds = 0;
  inode->mtime.nseconds = 0;
  inode->attr.size = 0;
  inode->attr.mode = 0;
  inode->attr.uid = 0;
  inode->attr.gid = 0;
  inode->attr.rdev.major = 0;
  inode->attr.rdev.minor = 0;
}

inline void
init_attr (cfs_inode_lnk *inode)
{
  inode->mtime.seconds = 0;
  inode->mtime.nseconds = 0;
  inode->attr.size = 0;
  inode->attr.mode = 0;
  inode->attr.uid = 0;
  inode->attr.gid = 0;
  inode->attr.rdev.major = 0;
  inode->attr.rdev.minor = 0;
}

inline void
set_attr (cfs_inode_reg *inode, cfsrw_sattr *ca)
{
  if (ca->valid_mtime) {
    inode->mtime.seconds = ca->mtime;
    inode->mtime.nseconds = 0;
  }
  if (ca->valid_size)
    inode->attr.size = ca->size;
  if (ca->valid_mode)
    inode->attr.mode = ca->mode;
  if (ca->valid_uid)
    inode->attr.uid = ca->uid;
  if (ca->valid_gid)
    inode->attr.gid = ca->gid;
}

inline void
set_attr (cfs_inode_lnk *inode, cfsrw_sattr *ca)
{
  if (ca->valid_mtime) {
    inode->mtime.seconds = ca->mtime;
    inode->mtime.nseconds = 0;
  }
  if (ca->valid_size)
    inode->attr.size = ca->size; // XXX what do i do?
  if (ca->valid_mode)
    inode->attr.mode = ca->mode;
  if (ca->valid_uid)
    inode->attr.uid = ca->uid;
  if (ca->valid_gid)
    inode->attr.gid = ca->gid;
}

inline bool 
dup_entry (rpc_ptr<cfs_rpc_dirent> *dep, cfs_dirent *e)
{
  if (!dep) 
    return false;
  rpc_ptr<cfs_rpc_dirent> de = *dep;
  while (de && de->name != e->name)
    de = de->nextentry;
  if (de)
    return true;
  return false;
}

inline ref<dbrec>
id2dbrec(bigint id)
{
  str whipme = id.getraw ();
  void *key = (void *)whipme.cstr ();
  int len = whipme.len ();

  ptr<dbrec> q = New refcounted<dbrec> (key, len);
  return q;
}

inline bigint
dbrec2id (ptr<dbrec> r)
{
  str raw = str ( (char *)r->value, r->len);
  bigint ret;
  ret.setraw (raw);
  return ret;
}

inline void 
getrnd (bigint *n, uint len) 
{
  char xx[len];
  rnd.getbytes (xx, len);
  mpz_set_rawmag_be (n, xx, len);
}

inline str 
get_filename (nfspath3 path)
{
  unsigned short *a;
  a = (unsigned short*) path.cstr();
  unsigned short d = (*a) % 240;
  return strbuf ("/var/tmp/cfsrw-scache/%d/%s", d, path.cstr());
}

#endif /* CFSRW_SNAPSHOT_H */

