/*
 *  Copyright (C) 2002-2003  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "dbfe.h"
#include "definc.h"
#include "snapshot.h"

#define WARN(x) if (sn_verbose >= (x)) warn
extern int sn_verbose;

void 
snapshot::store_new (snode *e)
{
  if (e->ent->type == CFS_INODE) {
    cfs_inode *inode = e->ent->inode;
    if ((inode->type == CFSLNK && inode->lnk->nlink > 0) || 
	(inode->type != CFSLNK && inode->reg->nlink > 0))
      store (e->inum);
    else {
      ref<dbrec> q = id2dbrec(e->inum);
      fmap_db->del (q);
    }
  }
}

void 
snapshot::store (bigint n) 
{
  WARN(1) << "store: inum = " << n << "\n";
  snode *e = active_snodes[n];
  if (!e) {
    warn << "snapshot::store: entry does not exist in active_snodes.\n";
    insert_failures++;
    return;
  }
  switch (e->ent->type) {
  case CFS_INODE: {
    ftypero type = e->ent->inode->type;
    cfs_inode *inode = e->ent->inode;
    if (type == CFSLNK)
      store_inode (n, inode);
    else if ((type == CFSREG || type == CFSREG_EXEC))
      store_file (n, inode);
    else
      store_directory (n, inode);
  }
  break;
  default:
    warn << "entry type should not have been in active_snodes.\n";
    insert_failures++;
  }
}

void
snapshot::insert (bigint hash, char *data, size_t len, ptr<cfs_data> d)
{
  outstanding_inserts++;
  inserts++;
  ns_cfsrw::insert_cfs
    (hash, data, len, wrap (mkref(this), &snapshot::insert_cb, false, 0), d);
  xfree (data);
}

void
snapshot::insert_and_addtofmap (char *data, size_t len, bigint inum,
                                ptr<cfs_data> d)
{
  outstanding_inserts++;
  inserts++;
  ns_cfsrw::insert_cfs
    (data, len, wrap (mkref(this), &snapshot::insert_cb, true, inum), d);
  xfree (data);
}

void 
snapshot::insert_cb (bool map, bigint inum, ns_stat status, bigint key)
{
  outstanding_inserts--;
  if (status == NS_OK) {
    if (map) {
      // insert inum->chash to fmap
      ptr<dbrec> k = id2dbrec (inum);
      ptr<dbrec> d = id2dbrec (key);
      if (fmap_db->insert (k, d))
	insert_failures++;
    }
  }
  else {
    warn << "snapshot::insert_cb: dhash insert failed.\n";
    insert_failures++;
  }
  if (outstanding_inserts == 0)
    insert_snapshot ();
}

void 
snapshot::insert_snapshot ()
{
  WARN(1) << "snapshot: " << inserts << " cfs blocks inserted\n";
  if (insert_failures)
    bcb (0, (ptr<snapshot>)0);
  else
    serialize_and_store();
}

bigint
snapshot::store_one_fmap_block (str s)
{
  fmap_block_inserts++;
  bigint n = compute_sha1_hash (s.cstr (), s.len ());
  ns::insert_block
    (n, s.cstr (), s.len (),
     wrap (mkref(this), &snapshot::insert_fmapblock_cb));
  return n;
}

void
snapshot::serialize_and_store ()
{
  fmap_block_inserts = 0;
  fmap_block_insert_failures = 0;

  vec<bigint> inum;
  vec<bigint> chash;
  ptr<dbEnumeration> it = fmap_db->enumerate ();
  ptr<dbPair> p = it->nextElement ();
  while (p) {
    ptr<dbrec> d = fmap_db->lookup (p->key);
    inum.push_back (dbrec2id (p->key));
    chash.push_back (dbrec2id (d));
    p = it->nextElement ();
  }

  bigint next = 0;
  unsigned total = 0;
  while (inum.size () > FMAP_ENTRIES_PER_PAGE) {
    unsigned n=0;
    strbuf sb;
    sint (sb, FMAP_ENTRIES_PER_PAGE);  // entries in this block
    sint (sb, total);                  // remaining entries after this block
    sbigint  (sb, next);                   // address of next block
    while (n < FMAP_ENTRIES_PER_PAGE) {
      bigint i = inum.pop_back ();
      bigint c = chash.pop_back ();
      sbigint (sb, i);
      sbigint (sb, c);
      total++;
      n++;
    }
    next = store_one_fmap_block (sb);
    WARN(1) << "stored extra fmap page at " << next << "\n";
  }

  strbuf sb;
  sbigint (sb, prev ());
  sbigint (sb, rootdir);
  _version.serialize(sb);
  sint (sb, inum.size ()); // entries in this block
  sint (sb, total);        // remaining entries after this block
  sbigint (sb, next);         // address of next block
  while (inum.size () > 0) {
    bigint i = inum.pop_back ();
    bigint c = chash.pop_back ();
    sbigint (sb, i);
    sbigint (sb, c);
  }
  id = store_one_fmap_block (sb);
}

void 
snapshot::insert_fmapblock_cb (ns_stat status, bigint key)
{
  assert (fmap_block_inserts > 0);
  fmap_block_inserts--;
  if (status != NS_OK)
    fmap_block_insert_failures++;

  if (fmap_block_inserts == 0) {
    if (!fmap_block_insert_failures) {
      WARN(2) << "stored snapshot: " << id << "\n";
      bcb (id, mkref (this));
    }
    else {
      warn << "snapshot: dhash insert failed.\n";
      bcb (0, (ptr<snapshot>)0);
    }
  }
}

void
snapshot::store_inode (bigint inum, cfs_inode *inode) 
{
  inode_stored++;

  ptr<cfs_data> dat = New refcounted<cfs_data> (CFS_INODE);
  *(dat->inode) = *inode;

  size_t calllen = 0;
  char *callbuf = NULL;
  xdrsuio x (XDR_ENCODE);
  if (xdr_cfs_data (x.xdrp (), dat)) {
    calllen = x.uio ()->resid ();
    callbuf = suio_flatten (x.uio ());
  }
  insert_and_addtofmap (callbuf, calllen, inum, dat);
}

void 
snapshot::store_directory (bigint inum, cfs_inode *inode)
{
  WARN(1) << "storing directory " << inum << "\n";
  unsigned blk = 0;
  unsigned dirsize = 0;

  for (blk = 0; blk < inode->reg->direct.size(); blk++) {
    bigint n = inode->reg->direct[blk];
    snode *dirblk_ent = active_snodes[n];
    if (!dirblk_ent) // dirblk not modified or loaded
      continue;

    if (dirblk_ent->ent->type != CFS_DIRBLK) {
      warn << "snapshot::store_directory: type is not dirblk.\n";
      insert_failures++;
      return;
    }
    cfs_directory *dir = dirblk_ent->ent->dir;
    cfs_dirent *dirent = dir->entries;
    
    while (dirent) {
      n = dirent->fh;
      // in build from prev, already storing everything in
      // active_snodes
      if (!prev_snapshot) 
        store (n); 
      dirent = dirent->nextentry;
    }
  
    size_t calllen = 0;
    char *callbuf = NULL;
    xdrsuio x (XDR_ENCODE);
  
    if (xdr_cfs_data (x.xdrp (), dirblk_ent->ent)) {
      calllen = x.uio ()->resid ();
      callbuf = suio_flatten (x.uio ());
    }
    n = compute_sha1_hash (callbuf, calllen);
    insert (n, callbuf, calllen, dirblk_ent->ent);
    dirsize += calllen;
    inode->reg->direct[blk] = n;
    WARN(1) << "storing block " << blk << " of directory: " << n << "\n";
  }
   
  // XXX not sure if this matters, but dirsize should be total size of
  // all dirblks, not just the total size of dirblks loaded
  inode->reg->attr.size = dirsize;
  store_inode (inum, inode);
}

void 
snapshot::store_file (bigint inum, cfs_inode *inode)
{
  inode->reg->direct.setsize (CFS_NDIR);

  if (inode->reg->attr.size > 0) {
    // make sure snapshot building will not finish until we do
    outstanding_inserts++;
    str fname = get_filename (inode->reg->path);
    snapshot::a->open (fname, O_RDONLY, 0666,
	               wrap (mkref(this), &snapshot::store_file_open_cb,
			     inum, inode));
  }
  else
    store_inode (inum, inode);
}

void
snapshot::store_file_open_cb (bigint inum, cfs_inode *inode,
                              ptr<aiofh> fd, int err)
{
  if (err) {
    warn << "cannot open snapshot file " << err << "\n";
    insert_failures++;
    outstanding_inserts--; // incremented on file open
    if (outstanding_inserts == 0)
      insert_snapshot ();
    return;
  }

  ptr<aiobuf> buf = snapshot::a->bufalloc (SNAPSHOT_BLOCKSIZE);
  if (!buf) {
    // warn << "no buf, bufwait\n";
    snapshot::a->bufwait
      (wrap (mkref(this), &snapshot::store_file_open_cb,
	     inum, inode, fd, 0));
    return;
  }
  
  vec<bigint>* fhs = New vec<bigint>;
  fd->read (0, buf, wrap (mkref(this), &snapshot::store_file_read_block,
	                  inum, inode, 0, fhs, fd));
}

void
snapshot::store_file_read_block (bigint inum, cfs_inode *inode,
                                 uint32 blocknum, vec<bigint>* fhs,
				 ptr<aiofh> fd, ptr<aiobuf> buf,
				 ssize_t sz, int err)
{
  if (err) {
    warn << "cannot read snapshot file " << err << "\n";
    insert_failures++;
    store_file_close (inum, inode, fhs, fd);
    return;
  }
  
  blocknum++;
  bool done = false;
  if (sz < SNAPSHOT_BLOCKSIZE ||
      blocknum*SNAPSHOT_BLOCKSIZE >= inode->reg->attr.size)
    done = true;

  assert (sz >= 0);
  bigint fh;
  store_file_block (inode, blocknum-1, buf->base(), sz, &fh);
  fhs->push_back (fh);

  if (done)
    store_file_close (inum, inode, fhs, fd);
  else
    fd->read (blocknum*SNAPSHOT_BLOCKSIZE, buf,
	      wrap (mkref(this), &snapshot::store_file_read_block,
		    inum, inode, blocknum, fhs, fd));
}

void 
snapshot::store_file_block (cfs_inode *inode, uint32 blocknum,
                            const char *block, ssize_t size, bigint *fh)
{
  ptr<cfs_data> res = New refcounted<cfs_data> (CFS_FILEBLK);
  res->data->setsize (size);
  memcpy (res->data->base (), block, size);

  size_t calllen = 0;
  char *callbuf = NULL;
  xdrsuio x (XDR_ENCODE);

  if (xdr_cfs_data (x.xdrp (), res)) {
    calllen = x.uio ()->resid ();
    callbuf = suio_flatten (x.uio ());
  }

  bigint n = compute_sha1_hash (callbuf, calllen);
  *fh = n;
 
  if ((inode->reg->mstart == 0 && inode->reg->mend == 0) ||
      ((blocknum+1) * SNAPSHOT_BLOCKSIZE <= inode->reg->mstart) ||
      (blocknum * SNAPSHOT_BLOCKSIZE >= inode->reg->mend)) {
    WARN(2) << "block " << blocknum << " not inserted\n";
    WARN(2) << "content hash is " << n << "\n";
    xfree (callbuf);
  }
  else
    insert (n, callbuf, calllen, res);
}

void
snapshot::store_file_close_cb (int)
{
}

void
snapshot::store_file_close (bigint inum, cfs_inode *inode,
                            vec<bigint>* fhs, ptr<aiofh> fd)
{
  fd->close (wrap (mkref(this), &snapshot::store_file_close_cb));

  inode->reg->mstart = 0;
  inode->reg->mend = 0;

  unsigned blocknum = 0;
  while (blocknum < CFS_NDIR && blocknum < fhs->size()) {
    inode->reg->direct[blocknum] = (*fhs)[blocknum];
    blocknum++;
  }

  if (blocknum >= CFS_NDIR && blocknum < CFS_NDIR+NFH)
    store_sind (fhs, &inode->reg->indirect, blocknum-CFS_NDIR, blocknum);
  if (blocknum >= CFS_NDIR+NFH && blocknum < CFS_NDIR+NFH+NFH*NFH)
    store_dind (fhs, &inode->reg->double_indirect,
	        blocknum-CFS_NDIR-NFH, blocknum);
  if (blocknum >= CFS_NDIR+NFH+NFH*NFH)
    store_tind (fhs, &inode->reg->triple_indirect,
	        blocknum-CFS_NDIR-NFH-NFH*NFH, blocknum);
  store_inode (inum, inode);

  delete fhs;
  outstanding_inserts --; // incremented on file open
      
  if (outstanding_inserts == 0)
    insert_snapshot ();
}

void
snapshot::store_sind (vec<bigint> *fhs, bigint *si_fh,
                      uint32 off, uint32& blocknum)
{
  uint32 blk = off / NFH;
  ptr<cfs_data> sindir = New refcounted<cfs_data> (CFS_INDIR);
  sindir->indir->handles.setsize (NFH);
  
  while (blk < NFH && blocknum < fhs->size()) {
    sindir->indir->handles[blk] = (*fhs)[blocknum];
    blk++;
    blocknum++;
  }

  if (blk > 0) {
    size_t calllen = 0;
    char *callbuf = NULL;
    xdrsuio x (XDR_ENCODE);
    if (xdr_cfs_data (x.xdrp(), sindir)) {
      calllen = x.uio ()->resid ();
      callbuf = suio_flatten (x.uio ());
    }
    bigint n = compute_sha1_hash (callbuf, calllen);
    *si_fh = n;
    insert (n, callbuf, calllen, sindir);
  }
}

void
snapshot::store_dind (vec<bigint> *fhs, bigint *di_fh,
                      uint32 off, uint32& blocknum)
{
  ptr<cfs_data> dindir = New refcounted<cfs_data> (CFS_INDIR);
  dindir->indir->handles.setsize (NFH);
  uint32 sblk = off/NFH;
  uint32 oldbn = blocknum;

  while (sblk < NFH && blocknum < fhs->size()) {
    off = (blocknum-CFS_NDIR)%NFH;
    store_sind
      (fhs, &dindir->indir->handles[sblk], off, blocknum);
    sblk++;
  }

  if (blocknum > oldbn) {
    size_t calllen = 0;
    char *callbuf = NULL;
    xdrsuio x (XDR_ENCODE);
    if (xdr_cfs_data (x.xdrp(), dindir)) {
      calllen = x.uio ()->resid ();
      callbuf = suio_flatten (x.uio ());
    }
    bigint n = compute_sha1_hash (callbuf, calllen);
    *di_fh = n;
    insert (n, callbuf, calllen, dindir);
  }
}

void
snapshot::store_tind (vec<bigint> *fhs, bigint *ti_fh,
                      uint32 off, uint32& blocknum)
{
  ptr<cfs_data> tindir = New refcounted<cfs_data> (CFS_INDIR);
  tindir->indir->handles.setsize (NFH);
  uint32 dblk = off/(NFH*NFH);
  uint32 oldbn = blocknum;

  while (dblk < NFH && blocknum < fhs->size()) {
    off = blocknum%(NFH*NFH);
    store_dind
      (fhs, &tindir->indir->handles[dblk], off, blocknum);
    dblk++;
  }

  if (blocknum > oldbn) {
    size_t calllen = 0;
    char *callbuf = NULL;
    xdrsuio x (XDR_ENCODE);
    if (xdr_cfs_data (x.xdrp(), tindir)) {
      calllen = x.uio ()->resid ();
      callbuf = suio_flatten (x.uio ());
    }
    bigint n = compute_sha1_hash (callbuf, calllen);
    *ti_fh = n;
    insert (n, callbuf, calllen, tindir);
  }
}

